/****************************************************************************
   Copyright & License:
   ====================
   
   Copyright 2009 - 2016 gAGE/UPC & ESA
   
   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at
   
   http://www.apache.org/licenses/LICENSE-2.0
   
   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
 ***************************************************************************/

/****************************************************************************
 * Copyright: gAGE/UPC & ESA
 * Project: EDUNAV GNSS Lab Tool 
 * Author: Pere Ramos-Bosch (group of Astronomy and GEomatics - gAGE/UPC)
 * Curator: Adria Rovira-Garcia ( gAGE/UPC )
 * Developer: Deimos Ibanez Segura ( gAGE/UPC )     
 *          glab.gage @ upc.edu
 * File: dataHandling.h
 * Code Management Tool File Version: 2.2  Revision: 8
 * Date: 2016/04/12
 ***************************************************************************/

/****************************************************************************
 * MODULE DESCRIPTION
 *
 * Name: dataHandling
 * Language: C
 *
 * Purpose:
 *  The purpose of this module is to centralize all the data structures of the 
 *  modules and allow an easy interface to them.
 *
 * Function:
 *  This class is in charge of declaring all the data structures used in all the
 *  modules. It defines all the required functions to access the information 
 *  inside the data structures, and a group of general functions.
 *
 * Dependencies:
 *  None
 *
 * Files modified:
 *  None
 *
 * Files read:
 *  None
 *
 * Resources usage:
 *  See Design Document
 ****************************************************************************/

/****************************
 *       RELEASE_HISTORY
 * -------------
 *          gLAB v1.3.1
 * Release: 2010/01/22
 * Change Log: First public version
 * -------------
 *          gLAB v1.3.2
 * Release: 2010/02/23
 * Change Log: Inertial/Non-inertial velocity switch included
 *             Included DCB data sources
 *             Included SINEX, DCBs and GPS_Receiver_Type data files
 *             Included strict radome option to enable/disable a relaxed matching
 *                  on ANTEX files
 *             Forward/Backward processing included
 * -------------
 *          gLAB v1.4.0
 * Release: 2010/06/21
 * Change Log: C1/P1 flexibility option is added to allow flexible handling of C1 and P1
 *             Included usable measurement index in TMeasOrder structure to select/deselect
 *                  individual measurements
 *             Included prealignCP option to set/unset carrier-phase prealignment 
 * -------------
 *          gLAB v1.4.1
 * Release: 2010/07/09
 * Change Log: No changes in this file.
 * -------------
 *          gLAB v1.4.2
 * Release: 2010/07/31
 * Change Log: No changes in this file.
 * -------------
 *          gLAB v1.4.3
 * Release: 2010/08/31
 * Change Log: Maximum number of filter iterations added.
 * 	       Improved DCB Management.
 * -------------
 *          gLAB v1.4.4
 * Release: 2010/09/22
 * Change Log: Included new DCB management.
 * -------------
 *          gLAB v1.4.5
 * Release: 2010/10/31
 * Change Log: Tropospheric correction option is extended to select independently the nominal and mapping used.
 *             Dilution Of Precision extension: (GDOP, PDOP, TDOP, VDOP, HDOP) new variables.
 * 	       Default precise orbit interpolation degree changed from 9 to 10.
 *   	       SIGMA_INF splitted into different SIGMA_INF_DR and SIGMA_INF_DT: receiver coordinates and clock.
 * -----------
 *          gLAB v2.0.0
 * Release: 2012/12/31
 * Change Log: gLAB version released together with the Educational Book:
 *             "GNSS Data Processing" (Vol. 1 and Vol. 2). ESA TM-23.
 *             Authors: J. Sanz Subirana, J.M. Juan Zornoza and M. Hernandez-Pajares
 *             ISBN: 978-92-9221-885-0 (two volumes)
 *             ISSN: 1013-7076
 *             December 2012.
 * -----------
 *          gLAB v2.2.0
 * Release: 2014/09/22
 * Change Log: Modified structure definitions (by adding new variables) for handling RINEX 3.02 files.
 *             Added new constants.
 *             Added new structures for handling IONEX and F-PPP data
 *             Added functions for initialising IONEX and F-PPP structures (both for ionosphere and DCB data).
 *             Added functions for handling IONEX and F-PPP structures (both for ionosphere and DCB data).
 *             Added function for computing the ionosphere mapping value.
 *             Added function for computing the pierce point value.
 *             Added functions for initialising ESA's troposphere model structures.
 *             Added functions for handling ESA's troposphere model structures.
 * -----------
 *          gLAB v2.2.1
 * Release: 2014/12/23
 * Change Log: No changes in this file.
 * -----------
 *          gLAB v2.2.2
 * Release: 2015/03/02
 * Change Log: No changes in this file. 
 * -----------
 *          gLAB v2.2.3
 * Release: 2015/05/28
 * Change Log: Added variables in TConstellationElement, TAntenna and TOptions structs for processing phase variatons in satellites and receivers.
 * -----------
 *          gLAB v2.2.4
 * Release: 2015/07/01
 * Change Log: No changes in this file.
 * -----------
 *          gLAB v2.2.5
 * Release: 2015/10/20
 * Change Log: No changes in this file
 * -----------
 *          gLAB v2.2.6
 * Release: 2015/11/25
 * Change Log: No changes in this file.
 * -----------
 *          gLAB v2.2.7
 * Release: 2016/03/01
 * Change Log: No changes in this file.
 * -----------
 *          gLAB v2.2.8
 * Release: 2016/04/12
 * Change Log: No changes in this file.
 * -----------
 *       END_RELEASE_HISTORY
 *****************************/

#ifndef DATAHANDLING_H_
#define DATAHANDLING_H_

/* System modules */
#include <string.h>
#include <time.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>

/* Internal types, constants and data */
#define MAX_SATELLITES_IN_VIEW				50
#define MAX_GNSS							6
#define MAX_MEASUREMENTS_PER_SATELLITE		50
#define MAX_INTERPOLATION_DEGREE            19
#define MAX_RECEIVER_NAME					60
#define MAX_ANTENNA_NAME					60
#define MAX_MEASUREMENTS					150
#define MAX_SATELLITES_PER_GNSS				50
#define MAX_FREQUENCIES_PER_GNSS            10
#define MAX_CS_MEMORY						3
#define MAX_FILTER_MEASUREMENTS				4
#define MAX_UNK 							(4+MAX_SATELLITES_IN_VIEW+1)
#define MAX_VECTOR_UNK 						((MAX_UNK)*(MAX_UNK+1)/2)
#define	MAX_PAR								4
#define SIGMA_INF							1e7
#define SIGMA_INF_DR							1e4
#define SIGMA_INF_DT							3e5
#define MAX_INPUT_LINE						4096
#define MAX_MESSAGE_STR                     4096
#define INVALID_CLOCK						999999.999999
#define MAX_DCB_TYPES						4

#define c									299792458.0
#define MU									3.986005e14
#define f0									10.23e6
#define cf0 								(c/f0)
#define EARTH_ROTATION						7.2921151467e-5
#define MAX_CLOCK_DISTANCE					0.3
#define AU									1.49597870e8
#define Pi									3.14159265358979
#define d2r									(Pi/180.)
#define r2d									(180./Pi)
#define EARTH_RADIUS                        6378137.
#define MIN_EARTH_RADIUS                    6356752.
#define KLOBUCHAR_HEIGHT                    350000.
#define BEIDOU_HEIGHT                       375000.
#define GEOMAGNETIC_POLE_LAT                78.3*d2r
#define GEOMAGNETIC_POLE_LON                291.*d2r
#define NUMBERSECONDSPERDAY						86400
#define NUMBERSECONDSPERHOUR						3600
#define NUMBERFPPPMAPS							2
#define TECU2MLI							0.105045952848732
#define DIFFGAL2GPSWEEK							1024

#define GPSmf1								154
#define GPSmf2								120
#define GPSmf5								115
#define GPSmfw 								(GPSmf1-GPSmf2)
#define GALmf1								154
#define GALmf5a 							115
#define GALmf5b								118
#define GALmf5								116.5
#define GALmf6								125
#define SBASmf1								154
#define SBASmf5								115

#define GPSf1 								(GPSmf1*f0)
#define GPSf2								(GPSmf2*f0) 
#define GPSf5								(GPSmf5*f0)
#define GPSfw 								(GPSmfw*f0)
#define GALf1 								(GALmf1*f0)
#define GALf5a								(GALmf5a*f0)
#define GALf5b								(GALmf5b*f0)
#define GALf5								(GALmf5*f0)
#define GALf6								(GALmf6*f0)
#define SBASf1								(SBASmf1*f0)
#define SBASf5								(SBASmf5*f0)

#define GPSl1								(cf0/GPSmf1)
#define GPSl2								(cf0/GPSmf2)
#define GPSl5								(cf0/GPSmf5)
#define GPSlw 								(cf0/GPSmfw)
#define GALl1								(cf0/GALmf1)
#define GALl5a								(cf0/GALmf5a)
#define GALl5b								(cf0/GALmf5b)
#define GALl5								(cf0/GALmf5)
#define GALl6 								(cf0/GALmf6)
#define SBASl1 								(cf0/SBASmf1)
#define SBASl5 								(cf0/SBASmf5)

#define GAMMAGPS12							((GPSf1*GPSf1)/(GPSf2*GPSf2))
#define ALPHAGPS12							(1/(GAMMAGPS12-1))
#define JULIANDAYDIFF							2400000.5

#define FLAG_CYCLESLIP							1
#define FLAG_FAIL							2



// Enumerator for file types
enum fileType {
	ftINVALID,
	ftUNKNOWN,
	ftRINEXobservation,
	ftRINEXbroadcast,
	ftRINEXclocks,
	ftSINEX,
	ftIONEX,
	ftFPPP,
	ftConstellation,
	ftSP3,
	ftANTEX,
	ftGENDCB,
	ftP1C1DCB,
	ftP1P2DCB,
	ftRECEIVERTYPE
};

// Enumerator for different GNSS systems
enum GNSSystem {
	GPS = 0,
	Galileo = 1,
	GLONASS = 2,
	GEO = 3,
	BDS = 4,
	QZSS = 5
};

// Enumerator for different GNSS products type
enum ProductsType{
	BRDC,
	SP3
};

// Enumerator for different data sources
enum Source{
	UNKNOWN_SOURCE,
	RINEX2,
	RINEX3
};

//Enumerator for different time system corrections (Rinex Nav v3 header data)
enum TimeSystemCorrection{
	GAUT = 0,
	GPUT = 1,
	SBUT = 2,
	GLUT = 3,
	GPGA = 4,
	GLGP = 5,
	UNKNOWN_TIME_CORR = 6
};

// Enumerator for different satellite blocks
enum SatelliteBlock{
	UNKNOWN_BLOCK,
	GPS_BLOCK_I,
	GPS_BLOCK_II,
	GPS_BLOCK_IIA,
	GPS_BLOCK_IIR,
	GPS_BLOCK_IIRM,
	GLONASS_BLOCK,
	GLONASS_BLOCK_M
};

// Enumerator for ionospheric model
enum IonoModel {
	UNKNOWN_IM,
	NoIono,
	Klobuchar,
	IONEX,
	FPPP,
	neQuick,
	BeiDou
};

// Enumerators for troposheric model and mapping
enum TropModelNominal {
	UNKNOWN_TMNominal,
	SimpleNominal,
	UNB3Nominal,
	TropoESA
};
enum TropModelMapping {
	UNKNOWN_TMMapping,
	SimpleMapping,
	NiellMapping
};


// Enumerator for P1C1 DCB model
enum P1C1DCBModel {
	p1c1UNKNOWN,
	p1c1NONE,
	p1c1FLEXIBLE,
	p1c1STRICT
};

// Enumerator for P1P2 DCB model
enum P1P2DCBModel {
	p1p2UNKNOWN,
	p1p2NONE,
	p1p2RINEX,
	p1p2DCB,
	p1p2IONEX,
	p1p2FPPP
};

// Enumerator for receiver type
enum ReceiverType {
	rtNA,
	rtUNKNOWN,
	rtCROSS,
	rtNOP1,
	rtNOMINAL
};

// Enumerator for different measurements
enum MeasurementType {
	NA = 0,					// N/A
	C1A =  1, L1A =  2, D1A =  3, S1A =  4, // GAL         A PRS     @ E1
	C1B =  5, L1B =  6, D1B =  7, S1B =  8, // GAL         B I/NAV   @ E1
	C1C =  9, L1C = 10, D1C = 11, S1C = 12, // GP/GL/GA/SB C/A       @ L1/G1/E1
	C1P = 13, L1P = 14, D1P = 15, S1P = 16, // GP/GL       P         @ L1/G1
	C1W = 17, L1W = 18, D1W = 19, S1W = 20, // GPS         Z         @ L1
	C1X = 21, L1X = 22, D1X = 23, S1X = 24, // GAL         B+C       @ E1
	C1Y = 25, L1Y = 26, D1Y = 27, S1Y = 28, // GPS         Y         @ L1
	C1Z = 29, L1Z = 30, D1Z = 31, S1Z = 32, // GAL         A+B+C     @ E1
	C1M = 33, L1M = 34, D1M = 35, S1M = 36, // GPS         M         @ L1
	          L1N = 38, D1N = 39, S1N = 40, // GPS         codeless  @ L1
	C2C = 41, L2C = 42, D2C = 43, S2C = 44, // GP/GL       C/A       @ L2/G2
	C2D = 45, L2D = 46, D2D = 47, S2D = 48, // GPS         L1(C/A)+(P2-P1) @ L2
	C2S = 49, L2S = 50, D2S = 51, S2S = 52, // GPS         L2C (M)   @ L2frequency
	C2L = 53, L2L = 54, D2L = 55, S2L = 56, // GPS         L2C (L)   @ L2
	C2X = 57, L2X = 58, D2X = 59, S2X = 60, // GPS         L2C (M+L) @ L2
	C2P = 61, L2P = 62, D2P = 63, S2P = 64, // GP/GL       P         @ L2/G2
	C2W = 65, L2W = 66, D2W = 67, S2W = 68, // GPS         Z         @ L2
	C2Y = 69, L2Y = 70, D2Y = 71, S2Y = 72, // GPS         Y         @ L2
	C2M = 73, L2M = 74, D2M = 75, S2M = 76, // GPS         M         @ L2
	          L2N = 78, D2N = 79, S2N = 80, // GPS         codeless  @ L2
	C5I = 81, L5I = 82, D5I = 83, S5I = 84, // GP/GA/SB    I         @ L5/E5a
	C5Q = 85, L5Q = 86, D5Q = 87, S5Q = 88, // GP/GA/SB    Q         @ L5/E5a
	C5X = 89, L5X = 90, D5X = 91, S5X = 92, // GP/GA/SB    I+Q       @ L5/E5a 
	C7I = 93, L7I = 94, D7I = 95, S7I = 96, // GAL         I         @ E5b
	C7Q = 97, L7Q = 98, D7Q = 99, S7Q =100, // GAL         Q         @ E5b
	C7X =101, L7X =102, D7X =103, S7X =104, // GAL         I+Q       @ E5b
	C8I =105, L8I =106, D8I =107, S8I =108, // GAL         I         @ E5 (E5a+E5b)
	C8Q =109, L8Q =110, D8Q =111, S8Q =112, // GAL         Q         @ E5 (E5a+E5b)
	C8X =113, L8X =114, D8X =115, S8X =116, // GAL         I+Q       @ E5 (E5a+E5b)
	C6A =117, L6A =118, D6A =119, S6A =120, // GAL         A PRS     @ E6
	C6B =121, L6B =122, D6B =123, S6B =124, // GAL         B         @ E6
	C6C =125, L6C =126, D6C =127, S6C =128, // GAL         C         @ E6
	C6X =129, L6X =130, D6X =131, S6X =132, // GAL         B+C       @ E6
	C6Z =133, L6Z =134, D6Z =135, S6Z =136, // GAL         A+B+C     @ E6
	
	// Combinations
	
	ENDMEAS = 137,		// Internal marker for combinations
	LW = 138,			// GPS         Wide-lane carrier-phase combination (f1*P1+f2*P2)/(f1+f2)
	PW = 139,			// GPS         Short-lane pseudorange combination (f1*L1-f2*L2)/(f1-f2)
	BW = 140,			// GPS         Melbourne-Wübbena combination (Lw-Pw)
	LI = 141,			// GPS         Ionospheric carrier-phase combination (L1-L2)
	PI = 142,			// GPS         Ionospheric pseudorange combination (P2-P1)
	LC = 143,			// GPS         Ionospheric-free carrier-phase combination (f1^2*L1-f2^2*L2)/(f1^2-f2^2)
	PC = 144,			// GPS         Ionospheric-free pseudorange combination (f1^2*P1-f2^2*P2)/(f1^2-f2^2)
	PCC = 145,			// GPS         Ionospheric-free pseudorange combination (f1^2*C1-f2^2*P2)/(f1^2-f2^2)
	G1C = 146,			// GPS         GRAPHIC combination between C1 and L1 (C1+L1)/2
	G1P = 147,			// GPS         GRAPHIC combination between P1 and L1 (P1+L1)/2
	G2C = 148,			// GPS         GRAPHIC combination between C2 and L2 (C2+L2)/2
	G2P = 149			// GPS         GRAPHIC combination between P2 and L2 (P2+L2)/2

};

// Enumerator for different kind of measurements
enum MeasurementKind {
	UNKNOWN_MEAS,
	CarrierPhase,
	Pseudorange,
	SNR,
	Doppler
};

// Enumerator for the different filter parameters
enum FilterParameter {
	PHI_PAR = 0,
	Q_PAR = 1,
	P0_PAR = 2
};

// Enumerator for the different filter unknowns
enum FilterUnknown {
	DR_UNK = 0,
	DT_UNK = 1,
	TROP_UNK = 2,
	BIAS_UNK = 3
};

// Enumerator for the different receiver position sources
enum ReceiverPositionSrc {
	rpUNKNOWN,
	rpRINEX,
	rpSINEX,
	rpSET,
	rpCALCULATE
};

// Enumerator for the different antenna data sources
enum AntennaData {
	adUNKNOWN,
	adNONE,
	adSET,
	adANTEX
};

// Enumerator for ARP
enum ARP {
	arpUNKNOWN,
	arpNONE,
	arpSET,
	arpRINEX
}; 

// Enumerator for the program work mode
enum WorkMode {
	wmUNKNWON,
	wmDOPROCESSING,
	wmSHOWINPUT,
	wmCOMPAREORBITS,
	wmSHOWPRODUCT
};

// Enumerator for satellite velocity
enum SatelliteVelocity {
	svUNKNOWN,
	svINERTIAL,
	svITRF
};

// Enumerator for processing direction
enum ProcessingDirection {
	pFORWARD,
	pBACKWARD
};

// Measurements structure
typedef struct {
//	enum MeasurementType	type;
	double 					value;
	double					model;
	int						LLI;
	int						SNR;
	int						dataFlag;   // see FLAG #defines
} TMeasurement;

// Time handling structure
typedef struct {
	int				MJDN;
	double			SoD;
} TTime;

// Satellite structure
typedef struct {
	enum GNSSystem 		GNSS;
	int 				SVN;
	int 				PRN;
	int					block;
	TMeasurement		meas[MAX_MEASUREMENTS_PER_SATELLITE];
	double				phasePrealign[MAX_MEASUREMENTS_PER_SATELLITE];
	int					arcLength;
	int					hasBeenCycleslip;
	TTime				transTime;
	double				position[3];
	double				velocity[3];
	double				ITRFvel[3];
	double				LoS[3];
	double				clockCorrection;
	double				orientation[3][3];
	double				elevation;
	double				azimuth;
	double				windUpRadAccumReceiver;
	double				windUpRadAccumSatellite;
	double				tropWetMap;
	double				measurementWeights[MAX_FILTER_MEASUREMENTS];
	double				smoothedMeas[MAX_FILTER_MEASUREMENTS];
	double				geometricDistance;
	double				ionoSigma;
	int					validIonex;
	int					hasOrbitsAndClocks;
	int					hasDCBs;
	int					hasDCBsofC1P1;
	int					available;
} TSatellite;

// Antenna structure
typedef struct {
	char			type[MAX_ANTENNA_NAME];
	char			type_ant[MAX_ANTENNA_NAME];
	char			type_radome[MAX_ANTENNA_NAME];
	double			PCO[MAX_FREQUENCIES_PER_GNSS][3];	// Antenna Phase Center Offset [PCO]  (North - East - Up)

        double                  azi1;  //Stored in degrees
        double                  azi2;  //Stored in degrees
        double                  dazi;  //Stored in degrees
        int                     nazi;

        double                  zen1;  //Stored in degrees
        double                  zen2;  //Stored in degrees
        double                  dzen;  //Stored in degrees
        int                     nzen;

        double                  ***noazi;               //Non-azimuth-dependent pattern values (dependent on constellation and frequency)
        double                  ****azi;                //Azimuth-dependent pattern values (dependent on constellation,frequency and azimuth)

} TAntenna;

// Receiver structure
typedef struct {
	char 				name[MAX_RECEIVER_NAME];
	char				type[MAX_RECEIVER_NAME];
	double				aproxPosition[3];		// X - Y - Z
	double				aproxPositionError;
	double				aproxPositionNEU[3];	// Latitude - Longitude - Height
	double				interval;
	double				orientation[3][3];
	double				ARP[3];  // Antenna Reference Point
	enum ReceiverType	recType;
	TAntenna			antenna;
	int					equivalentC1P1;
} TReceiver;

// Measurement order structure
typedef struct {
	enum GNSSystem			GNSS;
	int						nDiffMeasurements;
	int						meas2Ind[MAX_MEASUREMENTS];
	enum MeasurementType	ind2Meas[MAX_MEASUREMENTS_PER_SATELLITE];
	double					conversionFactor[MAX_MEASUREMENTS]; 
							// This factor is to allow changing from carrier-phase cycles
							// to meters.
	int						usable[MAX_MEASUREMENTS];
} TMeasOrder;

// Cycle slip data structure
typedef struct {
	double		nBwmean[MAX_SATELLITES_IN_VIEW];			// Melbourne-Wübbena ambiguity mean in cycles
	double		nBwmean2[MAX_SATELLITES_IN_VIEW];			// Melbourne-Wübbena ambiguity sigma square
	TTime		tPrev[MAX_CS_MEMORY];
	double		LiPrev[MAX_SATELLITES_IN_VIEW][MAX_CS_MEMORY];
	double		L1C1mean[MAX_SATELLITES_IN_VIEW];
	double		L1C1mean2[MAX_SATELLITES_IN_VIEW];
} TCS;

// SatInfo structure
typedef struct {
	TTime		lastEclipse;
} TSatInfo;

// Epoch structure
typedef struct {
	TTime		t;
	int			numSatellites;
	int			prevNumSatellites;
	int			usableSatellites;
	TSatellite	sat[MAX_SATELLITES_IN_VIEW];
	TSatellite	prevSat[MAX_SATELLITES_IN_VIEW];
	TReceiver	receiver;
	int			numGNSS;
	TMeasOrder	measOrder[MAX_GNSS];
	int			flag;
	enum Source	source;
	TCS			cycleslip;
	int			prealign;
	double		solidTideDisplacement[3];
	TSatInfo	satInfo[MAX_GNSS][MAX_SATELLITES_PER_GNSS];
	double		ZTD;
} TEpoch;

// SP3 orbits block
typedef struct {
	TTime		t;
	double		x[3];
	double		v[3];
} TSP3orbitblock;

// SP3 clocks block
typedef struct {
	TTime		t;
	double		clock;
	double		clockrate;
	double		clockdriftrate;
} TSP3clockblock;

// Satellite orbits structure
typedef struct {
	TTime			startTime;
	int				numSatellites;
	int				numRecords;
	int				velocityAvailable;
	int				index[MAX_GNSS][MAX_SATELLITES_PER_GNSS];
	TSP3orbitblock	**block;			// Satellite (from index) & Block
	double			*accuracy;			// Satellite (from index)
	int				interval;
} TSP3orbits;

// Satellite orbits structure
typedef struct {
	TTime			startTime;
	int				numSatellites;
	int				numRecords;
	int				vsizeblock;
	int				index[MAX_GNSS][MAX_SATELLITES_PER_GNSS];
	TSP3clockblock	**block;			// Satellite (from index) & Block
	int				interval;
} TSP3clocks;

// Satellite orbits and clocks SP3 structure
typedef struct {
	TSP3orbits	orbits;
	TSP3clocks	clocks;
} TSP3products;

// Time System Correction block (for Rinex Nav v3)
typedef struct {
        double          acoff[2];               //a0, a1 coefficients pf 1-deg polynomial
        int             timeref;                //Reference time for polynomial (Seconds in GPS/GAL week)
        int             weekref;                //Reference week number (GPS/GAL week)  
} TTimeSystemCorrection;

// Broadcast block
typedef struct {
	TTime		Ttoc;
	TTime		Ttoe;
	TTime		TtransTime;
	int		PRN;
	double		clockbias, clockdrift, clockdriftrate;		// First block row
	int		IODE,IODNav;					// Second block row (IODNav for Galileo)
	double		crs, deltan, M0;
	double		cuc, e, cus, sqrta;				// Third block row
	double		toe, cic, OMEGA, cis;				// Forth block row
	double		i0, crc, omega, OMEGADOT;			// Fifth block row
	double		IDOT;						// Sixth block row
	int		codesOnL2, GPSweek, L2Pdataflag;
	int		dataSources,GALweek;				// Galileo Sixth block row
	double		SVaccuracy;					// Seventh block row
	int		SVhealth;
	double		TGD;
	int		IODC;
	double		SISASignal;					// Galileo Seventh block row
	double		BGDE5a,BGDE5b;
	double		transTime;					// Eigth block row
	int		fitInterval;
	double		transTimeGAL;					// Galileo Eigth block row

	//GLONASS data (RINEX v3)
	double          relfreqbiasGLO,messageframetime;		// First block row
	double		satposX,satvelX,sataccX;			// Second block row
	double          satposY,satvelY,sataccY;		        // Third block row
	int		freqnumber;
	double		satposZ,satvelZ,sataccZ;		        // Fourth block row
	int		ageofoperation;	

	
} TBRDCblock;

// Satellite orbits and clocks broadcast structure
typedef struct {
	//Rinex Nav v2
	double			ionA[4];
	double			ionB[4];
	//Rinex Nav v3
	double         	 	ai[4];   	        //Galileo ionospheric correction
        double                  bdsA[4];                //BeiDou ionospheric correction
        double                  bdsB[4];                //BeiDou ionospheric correction	
	TTimeSystemCorrection 	timeSysCorr[6];		//6 positions, one for each correction type	
	int 			AT_LS;			//Number of leap seconds
	int			AT_LSF;			//Future or past leap seconds
	int			WN_LSF;			//Respective week number
	int			DN;			//Respective day number
	
	//Common blocks
	TBRDCblock		**block;		// Satellite & Block
	int			*numblocks;		// Per satellite
	int			numsats;
	int			index[MAX_GNSS][MAX_SATELLITES_PER_GNSS];
} TBRDCproducts;

// Satellite orbits and clocks generic structure
typedef struct {
	enum ProductsType	type;
	TSP3products		*SP3;
	TBRDCproducts		*BRDC;
} TGNSSproducts;

// Satellite DCB block
typedef struct {
	TTime					startTime;
	TTime					endTime;
	int						available[MAX_GNSS][MAX_SATELLITES_PER_GNSS];
	double 					dcb[MAX_GNSS][MAX_SATELLITES_PER_GNSS];
	enum MeasurementType	measFrom;
	enum MeasurementType	measTo;
} TDCBblock;

// Satellite DCBs
typedef struct {
	TDCBblock				block[MAX_DCB_TYPES];
	int						n;
} TDCBdata;

// Ionex DCB
typedef struct {
	int				DCBavailable;
        double                          DCB[MAX_GNSS][MAX_SATELLITES_PER_GNSS];
        double                          DCBRMS[MAX_GNSS][MAX_SATELLITES_PER_GNSS];
} TIonexDCB;

// FPPP DCB structure
typedef struct {
	int                             DCBavailable;
	//PRN  FRAC_1 FRAC_W IFB(P1-P2)
	double				frac1[MAX_GNSS][MAX_SATELLITES_PER_GNSS];
	double 				fracw[MAX_GNSS][MAX_SATELLITES_PER_GNSS];
	double				IFB[MAX_GNSS][MAX_SATELLITES_PER_GNSS];
} TFPPPDCB;

// TGD generic structure
typedef struct {
	enum P1C1DCBModel		p1c1dcbModel;
	enum P1C1DCBModel		p1p2dcbModel;
	TBRDCproducts			*BRDC;
	TDCBdata			*DCB;
	TIonexDCB			ionexDCB;	
	TFPPPDCB			fpppDCB;
} TTGDdata;

// Receiver data about the code biases managament
typedef struct {
	char 				name[MAX_RECEIVER_NAME];
	enum ReceiverType	type;
} TReceiverData;

// Receiver list structure
typedef struct {
	TReceiverData		*rec;
	int					n;
} TReceiverList;

// Station position
typedef struct {
	double				pos[3];
	char				name[5];
} TStation;

// Station list structure
typedef struct {
	TStation			*station;
	int					n;
} TStationList;

//Ionex Station / PRN / BIAS structure
typedef struct {
	char stationInitials [5];
	char stationcode [11];
	double stationBias;
	double stationRMS;
} TIonexStation;

// Ionospheric map structure
typedef struct {
	double				hgt1;	//Stored in Km
	double				hgt2;	//Stored in Km
	double				dhgt;	//Stored in Km
	int					nhgt;
	
	double				lat1;  //Stored in degrees
	double				lat2;  //Stored in degrees
	double				dlat;  //Stored in degrees
	int					nlat;
	
	double				lon1;  //Stored in degrees
	double				lon2;  //Stored in degrees
	double				dlon;  //Stored in degrees
	int					nlon;
	
	double				*TEC;   // Stored in TECU
	double				*RMS;	// Stored in TECU
	double				*HGT;	// Stored in Km
	TTime				t;
} TIonoMap;

// IONEX structure
typedef struct {
	int					numMaps;
	int				RMSavailable;
	int				HGTavailable;
	TIonoMap			*map;
	TTime				startTime;
	TTime                           endTime;
	int					interval;
	int					exponent;
	int                                     satNumber;
	int					stationNumber;
	int					mapDimension;
	int 					numStationData;
	double				conversionFactor;
	double				elevationCutoff;
	double				baseRadius;	//Stored in Km
	TIonexDCB			ionexDCB;
        char				mappingFunction[5];
	char				observablesUsed[61];
	TIonexStation			*ionexStation;
	
} TIONEX;

// FPPP map structure
typedef struct {
        double                          hgt1;   //Stored in Km
        double                          hgt2;   //Stored in Km
        double                          dhgt;   //Stored in Km (difference between both maps)

        double                          lat1;  //Stored in degrees
        double                          lat2;  //Stored in degrees
        double                          dlat;  //Stored in degrees
	double				nlat;

        double                          lt1;  //Stored in degrees (always 0)
        double                          lt2;  //Stored in degrees (always 360)
        double                          dlt;  //Stored in degrees
	double				nlt;

	double                          *TEC;   // Stored in TECU
        double                          *SIGMA; // Stored in TECU
        TTime                           t;
} TFPPPMap;


// FPPPIONEX structure
typedef struct {
        int                             numMaps;
        TFPPPMap                        *map;
        TTime                           startTime;
        TTime                           endTime;
        int                             interval;
        double                          baseRadius;     //Stored in Km
        char                            mappingFunction[5];
	TFPPPDCB			fpppDCB;
} TFPPPIONEX;

// IONEXPiercePoint structure
typedef struct {
	//Coordinates in X,Y,Z
	double x;
	double y;
	double z;
	//Coordinates in espheric phi,lambda,h
	double fi;		//Stored in degrees
	double longitude;	//Stored in degrees
	double elevation;	//Stored in degrees
	//Localtime position for FPPP maps
	double localtime;	//Stored in degrees
	//Mapping function for current height
	double mappingFunction;
	
} TIONEXPiercePoint;

//TROPOESA structure
typedef struct {
	int lat_pix_positions;			//Number of elements for latitude plus 1 (the beginning is the same as the end)
	int lon_pix_positions;			//Number of elements for longitude plus 1 (the beginning is the same as the end)
	double delta_lat_pix;			//Latitude increase in each position
	double delta_lon_pix;			//Longitude increase in each position
	double lat_pix_init;			//Lowest latitude position in map
	double lat_pix_end;			//Highest latitude position in map
	double lon_pix_init;			//Lowest longitude position in map
	double lon_pix_end;			//Highest longitude position in map
	double am_a_size[3];			//Size of each dimension for am_a
	double es_a_size[3];			//Size of each dimension for es_a
	double ld_a_size[3];			//Size of each dimension for ld_a
	double ps_a_size[3];			//Size of each dimension for ps_a
	double tm_a_size[3];			//Size of each dimension for tm_a
	double es_b_size[4];			//Size of each dimension for es_b
	double ld_b_size[4];			//Size of each dimension for ld_b
	double tm_b_size[4];			//Size of each dimension for tm_b

	double **hgt_pix_values;		//Height values in grid
	double **lnd_values;			//Land values in grid
	double ***am_a_values;			//am_a values in grid
	double ***es_a_values;			//es_a values in grid
	double ***ld_a_values;			//ld_a values in grid
	double ***ps_a_values;			//ps_a values in grid
	double ***tm_a_values;			//tm_a values in grid
	double ****es_b_values;			//es_b values in grid
	double ****ld_b_values;			//ld_b values in grid
	double ****tm_b_values;			//tm_b values in grid

} TTROPOESA;

//TROPOESAParameters Structure
typedef struct {

	//Final parameters
	double gm[4];				//Gravity acceleration in map points [m/s^2]
	double ph[4];				//Water Vapour Pressure in map points [hPa]
	double eh[4];				//Water Vapour Pressure in map points [hPa]
	double tmh[4];				//Average temperature of moist air [K]
	double zwd_map[4];			//Zenith wet delay
	double zhd_map[4];			//Zenith hydrostatic delay
	double ztd_map[4];			//Zenith wet delay + zenith hydrostatic delay

	//Specific time format variables
	double DoY;				//Day of the year [0..364.25] UT
	double hour_minute;			//Hour of the day [0..23.59] [UTC]

	//Variables for data calculation
	double g0[4];				//Acceleration in map points
	double ps[4];				//Air pressure in map points
	double tm[4];				//Mean temperature in map points  
	double am[4];				//Lapse rate of mean temperature in map points
	double es[4];				//Vapou pressure decresase factor in map points
	double ld[4];				//Values of harmonic model for seasonal fluctuations in map points
	double *a;				//Vector for retrieving data from ps_a,tm_a,am_a,es_a,ld_a matrix
	double *b;				//Vector for retrieving data from tm_b,es_b,ld_b
	//Auxiliary variable
	double validpoint[4];			//1->Valid map point 0-> Not a valid point

} TTROPOESAParameters;

// Individual element in constellation structure
typedef struct {
	TTime				tLaunch;
	TTime				tDecommissioned;
	int					PRN;
	int					SVN;
	enum SatelliteBlock	block;
	enum GNSSystem		GNSS;
	//double				phaseCenter[3];
	double				PCO[MAX_FREQUENCIES_PER_GNSS][3];	// Antenna Phase Center Offset [PCO]  (North - East - Up)

        double                  zen1;  //Stored in degrees
        double                  zen2;  //Stored in degrees
        double                  dzen;  //Stored in degrees
        int                     nzen;

        double                  azi1;  //Stored in degrees
        double                  azi2;  //Stored in degrees
        double                  dazi;  //Stored in degrees
        int                     nazi;

        double                  ***noazi;               //Non-azimuth-dependent pattern values (dependent on constellation and frequency)
        double                  ****azi;                //Azimuth-dependent pattern values (dependent on constellation,frequency and azimuth)

} TConstellationElement;

// Data on unknowns for the filter
typedef struct {
  int unk2par[MAX_UNK];
  int par2unk[MAX_PAR];
  int nunk;
  int nunkvector;
  double qnoise[MAX_UNK];
  double phi[MAX_UNK];
} TUnkinfo;

// Filter solution structure
typedef struct {
  double	x[MAX_UNK];
  double	parameters[MAX_UNK];
  double	correlations[MAX_VECTOR_UNK];
  double	dop,GDOP,PDOP,TDOP,HDOP,VDOP;
  int		prevNumSatellites;
} TFilterSolution;

// GNSS Constellation data
typedef struct {
	TConstellationElement	*sat;
	int				numSatellites;
} TConstellation;

// List of antennas
typedef struct {
	TAntenna			*ant;
	int				numAnt;
} TAntennaList;

// Statistics structure
typedef struct {
	double		mean;
	double		mean2;
	int 		n;
} TStat;

// Options structure
typedef struct {
	// Verbose options
	int	printModel;				// printModel = 1			=> Print all the model values (default)
							// printModel = 0			=> Do not print model values

	int	printPrefit;				// printPrefit = 1			=> Print prefilter residuals of measurements (default)
							// printPrefit = 0			=> Do not print prefilter residuals of measurements

	int	printPostfit;				// printPostfit = 1			=> Print postfilter residuals of measurements (default)
							// printPostfit = 0			=> Do not print postfilter residuals of measurements

	int	printFilterSolution;			// printFilterSolution = 1		=> Print the estimation of the parameters of the filter (default)
							// printFilterSolution = 0		=> Do not print the estimation of the parameters of the filter

	int	printInfo;				// printInfo = 1			=> Print general information (default)
							// printInfo = 0			=> Do not print info messages 

	int	printCycleslips;			// printCycleslips = 1			=> Print cycle slips (default)
							// printCycleslips = 0			=> Do not print cycle slips

	int	printSatellites;			// printSatellites = 1			=> Print satellites used for positioning for each epoch (default)
							// printSatellites = 0			=> Do not print satellites used for positioning for each epoch

	int	printInput;				// printInput = 1			=> Print information on measurements directly extracted from the RINEX (default)
							// printInput = 0			=> Do not print information on measurements

	int	printMeas;				// printMeas = 1			=> Print information on measurements and azimuth/elevation
							// printMeas = 0			=> Do not print information on measurements and azimuth/elevation (default)

	int	printOutput;				// printOutput = 1			=> Print output information (default)
							// printOutput = 0			=> Do not print output information

	int	printSatSel;				// printSatSel = 1			=> Print satellite selection information
							// printSatSel = 0			=> Do not print satellite selection information (default)

	int	printSatDiff;				// printSatDiff = 1			=> Print satellite products difference in comparison mode (default)
							// printSatDiff = 0			=> Do not print satellite products difference in comparison mode

	int	printSatStat;				// printSatStat = 1			=> Print statistics for each satellite in comparison mode (default)
							// printSatStat = 0			=> Do not print statistics for each satellite in comparison mode

	int	printSatStatTot;			// printSatStatTot = 1			=> Print summary for all satellites in comparison mode (default)
							// printSatStatTot = 0			=> Do not print summary for all satellites in comparison mode

	int	printSatPvt;				// printSatPvt = 1			=> Print satellite info in show product mode (default)
							// printSatPvt = 0			=> Do not print satellite info in show product mode
								
	enum SatelliteVelocity satVel;      // satVel = svINERTIAL          => Print inertial velocities if satellites
	                                    // satVel = svITRF              => Print ITRF (non-inertial) velocities of satellites 								

	// Processing and modeling options
	enum WorkMode		workMode;		// workMode = wmDOPROCESSING	=> Model and compute the navigation solution (default)
							// workMode = wmSHOWINPUT	=> Read and preprocess data only

	int	includeSatellite[MAX_GNSS][MAX_SATELLITES_PER_GNSS];	// Mask to select/deselect satellites, if =1 satellite selected, if =0 deselected (default, all selected)

	int	usableFreq[MAX_GNSS][MAX_FREQUENCIES_PER_GNSS];		// Mask to select/deselect frequencies, if =1 frequency selected, if =0 deselected (default, all selected)

	enum ReceiverPositionSrc	receiverPositionSource;			// receiverPositionSource = rpRINEX	=> Use the receiver position specified in the RINEX file for apriori position (default)
										// receiverPositionSource = rpSET	=> Specify the receiver apriori position (default)

	double	aprioriReceiverPosition[3];	// 3D vector with the apriori position of the receiver in meters. This parameter is required when receiverPositionSource == rpSET

	enum AntennaData	antennaData;			// antennaData = adNONE			=> Do not obtain nor use and receiver antenna information (default)
								// antennaData = adSET			=> Specify the receiver antenna data (such as phase center)
								// antennaData = adANTEX		=> Try to read the characteristics of the antenna from the ANTEX file

	double	receiverPCO[MAX_FREQUENCIES_PER_GNSS][3];	// 3D vector with the receiver antenna phase center [NEU] for each frequency. This parameter is required when antennaData == adSET

	int	strictradome;					// strictradome				=> When using ANTEX to correct the receiver phase center, this option will force a perfect
        		                        		        //  match between the antenna name and radome reported in the receiver RINEX file and the ones in the ANTEX file. If this option
                        		               			//  is disabled (=0), and the radome is not found, the corrections will be used considering the radome "NONE" (default=0)

	enum ARP	ARPData;				// ARPData = arpNONE			=> Do not apply any kind of Antenna Reference Point correction (default) 
								// ARPData = arpSET			=> Specify the ARP correction
								// ARPData = arpRINEX			=> Get the ARP from the RINEX header ('ANTENNA: DELTA H/E/N' field)

	double	receiverARP[3];					// 3D vector with the receiver Antenna Referece Point [NEU]. This parameter is required when ARPData == arpSET

	int		flexibleC1P1;				// flexibleC1P1 = 1				=> Allows a flexible handling of C1/P1 (C1C/C1P), i.e. providing one, one the RINEX observation file does not provide the other (default)
								// flexibleC1P1 = 0				=> Does not allow a flexible handling of C1/P1 (C1C/C1P)
	int	prealignCP;					// prealignCP = 1				=> Prealign carrier phase measurements with its corresponding pseudorange
								// prealignCP = 0				=> Take the carrier phase measurements "as is" without any kind of prealigment

	int	csLi;						// csLi = 1					=> Use the geometric-free carrier-phase combination to detect cycle-slips (default)
								// csLi = 0					=> Do not use the geometric-free carrier-phase combination to detect cycle-slips

	double	csLimin;					// Minimum threshold between estimated and measured Li values (default 0.034)

	double	csLimax;					// Maximum threshold between estimated and measured Li values (default 0.08)

	double	csLit;						// Time constant to set the threshold between maximum and minimum (default 60)

	int	csBW;						// csBW = 1					=> Use the Melbourne-Wübbena combination to detect cycle-slips (default)
								// csBW = 0					=> Do not use the Melbourne-Wübbena combination to detect cycle-slips

	double	csBWmin;					// Minimum threshold between estimated and measured BW values in relation to its standard deviation (default 0.9)

	double	csBWmax;					// Maximum threshold between estimated and measured BW values in relation to its standard deviation (default 18)

	double	csBWslope;					// Relation between estimated and measured BW values in relation to its standard deviation (default 3)

	int	csL1C1;						// csL1C1 = 1					=> Use the L1-C1 difference to detect cycle-slips [mainly for single-frequency receivers]
								// csL1C1 = 0					=> Do not use the L1-C1 difference to detect cycle-slips (default)

	double	csL1C1slope;				// Relation between estimated and measured (L1-C1) values in relation to its standard deviation (default 5)

	int	csL1C1window;				// Number of epochs to limit the L1-C1 ionosphere divergence (default 100)

	double	csL1C1max;				// Maximum standard deviation for the threshold calculation in the L1-C1 (default 15)

	double  csL1C1init;                             // Initial standard deviation for the threshold calculation in the L1-C1 (default 1.0)

	int		timeTrans;				// timeTrans = 1				=> Use transmission time for satellite positions (default)
								// timeTrans = 0				=> Use reception time for satellite positions

	int		earthRotation;			// earthRotation = 1			=> Correct the Earth rotation from travelTime (default 1)

	int		satellitePhaseCenter;		// satellitePhaseCenter = 1		=> Correct satellite phase centers in SP3 (default 1)

        int             satellitePhaseVar;              // 0 => Do not use zenital corrections for the antenna in the ANTEX file (default)
                                                        // 1 => Use zenital corrections for the satellite in the ANTEX file

        int             receiverPhaseVar;               // 0 => Do not use zenital corrections for the antenna in the ANTEX file (default)           
                                                        // 1 => Use zenital corrections for the antenna in the ANTEX file, both azimut and non azimut corrections

	int		relativisticCorrection;		// relativisticCorrection = 1	=> Correct the relativisty effects on the signal (default 1)

	int		windUpCorrection;		// windUpCorrection = 1			=> Correct the windup from phase measurements (default 1)

	int		gravitationalCorrection;	// gravitationalCorrection = 1  => Correct the gravitational delay from the general relativity (default 1) 

	int		solidTidesCorrection;		// solidTides = 1				=> Correct the solid tides effect (default 1)

	int		satelliteClockCorrection;	// satelliteClockCorrection = 1  => Correct the satellite clock error from the clocks products file (default 1)

	int		troposphericCorrection;		// troposphericCorrection = 1	=> Correct the troposheric delay (default 1)

	enum IonoModel		ionoModel;		// ionoModel = NoIono			=> Do not apply any ionospheric correction (default)
							// ionoModel = Klobuchar		=> Klobuchar standard model for ionospheric correction
							// ionoModel = NeQuick			=> NeQuick model for ionospheric correction
							// ionoModel = BeiDou			=> BeiDou model for ionospheric correction
							// ionoModel = IONEX			=> IONEX maps for ionospheric correction
							// ionoModel = FPPP			=> FPPP maps for ionospheric correction

	double		ionoRMS;			// Maximum error value for ionosphere model, only on FPPP (default 1)
	

	enum TropModelNominal		tropNominal;	// tropNominal 				=> Tropospheric nominal computation
							// tropNominal = Simple			=> Simple tropospheric nominal correction
							// tropNominal = UNB3			=> UNB3 tropospheric nominal correction (default)
							// tropNominal = TropoESA		=> ESA tropospheric correction

	enum TropModelMapping		tropMapping;	// tropMapping  			=> Tropospheric mapping
							// tropMapping = Simple			=> Simple tropospheric mapping 
							// tropMapping = Niell			=> Niell tropospheric mapping (default)


        enum P1C1DCBModel   p1c1dcbModel;   // p1c1dcbModel = p1c1NONE      => No correction of C1-P1 DCBs, no identification of C1 and P1
                                            // p1c1dcbModel = p1c1FLEXIBLE  => No correction of C1-P1 DCBs, identification of C1 and P1
                                            // p1c1dcbModel = p1c1STRICT    => Correction of C1-P1 DCBs, identification of C1 and P1 obtained from precise DCB files

        enum P1P2DCBModel   p1p2dcbModel;   // p1p2dcbModel = p1p2NONE      => No correction of P1-P2 DCBs 
                                            // p1p2dcbModel = p1p2RINEX     => Correction of P1-P2 DCBs, obtained from RINEX files
                                            // p1p2dcbModel = p1p2DCB       => Correction of P1-P2 DCBs, obtained from precise DCB files
                                            // p1p2dcbModel = p1p2IONEX     => Correction of P1-P2 DCBs, obtained from IONEX files
                                            // p1p2dcbModel = p1p2FPPP      => Correction of P1-P2 DCBs, obtained from FPPP files

	int	decimate;				//			=> Decimate the data to this interval (decimate = 0 do not decimate, default = 0)

	enum MeasurementType	smoothMeas[MAX_FILTER_MEASUREMENTS];	// Vector containing the measurements used to smooth each of the filter measurements (default: NA and NA)

	int	smoothEpochs;			// Value to limit the memory of the smoothing (default: 0, which means no smoothing)

	int	orbitInterpolationDegree;	// Orbit interpolation degree for precise products (default 10)

	int	clockInterpolationDegree;	// Clock interpolation degree for precise products (default 0 - no interpolation)

	int     satelliteHealth;            // satelliteHealth = 1          => Use the satellite healthy flag of the navigation message (default)
	                                    // satelliteHealth = 0          => Ignore the satellite healthy flag of the navigation message

	// Filtering options
	int	totalFilterMeasurements;	// Total number of measurements per station that will be provided to the filter (default 2)

	enum MeasurementType	measurement[MAX_FILTER_MEASUREMENTS];		// Vector containing the measurements given to the filter (default: PC and LC)

	double	fixedMeasurementWeight[MAX_FILTER_MEASUREMENTS];		// Vector containing the weights of the measurements for the filter (default: 1 and 0.01)

	double	elevationMeasurementWeight[MAX_FILTER_MEASUREMENTS][2];		// Vector containing the weights of the measurements for the filter depending on elevation
										// The weigth is computed as:  fixedMeasurementWeight + elevationMeasurementWeight[0]*e^(elevationMeasurementWeight[1])
										// Defaults are elevationMeasurementWeight[0] = 0 elevationMeasurementWeight[1] = 30 

	int	minArcLength;				// Minimum arc length to get the measure in the filter (default 3)

	double	maxdr2;					// Max distance (squared) from initial approximate position to estimated position (default 10000)

	int	estimateTroposphere;			// estimateTroposphere = 1		=> Estimate the troposphere in the filter (default 1)

	int	useSigmaIono;				// useSigmaIono = 1	=> Use ionosphere Sigma values in filter when using IONEX or FPPP model (default 1)

	int	usePhase;				// Internal check. usePhase = 1 if any measurement is of carrier-phase type (default 1)

	int	filterIterations;			// This will set the number of forward/backward runs in the filter, being 1 only forward (dafault 1)

	double	filterParams[3][MAX_PAR];	// filterParams stores the data of the Kalman filter. The first dimension is the parameter type of the
										// filter, the second is the specific unknown it is refering to:
										// To access the first dimension: PHI_PAR, Q_PAR and P0_PAR
										// Phi is the propagation of the uknown between epochs
										// Q is the process noise included in the propagation
										// P0 is the initial standard deviation of the parameter
										// To access the second dimension: DR_UNK, DT_UNK, TROP_UNK and BIAS_UNK
										// DR: Position unknowns
										// DT: Receiver clock unknown
										// TROP: Troposphere estimation unknown (if it is estimated)
										// BIAS: Carrierphase ambiguities (if carrierphase is used)
										// Default values are as follows:
										// PHI_DR:   1 [static positioning]       0 [kinematic positioning]
										// PHI_DT:   0
										// PHI_TROP: 1
										// PHI_BIAS: 1
										// Q_DR:     0 [static positioning]       inf [kinematic positioning]
										// Q_DT:     inf
										// Q_TROP:   1e-4   (This value is the variation per hour)
										// Q_BIAS:   0
										// P0_DR:    inf
										// P0_DT:    inf
										// P0_TROP:  0.5^2
										// P0_BIAS:  20^2

	double	elevationMask;				// 			=> Satellites below this elevation mask (in radians) will be discarded (default 0)

	int	discardEclipsedSatellites;	// discardEclipsedSatellites = 1		=> Discard satellites under Earth eclipes (default in PPP)					
						// discardEclipsedSatellites = 0		=> Do NOT discard satellites under Earth eclipes (default in Standalone)

	int   maxKalmanIterations;		// Maximum iterations of the filter trying to converge to a solution
	
} TOptions;

char messagestr[MAX_MESSAGE_STR];

/**************************************
 * Declarations of internal operations
 **************************************/

// Initialisations
void initOptions (TOptions *options);
void initSatellite (TEpoch *epoch, int satIndex);
void initEpoch(TEpoch *epoch);
void initGNSSproducts(TGNSSproducts *products);
void initBRDCproducts(TBRDCproducts *BRDC);
void initBRDCblock(TBRDCblock *BRDCblock);
void initSP3products(TSP3products *SP3);
void initSP3orbitblock(TSP3orbitblock *SP3orbitblock);
void initSP3clockblock(TSP3clockblock *SP3clockblock);
void initIONEXDCB(TIonexDCB *IonexDCB);
void initIONEX(TIONEX *IONEX);
void initIonoMap(TIonoMap *ionoMap);
void initIonexStation(TIonexStation *ionexStation);
void initFPPPDCB(TFPPPDCB *FPPPDCB);
void initFPPPIONEX(TFPPPIONEX *FPPP);
void initFPPPMap(TFPPPMap *FPPPMap);
void initFPPPMapData(TFPPPMap *FPPPMap);
void initIonexPiercePoint(TIONEXPiercePoint *ionexPiercePoint);
void initTropoESAData(TTROPOESA *TropoEsa);
void initTropoESAParameters(TTROPOESA *TropoEsa,TTROPOESAParameters *ESAParam);
void initFilterSolution (TFilterSolution *solution);
void initConstellation(TConstellation *constellation);
void initConstellationElement(TConstellationElement *sat);
void initAntennaList(TAntennaList *antennaList);
void initAntenna(TAntenna *antenna);
void initStat(TStat *stat);
void initTGD (TTGDdata *tgdData, enum P1C1DCBModel p1c1dcbModel, enum P1P2DCBModel p1p2dcbModel);
void initDCB (TDCBdata *dcbData);
void initReceiverList (TReceiverList *recList);
void initStationList (TStationList *stationList);

// General
void changeBase (double base[3][3], double vector[3]);
double min (double a, double b);
double max (double a, double b);
double modulo (double a, double mod);
void unitary(double *vect);
void vectProd(double *a, double *b, double *res);
double scalarProd(double *a, double *b);
char *trim(char *line);
char *getstr(char *out,char *line,int ini, int length);
void getnumericalelements(char *line, int *numelements, double *out);
void getwordelements(char *line, int *numelements, char *datavector);

// Measurements
enum MeasurementKind whatIs(enum MeasurementType meas);
double getLambda(enum GNSSystem GNSS, enum MeasurementType meas);
double getFrequency(enum GNSSystem GNSS, enum MeasurementType meas);
int getFrequencyInt(enum MeasurementType meas);
enum MeasurementType measstr2meastype(char *str);
char *meastype2measstr(enum MeasurementType meas);
enum GNSSystem gnsschar2gnsstype(char system);
char *gnsstype2gnssstr(enum GNSSystem GNSS);
char gnsstype2char(enum GNSSystem GNSS);
enum SatelliteBlock satblockstr2satblock(char *str);
char *satblock2satblockstr(enum SatelliteBlock satBlock);
enum TimeSystemCorrection timecorrstr2timecorrtype(char *str);
char *timecorrtype2timecorrstr(enum TimeSystemCorrection timecorr);

// Epoch handling
int getMeasModelValue(TEpoch *epoch, enum GNSSystem system, int PRN, enum MeasurementType measType, double *measurement, double *model);
double getModelValue(TEpoch *epoch, enum GNSSystem system, int PRN, enum MeasurementType measType);
double getMeasurementValue(TEpoch *epoch, enum GNSSystem system, int PRN, enum MeasurementType measType);
int satWas(enum GNSSystem GNSS,int PRN,TSatellite *prevSat,int prevNumSatellites);

// OC Products handling
TTime getProductsFirstEpochBRDC (TGNSSproducts *products);
TTime getProductsFirstEpochSP3 (TGNSSproducts *products);
TTime getProductsFirstEpoch (TGNSSproducts *products);
TTime getProductsLastEpochBRDC (TGNSSproducts *products);
TTime getProductsLastEpochSP3 (TGNSSproducts *products);
TTime getProductsLastEpoch (TGNSSproducts *products);
TBRDCblock *selectBRDCblock (TBRDCproducts *products, TTime *t, enum GNSSystem GNSS, int PRN, TOptions *options);

// Time handling
int MJDN(struct tm *tm);
void t2tm(TTime *t, struct tm *tm, double *seconds);
void t2doy(TTime *t, int *year, double *doy);
char *t2doystr (TTime *t);
TTime gpsws2ttime(int GPSweek, double SoW);
TTime cal2t (int year, int month, int day, int hour, int minute, double second);
double tdiff(TTime *a, TTime *b);
TTime tdadd(TTime *a, double b);
int ESA_loc_month(double doy, int *im, double *id);

// IONEX
int numSteps(double a1, double a2, double da);
void allocateIonoMap (TIonoMap *ionoMap);
double E(int exponent);
int	conv3Dto1D (int i, int j, int k, int ni, int nj);
int ionoConv3Dto1D (TIonoMap *ionoMap, int ihgt, int ilat, int ilon);
int ionoHLLto1D (TIonoMap *ionoMap, double hgt, double lat, double lon);
int setIonoValue(TIonoMap *ionoMap, int type, double hgt, double lat, double lon, double value);
double getIonoValue(TIonoMap *ionoMap, int type, double hgt, double lat, double lon);
int getIonoMapTotalSize(TIonoMap *ionoMap);
int FPPPConv3Dto1D (TFPPPMap *FPPPMap, int ihgt, int ilat, int ilt);
int FPPPHLLto1D (TFPPPMap *FPPPMap, double hgt, double lat, double lt);
int setFPPPValue(TFPPPMap *FPPPMap, int type, double hgt, double lat, double lt, double value);
double getFPPPValue(TFPPPMap *FPPPMap, int type, double hgt, double lat, double lt);
int getFPPPMapTotalSize(TFPPPMap *FPPPMap);
void getIonoPiercePoint(TEpoch *epoch, TIONEXPiercePoint *ionexPiercePoint, int satIndex, double earthRadius, double altitude);
void getMappingFunction(TIONEXPiercePoint *ionexPiercePoint, double earthRadius, double altitude, char functionType[5], TOptions *options);

// Constellation
TConstellationElement* getConstellationElement(enum GNSSystem GNSS, int PRN, TTime *t, TConstellation *constellation);

// Station
TStation *getStation(char *name, TStationList *stationList);

// Statistics
void addStat (TStat *stat, double value);

#endif /*DATAHANDLING_H_*/
