/****************************************************************************
   Copyright & License:
   ====================
   
   Copyright 2009 - 2016 gAGE/UPC & ESA
   
   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at
   
   http://www.apache.org/licenses/LICENSE-2.0
   
   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
 ***************************************************************************/

/****************************************************************************
 * Copyright: gAGE/UPC & ESA
 * Project: EDUNAV GNSS Lab Tool 
 * Author: Pere Ramos-Bosch (group of Astronomy and GEomatics - gAGE/UPC)
 * Curator: Adria Rovira-Garcia ( gAGE/UPC )
 * Developer: Deimos Ibanez Segura ( gAGE/UPC )              
 *          glab.gage @ upc.edu
 * File: model.c
 * Code Management Tool File Version: 2.2  Revision: 8
 * Date: 2016/04/12
 ***************************************************************************/

/****************************************************************************
 * MODULE DESCRIPTION
 *
 * Name: model
 * Language: C
 *
 * Purpose:
 *  The purpose of this module is to fully model an epoch of data
 *
 * Function:
 *  This class is in charge of modeling an epoch of data. This module is fully
 *  configurable through a Toptions structure, being able to connect or disconnect
 *  each model separately.
 *
 * Dependencies:
 *  dataHandling, preprocessing
 *
 * Files modified:
 *  None
 *
 * Files read:
 *  None
 *
 * Resources usage:
 *  See Design Document
 ****************************************************************************/

/****************************
 *       RELEASE_HISTORY
 * -------------
 *          gLAB v1.3.1
 * Release: 2010/01/22
 * Change Log: First public version
 * -------------
 *          gLAB v1.3.2
 * Release: 2010/02/23
 * Change Log: Enhanced parsing of antenna in ANTEX file, it is now able to 
 *                  separate the name from the dome
 *             Inertial/non-inertial velocity switch included
 * -------------
 *          gLAB v1.4.0
 * Release: 2010/06/21
 * Change Log: Changed the way rotation during signal flight time is done, it
 *                  is now possible to set/unset this rotation, and the reception/
 *                  transmission time separately
 * -------------
 *          gLAB v1.4.1
 * Release: 2010/07/09
 * Change Log: No changes in this file.
 * -------------
 *          gLAB v1.4.2
 * Release: 2010/07/31
 * Change Log: No changes in this file.
 * -------------
 *          gLAB v1.4.3
 * Release: 2010/08/31
 * Change Log: Improved DCB Management. 
 * -------------
 *          gLAB v1.4.4
 * Release: 2010/09/22
 * Change Log: Included new DCB management.
 * -------------
 *          gLAB v1.4.5
 * Release: 2010/10/31
 * Change Log: No changes in this file.
 * -----------
 *          gLAB v2.0.0
 * Release: 2012/12/31
 * Change Log: gLAB version released together with the Educational Book:
 *             "GNSS Data Processing" (Vol. 1 and Vol. 2). ESA TM-23.
 *             Authors: J. Sanz Subirana, J.M. Juan Zornoza and M. Hernandez-Pajares
 *             ISBN: 978-92-9221-885-0 (two volumes)
 *             ISSN: 1013-7076
 *             December 2012.
 * -----------
 *          gLAB v2.2.0
 * Release: 2014/09/22
 * Change Log: Added functions for BeiDou, IONEX, Fast-PPP and NeQuick ionosphere models.
 *             Added functions for using Fast-PPP and IONEX DCBs.
 *             Added functions for ESA's troposphere model.
 *                      (Not available in this version due to ESA's restrictions)
 *             Modified Klobuchar model function.
 * -----------
 *          gLAB v2.2.1
 * Release: 2014/12/23
 * Change Log: No changes in this file.
 * -----------
 *          gLAB v2.2.2
 * Release: 2015/03/02
 * Change Log: No changes in this file.
 * -----------
 *          gLAB v2.2.3
 * Release: 2015/05/28 
 * Change Log: Added satellitePhaseCenterVarCorrection and receiverPhaseCenterVarCorrection functions for phase variation correction.
 * -----------
 *          gLAB v2.2.4
 * Release: 2015/07/01
 * Change Log: Fixed iono sigma not being converted to meters.
 * 	       Corrected bug in troposphericCorrectionMOPS function where seconds of day where added twice.
 * -----------
 *          gLAB v2.2.5
 * Release: 2015/10/20
 * Change Log: Fixed iono sigma not being added to the total sigma weight.
 * -----------
 *          gLAB v2.2.6
 * Release: 2015/11/25
 * Change Log: No changes in this file.
 * -----------
 *          gLAB v2.2.7
 * Release: 2016/03/01
 * Change Log: Fixed iono sigma not being transformed from TECUs to meters of the frequency of the measurement in the filter
 * -----------
 *          gLAB v2.2.8
 * Release: 2016/04/12
 * Change Log: Fixed computation of Galileo elevation and azimuth.
 * -----------
 *       END_RELEASE_HISTORY
 *****************************/

/* External classes */
#include "model.h"

/**************************************
 * Declarations of internal operations
 **************************************/
 
 /*****************************************************************************
 * Name        : getSatelliteOrientationACR
 * Description : Obtains the 3 vectors determining the reference frame of a GNSS 
 *               satellite in the reference system:
 *                  Z axis => Radial: Pointing to the opposite direction of Earth
 *                  X axis => Along-track: Perpendicular to Z in the plane 
 *                            Velocity-Satellite-Earth (similar to velocity)
 *                  Y axis => Cross-track: Completing the system
 * Parameters  :
 * Name                           |Da|Unit|Description
 * double  *position               I  m    Satellite position
 * double  *velocity               I  m    Satellite velocity
 * double  orientation[3][3]       O  N/A  Vectors determining the attitude of 
 *                                         the satellite
 *****************************************************************************/
void getSatelliteOrientationACR (double *position, double *velocity, double orientation[3][3]) {
	double		posUni[3],velUni[3];
	int			i;
	
	memcpy(posUni,position,sizeof(double)*3);
	unitary(posUni);
	memcpy(velUni,velocity,sizeof(double)*3);
	unitary(velUni);
	
	for (i=0;i<3;i++)
		orientation[2][i] = posUni[i];

	vectProd(posUni,velUni,orientation[1]);
	vectProd(orientation[1],orientation[2],orientation[0]);
	
	unitary(orientation[0]);
	unitary(orientation[1]);
}
 
/*****************************************************************************
 * Name        : getSatelliteOrientation
 * Description : Obtains the 3 vectors determining the attitude of a GNSS satellite.
 *                  Z axis => Nadir (pointing to Earth)
 *                  X axis => Perpendicular to Z in the plane Sun-Satellite-Earh
 *                  Y axis => Completing the system (along the solar panel)
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TTime  *t                       I  N/A  Reference time
 * double  *position               I  m    Satellite position
 * double  orientation[3][3]       O  N/A  Vectors determining the attitude of 
 *                                         the satellite
 *****************************************************************************/
void getSatelliteOrientation (TTime *t, double *position, double orientation[3][3]) {
	double		sunPos[3],relSunPos[3];
	double		posUni[3];
	int			i;
	
	findSun(t,sunPos);
	relSunPos[0] = sunPos[0] - position[0];
	relSunPos[1] = sunPos[1] - position[1];
	relSunPos[2] = sunPos[2] - position[2]; 
	unitary(relSunPos);
			
	memcpy(posUni,position,sizeof(double)*3);
	unitary(posUni);
	
	for (i=0;i<3;i++)
		orientation[2][i] = -posUni[i];

	vectProd(orientation[2],relSunPos,orientation[1]);
	vectProd(orientation[1],orientation[2],orientation[0]);
	
	unitary(orientation[0]);
	unitary(orientation[1]);
}

/*****************************************************************************
 * Name        : fillSatelliteOrientation
 * Description : Obtains the 3 vectors determining the attitude of a GNSS 
 *               satellite and fills it to TEpoch class. 
 *                  Z axis => Nadir (pointing to Earth)
 *                  X axis => Perpendicular to Z in the plane Sun-Satellite-Earh
 *                  Y axis => Completing the system (along the solar panel)
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  IO N/A  TEpoch structure
 * int  satIndex                   I  N/A  Index of the satellite
 *****************************************************************************/
void fillSatelliteOrientation (TEpoch *epoch, int satIndex) {
	double	orientation[3][3];
	
	getSatelliteOrientation (&epoch->t, epoch->sat[satIndex].position, orientation);
	
	memcpy(epoch->sat[satIndex].orientation,orientation,sizeof(double)*3*3);
}

/*****************************************************************************
 * Name        : XYZ2NEU
 * Description : Obtains the orientation of a fixed ground station.
 *                  Z axis => Zenith (Up)  [m]
 *                  X axis => North                            - Latitude  [rad]
 *                  Y axis => Completing the system (East)     - Longitude [rad]
 * Parameters  :
 * Name                           |Da|Unit|Description
 * double  *positionXYZ            I  m    Station position in XYZ
 * double  *positionNEU            O  rd/m Station position in NEU
 *                                         0=>N(Lat)  1=>E(Lon)   2=>U
 *****************************************************************************/
void XYZ2NEU (double *positionXYZ, double *positionNEU) {
	const double	tol = 1e-11;
	const double	a = EARTH_RADIUS;
	const double	f = 1/298.257223563;
	const double	b = a*(1-f);
	const double	e2 = (a*a-b*b)/(a*a);
	double			p;
	double			longitude;
	double			latitude,latitudePrev;
	double			xn;
	double			heightOverElipsoide;
	// WGS84 parameters (in meters)

	longitude = atan2(positionXYZ[1],positionXYZ[0]);
	p = sqrt(positionXYZ[0]*positionXYZ[0]+positionXYZ[1]*positionXYZ[1]);
	if (p==0) p=0.00001;
	latitude = atan((positionXYZ[2]/p)/(1-e2));

	latitudePrev = latitude + tol*2; // To overbound the following while
	while (fabs(latitude-latitudePrev) > tol) {
		latitudePrev = latitude;
		xn = a*a/sqrt((a*cos(latitude))*(a*cos(latitude)) + (b*sin(latitude))*(b*sin(latitude)));
		heightOverElipsoide = p/cos(latitude)-xn;
		if ((xn+heightOverElipsoide)!=0)
			latitude = atan((positionXYZ[2]/p)/(1-e2*(xn/(xn+heightOverElipsoide))));
		else
			latitude = 0;
	}
	
	positionNEU[0] = latitude;
	positionNEU[1] = longitude;
	positionNEU[2] = heightOverElipsoide;
}

/*****************************************************************************
 * Name        : fillGroundStationNEU
 * Description : Fills in a TEpoch structure the NEU of a station from its 
 *               XYZ position.
 *                  Z axis => Zenith (Up)  [m]
 *                  X axis => North                            - Latitude  [rad]
 *                  Y axis => Completing the system (East)     - Longitude [rad]
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  IO N/A  TEpoch structure
 *****************************************************************************/
void fillGroundStationNEU (TEpoch *epoch) {
	XYZ2NEU(epoch->receiver.aproxPosition,epoch->receiver.aproxPositionNEU);
}

/*****************************************************************************
 * Name        : getGroundStationOrientation
 * Description : Obtains the 3 vectors determining the attitude of fixed ground
 *               station.
 * Parameters  :
 * Name                           |Da|Unit|Description
 * double  *positionNEU            I  N/A  Position of the station in NEU
 * double  orientation[3][3]       O  N/A  Vectors containing the attitude of
 *                                         the station
 *                                         0=>X(N)    1=>Y(E)    2=>Z(U) 
 *****************************************************************************/
void getGroundStationOrientation (double *positionNEU, double orientation[3][3]) {
	// Local coordinate frame
	
	// North
	orientation[0][0] = -sin(positionNEU[0])*cos(positionNEU[1]);
	orientation[0][1] = -sin(positionNEU[0])*sin(positionNEU[1]);
	orientation[0][2] =  cos(positionNEU[0]);
	
	// East
	orientation[1][0] = -sin(positionNEU[1]);
	orientation[1][1] =  cos(positionNEU[1]);
	orientation[1][2] =  0;
	
	// Up (Zenith)
	orientation[2][0] =  cos(positionNEU[0])*cos(positionNEU[1]);
	orientation[2][1] =  cos(positionNEU[0])*sin(positionNEU[1]);
	orientation[2][2] =  sin(positionNEU[0]);
}

/*****************************************************************************
 * Name        : fillGroundStationOrientation
 * Description : Fills the orientation of a station from its NEU position
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  IO N/A  TEpoch structure
 *****************************************************************************/
void fillGroundStationOrientation (TEpoch *epoch) {
	getGroundStationOrientation (epoch->receiver.aproxPositionNEU, epoch->receiver.orientation);
}

/*****************************************************************************
 * Name        : getAzimuthElevation
 * Description : Get the azimuth/elevation in relation to WGS84 between a 
 *               pair satellite-receiver
 * Parameters  :
 * Name                           |Da|Unit|Description
 * double  orientation[3][3]       I  N/A  Orientation of the receiver
 * double  *receiverPosition       I  N/A  Position of the receiver
 * double  *satellitePosition      I  N/A  Position of the satellite
 * double  *azimuth                O  rad  Azimuth
 * double  *elevation              O  rad  Elevation
 *****************************************************************************/
void getAzimuthElevation(double orientation[3][3], double *receiverPosition, double *satellitePosition, double *azimuth, double *elevation) {
	int			i;
	double		relativeLoS[3];
	double		distance;
	
	distance = 0;
	for (i=0;i<3;i++) {
		distance += (satellitePosition[i]-receiverPosition[i])*(satellitePosition[i]-receiverPosition[i]);
	}
	
	distance = sqrt(distance);
	
	for (i=0;i<3;i++) {
		relativeLoS[i] = ((satellitePosition[0]-receiverPosition[0])*orientation[i][0] + 	
						  (satellitePosition[1]-receiverPosition[1])*orientation[i][1] + 	
						  (satellitePosition[2]-receiverPosition[2])*orientation[i][2]) / distance;
	}
	
	*azimuth = atan2(relativeLoS[1],relativeLoS[0]);
	
	//if (*azimuth!=0) {
//		*elevation = atan2(relativeLoS[1]/sin(*azimuth),relativeLoS[2]);
//	} else {
//		*elevation = asin(relativeLoS[0]);
//	}
	
//	*elevation = Pi/2-*elevation;
          *elevation = asin(relativeLoS[2]);
}

/*****************************************************************************
 * Name        : fillAzimuthElevation
 * Description : Fill the azimuth/elevation in relation to WGS84 between a 
 *               pair satellite-receiver
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  IO N/A  TEpoch structure
 * int  satIndex                   I  N/A  Index of the satellite
 *****************************************************************************/
void fillAzimuthElevation(TEpoch *epoch, int satIndex) {
	getAzimuthElevation(epoch->receiver.orientation, epoch->receiver.aproxPosition, epoch->sat[satIndex].position, &epoch->sat[satIndex].azimuth, &epoch->sat[satIndex].elevation);
}

/*****************************************************************************
 * Name        : rotate
 * Description : Rotates a 3D vector an specified angle in an axis
 * Parameters  :
 * Name                           |Da|Unit|Description
 * double  *v                      IO N/A  Vector to rotate. Result overwrittes
 *                                         this vector
 * double  angle                   I  rad  Angle to rotate
 * int  axis                       I  N/A  Rotation axis (1->x  2->y  3->z)
 *****************************************************************************/
void rotate (double *v, double angle, int axis) {
	double		vRes[3];
	
	if (axis==1) {
		vRes[0] =  v[0];
		vRes[1] =  v[1]*cos(angle) + v[2]*sin(angle);
		vRes[2] = -v[1]*sin(angle) + v[2]*cos(angle);
	} else if (axis==2) {
		vRes[0] =  v[0]*cos(angle) - v[2]*sin(angle);
		vRes[1] =  v[1];
		vRes[2] =  v[0]*sin(angle) + v[2]*cos(angle);
	} else if (axis==3) {
		vRes[0] =  v[0]*cos(angle) + v[1]*sin(angle);
		vRes[1] = -v[0]*sin(angle) + v[1]*cos(angle);
		vRes[2] =  v[2];
	}
	
	memcpy(v,vRes,sizeof(double)*3);
}

/*****************************************************************************
 * Name        : fillTransmissionTimeSat
 * Description : Obtains the transmission time for a specific satellite
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  IO N/A  TEpoch structure
 * TGNSSproducts  *products        I  N/A  TGNSSproducts structure
 * int  ind                        I  N/A  Index position of the satellite in 
 *                                         the epoch struct
 * TOptions  *options              I  N/A  TOptions structure
 * Returned value (int)            O  N/A  Status of the function
 *                                         1 => Properly done
 *                                         0 => Not found
 *****************************************************************************/
int fillTransmissionTimeSat(TEpoch *epoch, TGNSSproducts *products, int ind, TOptions *options) {
	TTime	t;
	double	satCorrection;
	int		res;
	double	meas;
	
	t.SoD = 0;
	t.MJDN = 0;
	
	if (options->satelliteClockCorrection) {
		res = getSatellitePVT(products,&epoch->t,0.0,epoch->sat[ind].GNSS,epoch->sat[ind].PRN,NULL,NULL,NULL,&satCorrection,options);
	} else {
		res = 1;
		satCorrection = 0;
	}
	if (res!=0) {
		meas = getMeasurementValue(epoch,epoch->sat[ind].GNSS,epoch->sat[ind].PRN,C1C);
		// If C1P (P1) is not available, try to get C1C (C1)
		if (meas==-1) meas = getMeasurementValue(epoch,epoch->sat[ind].GNSS,epoch->sat[ind].PRN,C1P);
		if (meas!=-1) {
			t = tdadd(&epoch->t, (-meas-satCorrection)/c);
			memcpy(&epoch->sat[ind].transTime,&t,sizeof(TTime));
		}
	} else {
		epoch->sat[ind].hasOrbitsAndClocks = 0;
	}
	if (res==0 || meas==-1) return 0;
	else return 1;
}

/*****************************************************************************
 * Name        : fillTransmissionTime
 * Description : Obtains the transmission time for all satellites
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  IO N/A  TEpoch structure
 * TGNSSproducts  *products        I  N/A  TGNSSproducts structure
 * TOptions  *options              I  N/A  TOptions structure
 *****************************************************************************/
void fillTransmissionTime (TEpoch *epoch, TGNSSproducts* products, TOptions *options) {
	int		i;
	
	for (i=0;i<epoch->numSatellites;i++) {
		fillTransmissionTimeSat(epoch,products,i,options);
	}
}

/*****************************************************************************
 * Name        : computeFlightTime
 * Description : Computes the signal flight time from the GNSS satellite to 
 *               the receiver
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  I  N/A  TEpoch structure
 * TGNSSproducts  *products        I  N/A  TGNSSproducts structure
 * int  satIndex                   I  N/A  Index of the satellite
 * TOptions  *options              I  N/A  TOptions structure
 * Returned value (double)         O  s    Signal flight time
 *****************************************************************************/
double computeFlightTime (TEpoch *epoch, TGNSSproducts *products, int satIndex, TOptions *options) {
	double	position[3];
	int		res;
	double	dist;
	
	res = getSatellitePVT(products,&epoch->sat[satIndex].transTime,0.0,epoch->sat[satIndex].GNSS,epoch->sat[satIndex].PRN,position,NULL,NULL,NULL,options);
	if (res==0) return 0;
	
	dist = sqrt((position[0]-epoch->receiver.aproxPosition[0])*(position[0]-epoch->receiver.aproxPosition[0]) +
				(position[1]-epoch->receiver.aproxPosition[1])*(position[1]-epoch->receiver.aproxPosition[1]) +
				(position[2]-epoch->receiver.aproxPosition[2])*(position[2]-epoch->receiver.aproxPosition[2]));
	return dist/c;
}

/*****************************************************************************
 * Name        : correctEarthRotation
 * Description : Corrects the Earth rotation of a satellite position.
 * Parameters  :
 * Name                           |Da|Unit|Description
 * double  *x                      IO N/A  3 dimension vector (overwritten)
 * double  flightTime              I  N/A  Flight time of the signal (difference 
 *                                         between the reception time and transmission 
 *                                         time).
 *****************************************************************************/
void correctEarthRotation(double *x, double flightTime) {
	double angle;
	
	angle = EARTH_ROTATION * flightTime;
	rotate(x,angle,3);
}

/*****************************************************************************
 * Name        : nsteffensen
 * Description : Method to accelerate the Newton-Rapson convergence for
 *               equations like:  p=g(p)    (==> E=M+e*sin(E))
 *               It is required:  g'(p)<>1  (==> p simple root)
 * Parameters  :
 * Name                           |Da|Unit|Description
 * double  Mk                      I  N/A  Mean Anomaly
 * double  e                       I  N/A  Excentricity
 * Returned value (double)         O  N/A  Excentric anomaly
 *****************************************************************************/
double nsteffensen(double Mk, double e) {
	const double	tol=1e-14;
	
	double			p,p0,p1,p2;
	double			dd;
	
	Mk = atan2(sin(Mk),cos(Mk));
	p = Mk;
	
	while(1) {
		p0 = p;
		p1 = Mk + e*sin(p0);
		p2 = Mk + e*sin(p1);
		
		dd = fabs(p2-2*p1+p0);
		if (dd<tol) break;
		p = p0-(p1-p0)*(p1-p0)/(p2-2*p1+p0);
		if (fabs(p-p0)<=tol) break;
	}
	
	return p;

}

/*****************************************************************************
 * Name        : satellitePhaseCenterCorrection3D
 * Description : Get the satellite phase center correction 
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TTime  *t                       I  N/A  Reference time
 * GNSSystem  GNSS                 I  N/A  GNSS system of the satellite
 * int  PRN                        I  N/A  PRN identifier of the satellite
 * MeasurementType meas            I  N/A  Measurement type
 * double  orientation[3][3]       I  N/A  Satellite orientation
 * TConstellation  *constellation  I  N/A  TConstellation structure where 
 *                                         satellite data is stored
 * double  *dr                     O  m    3D phase center correction
 * Returned value (int)            O  N/A  Status of the function
 *                                         1 => Phase center correction found
 *                                         0 => Satellite not found on TConstellation
 *****************************************************************************/
int satellitePhaseCenterCorrection3D (TTime *t, enum GNSSystem GNSS, int PRN, enum MeasurementType meas, double orientation[3][3], TConstellation *constellation, double *dr) {
	TConstellationElement	*el;
	int						ind;
	int						i;
	
	el = getConstellationElement(GNSS,PRN,t,constellation);
	ind = getFrequencyInt(meas);
	
	if (el!=NULL) {
		for (i=0;i<3;i++) {
			dr[i] = orientation[2][i]*el->PCO[ind][2] + orientation[1][i]*el->PCO[ind][1] + orientation[0][i]*el->PCO[ind][0];
		}
		return 1;
	}
	dr[0] = dr[1] = dr[2] = 0;
	return 0;
}

/*****************************************************************************
 * Name        : satellitePhaseCenterCorrection
 * Description : Get the satellite phase center correction 
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TTime  *t                       I  N/A  Reference time
 * GNSSystem  GNSS                 I  N/A  GNSS system of the satellite
 * int  PRN                        I  N/A  PRN identifier of the satellite
 * MeasurementType meas            I  N/A  Measurement type
 * double  orientation[3][3]       I  N/A  Satellite orientation
 * double  *LoS                    I  N/A  Line of Sight satellite-receiver
 * TConstellation  *constellation  I  N/A  TConstellation structure where 
 *                                         satellite data is stored
 * double  *pc                     O  m    Phase center correction
 * Returned value (int)            O  N/A  Status of the function
 *                                         1 => Phase center corrected
 *                                         0 => Satellite not found on TConstellation
 *****************************************************************************/
int satellitePhaseCenterCorrection (TTime *t, enum GNSSystem GNSS, int PRN, enum MeasurementType meas, double orientation[3][3], double *LoS, TConstellation *constellation, double *pc) {
	double					dr[3];
	
	if (satellitePhaseCenterCorrection3D(t,GNSS,PRN,meas,orientation,constellation,dr)==1) {
		*pc = scalarProd(LoS,dr);
		return 1;

	}
	*pc = 0;
	return 0;
}

/*****************************************************************************
 * Name        : getClockBRDC
 * Description : Get the clock correction of the satellite from specified 
 *               ephemeris block
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TBRDCblock  *block              I  N/A  TBRDCblock structure
 * TTime  *t                       I  N/A  Reference time
 * Returned value (double)         O  s    Clock correction
 *                                         999999.999999 => Clock not available
 *****************************************************************************/
double getClockBRDC(TBRDCblock *block,TTime *t) {
	double diff;
	
	diff = tdiff(t,&block->Ttoc);
	return (block->clockbias + block->clockdrift*diff + block->clockdriftrate*diff*diff)*c;
}

/*****************************************************************************
 * Name        : satellitePhaseCenterVarCorrection
 * Description : Get the zenital satellite phase center correction 
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TTime  *t                       I  N/A  Reference time
 * GNSSystem  GNSS                 I  N/A  GNSS system of the satellite
 * int  PRN                        I  N/A  PRN identifier of the satellite
 * MeasurementType meas            I  N/A  Measurement type
 * double  satearthcenterdistance  I  N/A  Distance of satellite to Earth centre
 * double  elevation               I  N/A  Elevation of the satellite in relation to the station (radians)
 * TConstellation  *constellation  I  N/A  TConstellation structure where 
 *                                         satellite data is stored
 * double  *zc                     O  m    Zenital phase center correction
 * Returned value (int)            O  N/A  Status of the function
 *                                         1 => Zenital phase center corrected
 *                                         0 => Satellite not found on TConstellation
 *****************************************************************************/
int satellitePhaseCenterVarCorrection (TTime *t, enum GNSSystem GNSS, int PRN, enum MeasurementType meas, double satearthcenterdistance, double elevation, TConstellation *constellation, double *pvc) {

        TConstellationElement   *el;
        int                     ind;
        int                     pos;
        double                  diff;
        double                  nadir;

        el = getConstellationElement(GNSS,PRN,t,constellation);
        ind = getFrequencyInt(meas);


        if (el!=NULL) {

                //Only Non azimuth corrections implemented!!

                //Calculate nadir angle
                nadir=asin((EARTH_RADIUS/satearthcenterdistance)*cos(elevation));

                //Check if angle is above the maximum zenital angle in the ANTEX file (some antennas the data goes from 0 to 80º)RTH_RADIUS
                if(nadir*r2d>el->zen2) {
                        *pvc=0;
                        return 0;
                }
                ind = getFrequencyInt(meas);
                pos=(int)((nadir*r2d-el->zen1)/el->dzen);

                //Check for the cases that are in the extreme
                if(pos==el->nzen-1) {
                        *pvc=el->noazi[GNSS][ind][pos];
                        return 1;
                }

                //Interpolate values
                diff=((nadir*r2d-el->zen1)/el->dzen)-pos;

                *pvc=el->noazi[GNSS][ind][pos]*(1-diff)+el->noazi[GNSS][ind][pos+1]*diff;

                return 1;

        }

        *pvc=0;
        return 0;
}


/*****************************************************************************
 * Name        : getPositionBRDC
 * Description : Get the position of the satellite from specified ephemeris block
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TBRDCblock  *block              I  N/A  TBRDCblock structure
 * TTime  *t                       I  N/A  Reference time
 * double  *position               O  m    3D vector with the satellite position
 *****************************************************************************/
void getPositionBRDC(TBRDCblock *block,TTime *t, double *position) {
	const double	mu = 3.986005e+14;  // WGS-84 Earth Univi. Grav. parameter (m3/s2)
	const double	om_e = 7.2921151467e-5;  // WGS-84 Earth rotation rate (rad/s)
	
	double 			diff;
	double			Mk;			// Mean anomaly
	double			Ek;			// Excentric anomaly
	double			fk;			// True anomaly
	double			uk;			// Argument of latitude
	double			rk;			// Radial distance
	double			ik;			// Inclination
	double			xp,yp;
	double			OMk;		// Longitude of ascending node
	
	diff = tdiff(t,&block->Ttoe);
	// True anomaly fk:
	Mk = block->M0 + (sqrt(mu)/(block->sqrta*block->sqrta*block->sqrta)+block->deltan)*diff;
	Ek = nsteffensen(Mk,block->e);
	if (Ek<0) Ek+=2*Pi;
	if (Ek>2*Pi) Ek-=2*Pi;
	fk = atan2(sqrt(1-block->e*block->e)*sin(Ek),cos(Ek)-block->e);
	if (fk<0) fk+=2*Pi;
	if (fk>2*Pi) fk-=2*Pi;
	
	// Arg. of Latitude uk,radius rk, inclination ik:
	uk = block->omega + fk + block->cuc*cos(2*(block->omega+fk)) + block->cus*sin(2*(block->omega+fk));
	rk = (block->sqrta*block->sqrta)*(1.0-block->e*cos(Ek)) + block->crc*cos(2.0*(block->omega+fk)) + block->crs*sin(2.0*(block->omega+fk));
	ik = block->i0 + block->IDOT*diff + block->cic*cos(2*(block->omega+fk)) + block->cis*sin(2*(block->omega+fk));
	
	// Positions in orbital plane
	xp = rk * cos(uk);
	yp = rk * sin(uk);
	
	// Longitude of ascending node
	OMk = block->OMEGA + (block->OMEGADOT-om_e)*diff - om_e*block->toe;
	
	// CT-System coordinates
	position[0] = xp*cos(OMk) - yp*cos(ik)*sin(OMk);
	position[1] = xp*sin(OMk) + yp*cos(ik)*cos(OMk);
	position[2] = yp*sin(ik);
}

/*****************************************************************************
 * Name        : getVelocityBRDC
 * Description : Get the velocity of the satellite from specified ephemeris block
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TBRDCblock  *block              I  N/A  TBRDCblock structure
 * TTime  *t                       I  N/A  Reference time
 * double  *position               I  N/A  Satellite position (3D vector) at time t.
 *                                         It is used to avoid for optimisation
 *                                         purposes, to avoid recomputing it when
 *                                         obtaining satellite PVT. If this parameter
 *                                         us NULL, the position at t is computed.
 * enum SatelliteVelocity  sv      I  N/A  Chooses if svINERTIAL or svITRF velocity
 *                                         is computed 
 * double  *velocity               O  m/s  3D vector with the satellite velocity
 *****************************************************************************/
void getVelocityBRDC(TBRDCblock *block,TTime *t, double *position, double *velocity, enum SatelliteVelocity satVel) {
	const double	diffTime = 0.001;
	double			pos[3],posPrev[3];
	TTime			tPrev;
	int				i;
	
	// Position at t
	if (position==NULL) {
		getPositionBRDC(block,t,pos);
	} else {
		memcpy(pos,position,sizeof(double)*3);
	}
	
	// Position at t-diffTime
	memcpy(&tPrev,t,sizeof(TTime));
	tPrev.SoD -= diffTime;
	getPositionBRDC(block,&tPrev,posPrev);
	
	// Earth rotation in diffTime
	if (satVel==svINERTIAL) {
		correctEarthRotation(posPrev,diffTime);
	}
	 
	// Estimate velocity
	for (i=0;i<3;i++) velocity[i] = (pos[i]-posPrev[i])/diffTime;
}

/*****************************************************************************
 * Name        : getClockSP3
 * Description : Obtain the satellite clock correction for a specified 
 *               satellite in a given time from a specified SP3 structure
 *               By default does not use any kind of interpolation due to the
 *               stochastic nature of the clocks, but TOptions structure has a 
 *               field (clockInterpolationDegree) to allow doing an interpolation
 *               with a specified degree.
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TGNSSproducts  *products        I  N/A  TGNSSproducts structure
 * TTime  *t                       I  N/A  Reference time
 * GNSSystem  GNSS                 I  N/A  GNSS system of the satellite
 * int  PRN                        I  N/A  PRN identifier of the satellite
 * TOptions  *options              I  N/A  TOptions structure to configure behaviour
 * Returned value (double)         O  s    Clock for time t and satellite PRN
 *****************************************************************************/
double getClockSP3 (TGNSSproducts *products,TTime *t, enum GNSSystem GNSS, int PRN, TOptions *options) {
	TSP3clocks	*clocks;
	int			i,j;
	int			ind;
	double		tDistance;
	int			pointsBelow = 0;
	int			pointsAbove;
	int			indI;
	int			halfDeg;
	int			lowerMargin;
	double		y[MAX_INTERPOLATION_DEGREE+1];
	TTime		tPrev[MAX_INTERPOLATION_DEGREE+1];
	
	clocks = &products->SP3->clocks;
	
	ind = clocks->index[GNSS][PRN];
	if (ind==-1) return INVALID_CLOCK;
	if (options->clockInterpolationDegree==0) { // No clock interpolation
		for (i=0;i<clocks->numRecords;i++) {
			tDistance = tdiff(t,&clocks->block[ind][i].t);
			if (fabs(tDistance)<MAX_CLOCK_DISTANCE) {
				return clocks->block[ind][i].clock + clocks->block[ind][i].clockrate*tDistance + clocks->block[ind][i].clockdriftrate*tDistance*tDistance;
			}
		}
	} else { // Clock interpolation
		for (i=0;i<clocks->numRecords;i++) {
			tDistance = tdiff(t,&clocks->block[ind][i].t);
			if (tDistance>0) pointsBelow++;
			if (tDistance<=0) break;
		}
		
		indI = i;
		if (indI == clocks->numRecords) {
			if (tdiff(t,&clocks->block[ind][indI-1].t)<=MAX_CLOCK_DISTANCE) {
				pointsBelow--;
				indI--;
			} else {
				// Time too far in the future to extrapolate
				return INVALID_CLOCK;
			}
		}
	
		if (indI == 0) {
			if (tdiff(t,&clocks->block[ind][0].t)<-MAX_CLOCK_DISTANCE) {
				// Time too earlier to extrapolate
				return INVALID_CLOCK;
			}
		}
		
		halfDeg = (options->clockInterpolationDegree+1)/2;
		
		pointsAbove = clocks->numRecords - pointsBelow;	

		if (clocks->numRecords<(options->clockInterpolationDegree+1)) {
			return INVALID_CLOCK; 				
		}
			
		if (pointsBelow>=halfDeg) lowerMargin = halfDeg;
		else lowerMargin = pointsBelow;
		if (pointsAbove<(options->clockInterpolationDegree-halfDeg+1)) lowerMargin = options->clockInterpolationDegree - pointsAbove + 1;
		for (j=0,i=indI-lowerMargin;j<=options->clockInterpolationDegree;j++,i++) {
			memcpy(&tPrev[j],&clocks->block[ind][i].t,sizeof(TTime));
		}
		
		for (j=0,i=indI-lowerMargin;j<=options->clockInterpolationDegree;j++,i++) {
			if (clocks->block[ind][i].clock == INVALID_CLOCK) return INVALID_CLOCK;
			y[j] = clocks->block[ind][i].clock;
		}
			
		return lagrangeInterpolation(options->clockInterpolationDegree,*t,tPrev,y);
	}
	
	return INVALID_CLOCK;  // This is the invalid data flag of SP3 files
}

/*****************************************************************************
 * Name        : getPositionSP3
 * Description : Obtain the satellite position estimation for a specified 
 *               satellite in a given time from a specified SP3 structure.
 *               By default uses a 9 degree interpolator (5 points before the epoch 
 *               and 5 points after). The TOptions structure has a field
 *               (orbitInterpolationDegree) to allow changing the degree.
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TGNSSproducts  *products        I  N/A  TGNSSproducts structure
 * TTime  *t                       I  N/A  Reference time
 * GNSSystem  GNSS                 I  N/A  GNSS system of the satellite
 * int  PRN                        I  N/A  PRN identifier of the satellite
 * double  *position               O  m    3D vector with the satellite position
 * TOptions  *options              I  N/A  TOptions structure to configure behaviour
 * Returned value (int)            O  N/A  Status of the function
 *                                         1 => Position computed
 *                                         0 => Error in the position computation
 *****************************************************************************/
int getPositionSP3(TGNSSproducts *products,TTime *t,enum GNSSystem GNSS,int PRN, double *position, TOptions *options) {
	const double	maxextrapolationtime = 1;
	TSP3orbits		*orbits;
	int				ind;
	int				indI;
	int				i,j,k;
	double			tDistance;
	int				pointsBelow = 0;
	int				pointsAbove;
	int				lowerMargin;
	double			y[MAX_INTERPOLATION_DEGREE+1];
	TTime			tPrev[MAX_INTERPOLATION_DEGREE+1];
	int				halfDeg;
	
	// halfDeg calculation
	halfDeg = (options->orbitInterpolationDegree+1)/2;
	
	orbits = &products->SP3->orbits;
	
	ind = orbits->index[GNSS][PRN];

	if (ind==-1) return 0;
	
	for (i=0;i<orbits->numRecords;i++) {
		tDistance = tdiff(t,&orbits->block[ind][i].t);
		if (tDistance>0) pointsBelow++;
		if (tDistance<=0) break;
	}

	indI = i;
	if (indI == orbits->numRecords) {
		if (tdiff(t,&orbits->block[ind][indI-1].t)<=maxextrapolationtime) {
			pointsBelow--;
			indI--;
		} else {
			// Time too far in the future to extrapolate
			return 0;
		}
	}
	
	if (indI == 0) {
		if (tdiff(t,&orbits->block[ind][0].t)<-maxextrapolationtime) {
			// Time too earlier to extrapolate
			return 0;
		}
	}
	
	pointsAbove = orbits->numRecords - pointsBelow;	

	if (orbits->numRecords<(options->orbitInterpolationDegree+1)) {
		position[0] = position[1] = position[2] = 0;
		return 0; 				
	}
		
	if (pointsBelow>=halfDeg) lowerMargin = halfDeg;
	else lowerMargin = pointsBelow;
	if (pointsAbove<(options->orbitInterpolationDegree-halfDeg+1)) lowerMargin = options->orbitInterpolationDegree - pointsAbove + 1;
	for (j=0,i=indI-lowerMargin;j<=options->orbitInterpolationDegree;j++,i++) {
		memcpy(&tPrev[j],&orbits->block[ind][i].t,sizeof(TTime));
	}
	
	for (k=0;k<3;k++) {
		for (j=0,i=indI-lowerMargin;j<=options->orbitInterpolationDegree;j++,i++) {
			y[j] = orbits->block[ind][i].x[k];
			if (orbits->block[ind][i].x[0]==0 && orbits->block[ind][i].x[1]==0 && orbits->block[ind][i].x[2]==0) return 0;
		}
		
		position[k] = lagrangeInterpolation(options->orbitInterpolationDegree,*t,tPrev,y);
	}
	
	return 1;
}

/*****************************************************************************
 * Name        : getVelocitySP3
 * Description : Obtain the satellite velocity estimation for a specified 
 *               satellite in a given time from a specified SP3 structure.
 *               By default uses a 9 degree interpolator (5 points before the epoch 
 *               and 5 points after). The TOptions structure has a field
 *               (orbitInterpolationDegree) to allow changing the degree.
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TGNSSproducts  *products        I  N/A  TGNSSproducts structure
 * TTime  *t                       I  N/A  Reference time
 * GNSSystem  GNSS                 I  N/A  GNSS system of the satellite
 * int  PRN                        I  N/A  PRN identifier of the satellite
 * double  *position               O  N/A  Satellite position (3D vector) at time t.
 *                                         It is used to avoid for optimisation
 *                                         purposes, to avoid recomputing it when
 *                                         obtaining satellite PVT. If this parameter
 *                                         us NULL, the position at t is computed.
 * double  *velocity               O  m/s  3D vector with the satellite velocity
 * enum SatelliteVelocity  satVel  I  N/A  Chooses if svINERTIAL or svITRF velocity
 *                                         is computed 
 * TOptions  *options              I  N/A  TOptions structure to configure behaviour
 * Returned value (int)            O  N/A  Status of the function
 *                                         1 => Velocity computed
 *                                         0 => Error in the velocity computation
 *****************************************************************************/
int getVelocitySP3(TGNSSproducts *products,TTime *t,enum GNSSystem GNSS,int PRN, double *position, double *velocity, enum SatelliteVelocity satVel, TOptions *options) {
	const double	diffTime = 0.001;
	double			pos[3],posPrev[3];
	TTime			tPrev;
	int				i;
	int				res;
		
	// Position at t
	if (position==NULL) {
		res = getPositionSP3(products,t,GNSS,PRN,pos,options);
		if (res==0) return 0;
	} else {
		memcpy(pos,position,sizeof(double)*3);
	}
	
	// Position at t-diffTime
	memcpy(&tPrev,t,sizeof(TTime));
	tPrev.SoD -= diffTime;
	res = getPositionSP3(products,&tPrev,GNSS,PRN,posPrev,options);
	if (res==0) return 0;
	
	// Earth rotation in diffTime
	if (satVel == svINERTIAL) {
		correctEarthRotation(posPrev,diffTime);
	}
	
	// Estimate velocity
	for (i=0;i<3;i++) velocity[i] = (pos[i]-posPrev[i])/diffTime;
	
	return 1;
}

/*****************************************************************************
 * Name        : getSatellitePVTBRDCraw
 * Description : Obtains PVT information for a specific satellite from broadcast 
 *               products
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TGNSSproducts  *products        I  N/A  TGNSSproducts structure
 * TTime  *t                       I  N/A  Reference time
 * GNSSystem  GNSS                 I  N/A  GNSS system of the satellite
 * int  PRN                        I  N/A  PRN identifier of the satellite
 * double  *position               O  m    3D vector with the satellite position
 *                                         If the pointer was NULL this parameter
 *                                         is not computed
 * double  *velocity               O  m/s  3D vector with the satellite velocity
 *                                         If the pointer was NULL this parameter
 *                                         is not computed
 * double  *ITRFvel                O  m/s  3D vector with the ITRF satellite velocity
 *                                         If the pointer was NULL this parameter
 *                                         is not computed
 * double  *clock                  O  s    Satellite clock correction
 *                                         If the pointer was NULL this parameter
 *                                         is not computed
 * TOptions  *options              I  N/A  TOptions structure to configure behaviour
 * Returned value (int)            O  N/A  Status of the function
 *                                         0 => Information not available
 *                                         1 => OK
 *****************************************************************************/
int getSatellitePVTBRDCraw(TGNSSproducts *products, TTime *t, enum GNSSystem GNSS, int PRN, double *position, double *velocity, double *ITRFvel, double *clock, TOptions *options) {
	TBRDCblock		*block;
	
	if (products->type != BRDC) return 0;
	
	block = selectBRDCblock(products->BRDC,t,GNSS,PRN,options);
	
	if (block==NULL) return 0;

	// Obtain satellite clock correction
	if (clock!=NULL) {
		*clock = getClockBRDC(block,t);
	}
	
	// Obtain satellite position
	if (position!=NULL) {
		getPositionBRDC(block,t,position);
	}
	
	// Obtain satellite velocity
	if (velocity!=NULL) {
		getVelocityBRDC(block,t,position,velocity,svINERTIAL);
	}
	if (ITRFvel!=NULL) {
		getVelocityBRDC(block,t,position,ITRFvel,svITRF);
	}
	
	return 1;
}

/*****************************************************************************
 * Name        : getSatellitePVTSP3raw
 * Description : Obtains PVT information for a specific satellite from SP3 products
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TGNSSproducts  *products        I  N/A  TGNSSproducts structure
 * TTime  *t                       I  N/A  Reference time
 * GNSSystem  GNSS                 I  N/A  GNSS system of the satellite
 * int  PRN                        I  N/A  PRN identifier of the satellite
 * double  *position               O  m    3D vector with the satellite position
 *                                         If the pointer was NULL this parameter
 *                                         is not computed
 * double  *velocity               O  m/s  3D vector with the satellite velocity
 *                                         If the pointer was NULL this parameter
 *                                         is not computed
 * double  *ITRFvel                O  m/s  3D vector with the ITRF satellite velocity
 *                                         If the pointer was NULL this parameter
 *                                         is not computed
 * double  *clock                  O  s    Satellite clock correction
 *                                         If the pointer was NULL this parameter
 *                                         is not computed
 * TOptions  *options              I  N/A  TOptions structure to configure behaviour
 * Returned value (int)            O  N/A  Status of the function
 *                                         0 => Information not available
 *                                         1 => OK
 *****************************************************************************/
int getSatellitePVTSP3raw(TGNSSproducts *products, TTime *t, enum GNSSystem GNSS, int PRN, double *position, double *velocity, double *ITRFvel, double *clock, TOptions *options) {
	int res;
	
	if (products->type != SP3) return 0;
	
	// Obtain satellite clock correction
	if (clock!=NULL) {
		*clock = getClockSP3(products,t,GNSS,PRN,options);
		if (*clock==INVALID_CLOCK) return 0;
	}
	
	// Obtain satellite position
	if (position!=NULL) {
		res = getPositionSP3(products,t,GNSS,PRN,position,options);
		if (res==0) return 0;
	}
	
	// Obtain satellite velocity
	if (velocity!=NULL) {
		res = getVelocitySP3(products,t,GNSS,PRN,position,velocity,svINERTIAL,options);
		if (res==0) return 0;
	}
	if (ITRFvel!=NULL) {
		res = getVelocitySP3(products,t,GNSS,PRN,position,ITRFvel,svITRF,options);
		if (res==0) return 0;
	}
	
	return 1;
}

/*****************************************************************************
 * Name        : getSatellitePVT
 * Description : Obtains PVT information for a specific satellite
 *               Important note: It does NOT take into account the phase center 
 *                               correction
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TGNSSproducts  *products        I  N/A  TGNSSproducts structure
 * TTime  *t                       I  N/A  Reference time
 * double  flightTime              I  s    Time of flight, in order to properly 
 *                                         rotate the position
 * GNSSystem  GNSS                 I  N/A  GNSS system of the satellite
 * int  PRN                        I  N/A  PRN identifier of the satellite
 * double  *position               O  m    3D vector with the satellite position
 *                                         If the pointer was NULL this parameter
 *                                         is not computed
 * double  *velocity               O  m/s  3D vector with the inertial satellite velocity
 *                                         If the pointer was NULL this parameter
 *                                         is not computed
 * double  *ITRFvel                O  m/s  3D vector with the ITRF satellite velocity
 *                                         If the pointer was NULL this parameter
 *                                         is not computed
 * double  *clock                  O  s    Satellite clock correction
 *                                         If the pointer was NULL this parameter
 *                                         is not computed
 * TOptions  *options              I  N/A  TOptions structure to configure behaviour
 * Returned value (int)            O  N/A  Status of the function
 *                                         0 => Information not available
 *                                         1 => OK
 *****************************************************************************/
int getSatellitePVT(TGNSSproducts *products, TTime *t, double flightTime, enum GNSSystem GNSS, int PRN, double *position, double *velocity, double *ITRFvel, double *clock, TOptions *options) {
	int res;
	
	if (products->type == BRDC) {
		res = getSatellitePVTBRDCraw(products,t,GNSS,PRN,position,velocity,ITRFvel,clock,options);
	} else if (products->type == SP3) {
		res = getSatellitePVTSP3raw(products,t,GNSS,PRN,position,velocity,ITRFvel,clock,options);
	} else {
		res = 0;
	}
	if (res!=0 && options!=NULL) {
		if (options->earthRotation) {
			if (position!=NULL) { 
				correctEarthRotation(position,flightTime);
			}
			if (velocity!=NULL) {
				correctEarthRotation(velocity,flightTime);
			}
			if (ITRFvel!=NULL) {
				correctEarthRotation(ITRFvel,flightTime);
			}
		}
	}
	
	return res;
}

/*****************************************************************************
 * Name        : fillSatellitePVT
 * Description : Obtains PVT information for a specific satellite and fills 
 *               it in TSatellite structure.
 *               It DOES take into account the phase center correction
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  IO N/A  TEpoch structure
 * TGNSSproducts  *products        I  N/A  TGNSSproducts structure
 * TTime  *t                       I  N/A  Reference time
 * double  flightTime              I  s    Time of flight, in order to properly 
 *                                         rotate the position
 * int  satIndex                   I  N/A  Index of the satellite inside TEpoch structure
 * int  position                   I  N/A  Flag to compute position (0-> Do not)
 * int  velocity                   I  N/A  Flag to compute velocity (0-> Do not)
 * int  clock                      I  N/A  Flag to compute clock (0-> Do not)
 * TConstellation  *constellation  I  N/A  TConstellation structure
 * TOptions  *options              I  N/A  TOptions structure to configure the 
 *                                         satellite PVT
 * Returned value (int)            O  N/A  Status of the function
 *                                         0 => Information not available
 *                                         1 => OK
 *****************************************************************************/
int fillSatellitePVT(TEpoch *epoch, TGNSSproducts *products, TTime *t, double flightTime, int satIndex, int position, int velocity, int clock, TConstellation *constellation, TOptions *options) {
	int		res;
	double	*posPt;
	double	*velPt;
	double  *ITRFvelPt;
	double	*clkPt;
	double	posVal[3];
	double	velVal[3];
	double  ITRFvelVal[3];
	double	clkVal;
	
	if (position) posPt = posVal;
	else posPt = NULL;
	
	if (velocity) {
		velPt = velVal;
		ITRFvelPt = ITRFvelVal;
	} else {
		velPt = NULL;
		ITRFvelPt = NULL;
	}
	
	if (clock) clkPt = &clkVal;
	else clkPt = NULL;
	
	
	res = getSatellitePVT(products, t, flightTime, epoch->sat[satIndex].GNSS, epoch->sat[satIndex].PRN, posPt, velPt, ITRFvelPt, clkPt, options);
	
	if (res!=0) {
		
		
	
		// Store it in TEpoch structure
		if (position) {
			memcpy(epoch->sat[satIndex].position,posPt,sizeof(double)*3);
		}
		if (velocity) {
			memcpy(epoch->sat[satIndex].velocity,velPt,sizeof(double)*3);
		}
		if (velocity) {
			memcpy(epoch->sat[satIndex].ITRFvel,ITRFvelPt,sizeof(double)*3);
		}
		if (clock) {
			epoch->sat[satIndex].clockCorrection = -*clkPt;
		}
		
		// Satellite orientation
		fillSatelliteOrientation(epoch,satIndex);
		
		// Flag the satellite as "orbit and clock" available
		epoch->sat[satIndex].hasOrbitsAndClocks = 1;
	} else {
		// Flag the satellite as "orbit and clock" unavailable
		epoch->sat[satIndex].hasOrbitsAndClocks = 0;
	}
	return res;
}

/*****************************************************************************
 * Name        : getLineOfSight
 * Description : Obtains the Line of Sight vector from receiver to satellite.
 *               LoS is an unitary vector.
 * Parameters  :
 * Name                           |Da|Unit|Description
 * double  *satPos                 I  N/A  Transmitter position
 * double  *recPos                 I  N/A  Receiver position
 * double  *LoS                    O  N/A  Line of Sight
 *****************************************************************************/
void getLineOfSight(double *satPos, double *recPos, double *LoS) {
	int			i;
	
	for (i=0;i<3;i++)
		LoS[i] = satPos[i] - recPos[i];
	
	unitary(LoS);
}

/*****************************************************************************
 * Name        : fillLineOfSight
 * Description : Obtains the Line of Sight vector from receiver to satellite.
 *               LoS is an unitary vector.
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  IO N/A  TEpoch structure
 * int  satIndex                   I  N/A  Index of the satellite inside TEpoch structure
 *****************************************************************************/
void fillLineOfSight(TEpoch *epoch, int satIndex) {
	TSatellite	*sat;
	
	sat = &epoch->sat[satIndex];
	
	getLineOfSight(sat->position,epoch->receiver.aproxPosition,sat->LoS);
}

/*****************************************************************************
 * Name        : findMoon
 * Description : Obtains the position vector of the Moon in relation to Earth 
 *               (in ECEF). Source: 
 *               O. Montenbruck and G. Eberhard, Satellite Orbits, pp. 70-73, Springer, 2000
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TTime  *t                       I  N/A  Reference time
 * double  *moonPosition           O  Km   Moon position at t
 *****************************************************************************/
void findMoon (TTime *t, double* moonPosition) {
	double	T;
	double	L0;
	double 	l;
	double	lp;
	double	F;
	double	D;
	double	longitude;
	double	latitude;
	double	distance;
	double	obliquity = 23.43929111*d2r;
	double	fday;
	double	JDN;
	double	gstr;
	
	fday = t->SoD/86400;
	JDN = t->MJDN-15019.5 + fday;
	
	T 	= (t->MJDN - 51544.5  + t->SoD/86400)/36525.0;
	L0	= (218.31617 + 481267.88088*T - 1.3972*T)*d2r;
	l 	= (134.96292 + 477198.86753*T)*d2r;
	lp	= (357.52543 +  35999.04944*T)*d2r;
	F	= (93.27283  + 483202.01873*T)*d2r;
	D	= (297.85027 + 445267.11135*T)*d2r;
	
	longitude = L0 + (22640*sin(l) + 769*sin(2*l)
					- 4586*sin(l-2*D) + 2370*sin(2*D)
					- 668*sin(lp) - 412*sin(2*F)
					- 212*sin(2*l-2*D) - 206*sin(l+lp-2*D)
					+ 192*sin(l+2*D) - 165*sin(lp-2*D)
					+ 148*sin(l-lp) - 125*sin(D)
					- 110*sin(l+lp) - 55*sin(2*F-2*D)
				)/3600*d2r;
				
	latitude = (18520*sin(F + longitude - L0 + (412*sin(2*F) + 541*sin(lp))/3600*d2r)
				- 526*sin(F-2*D) + 44*sin(l+F-2*D)
				- 31*sin(-l+F-2*D) - 25*sin(-2*l+F)
				- 23*sin(lp+F-2*D) + 21*sin(-l+F)
				+ 11*sin(-lp+F-2*D)
				)/3600*d2r;
				
	distance = (385000 - 20905*cos(l) - 3699*cos(2*D-l)
				- 2956*cos(2*D) - 570*cos(2*l) + 246*cos(2*l-2*D)
				- 205*cos(lp-2*D) -171*cos(l+2*D)
				- 152*cos(l+lp-2*D)
				)*1;
	
	longitude = modulo(longitude,Pi*2);
	latitude = modulo(latitude,Pi*2);
	
	moonPosition[0] = distance * cos(longitude) * cos(latitude);
	moonPosition[1] = distance * sin(longitude) * cos(latitude);
	moonPosition[2] = distance * sin(latitude);
	
	rotate(moonPosition,-obliquity,1);
	
	// Rotate from inertial to non inertial system (ECI to ECEF)
	gstr = modulo(279.690983 + 0.9856473354*JDN + 360*fday + 180,360)*d2r;
	rotate(moonPosition,gstr,3);
}

/*****************************************************************************
 * Name        : findSun
 * Description : Obtains the position vector of the Sun in relation to Earth
 *               (in ECEF).
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TTime  *t                       I  N/A  Reference time
 * double  *sunPosition            O  Km   Sun position at t
 *****************************************************************************/
void findSun (TTime *t, double* sunPosition) {
	double	gstr;
	double	slong;
	double	sra;
	double	sdec;
	
	GSDtime_sun(t,&gstr,&slong,&sra,&sdec);
	
	sunPosition[0] = cos(sdec*d2r) * cos((sra)*d2r) * AU;
	sunPosition[1] = cos(sdec*d2r) * sin((sra)*d2r) * AU;
	sunPosition[2] = sin(sdec*d2r) * AU;
	
	// Rotate from inertial to non inertial system (ECI to ECEF)
	rotate(sunPosition,gstr*d2r,3);
}

/*****************************************************************************
 * Name        : GSDtime_sun
 * Description : Positions the sun
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TTime  *t                       I  N/A  Reference time
 * double*  gstr                   O  rad  GMST0 (to go from ECEF to inertial)
 * double*  slong			       O  rad  Sun longitude
 * double*  sra				       O  rad  Sun right Ascension
 * double*  sdec			       O  rad  Sun declination
 *****************************************************************************/
void GSDtime_sun (TTime *t, double *gstr, double *slong, double *sra, double *sdec) {
	static int		first = 1;
	static TTime	lastTimeComputed;
	static double	gstrRec,slongRec,sraRec,sdecRec;
	double			JDN;
	double			fday;
	double			vl;
	double			g;
	double			obliq;
	double			slp;
	double			sind;
	double			cosd;
	
	if (first) {
		first = 0;
		lastTimeComputed.MJDN = 0;
		lastTimeComputed.SoD = 0;
	}
	
	if (tdiff(t,&lastTimeComputed)==0) {
		*gstr = gstrRec;
		*slong = slongRec;
		*sra = sraRec;
		*sdec = sdecRec;
		return;
	}
	
	fday = t->SoD/86400;
	JDN = t->MJDN-15019.5 + fday;
	
	vl = modulo(279.696678 + 0.9856473354*JDN,360);
	*gstr = modulo(279.690983 + 0.9856473354*JDN + 360*fday + 180,360);
	g = modulo(358.475845 + 0.985600267*JDN,360)*d2r;
	
	*slong = vl + (1.91946-0.004789*JDN/36525)*sin(g) + 0.020094*sin(2*g);
	obliq = (23.45229-0.0130125*JDN/36525)*d2r;
	
	slp = (*slong-0.005686)*d2r;
	sind = sin(obliq)*sin(slp);
	cosd = sqrt(1-sind*sind);
	*sdec = atan2(sind,cosd)/d2r;
	
	*sra = 180 - atan2(sind/cosd/tan(obliq),-cos(slp)/cosd)/d2r;
	
	slongRec = *slong;
	sraRec = *sra;
	gstrRec = *gstr;
	sdecRec = *sdec;
	memcpy(&lastTimeComputed,t,sizeof(TTime));
}

/*****************************************************************************
 * Name        : receiverPhaseCenterCorrection
 * Description : Get the receiver phase center correction for a specific measurement 
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TAntenna *antenna               I  N/A  Antenna information
 * GNSSystem  GNSS                 I  N/A  GNSS system of the satellite
 * MeasurementType meas            I  N/A  Measurement type
 * double  orientation[3][3]       I  N/A  Receiver orientation
 * double  *LoS                    I  N/A  Line of Sight satellite-receiver
 * Returned value (double)         O  N/A  Receiver antenna phase center projection
 *                                         into the LoS
 *****************************************************************************/
double receiverPhaseCenterCorrection (TAntenna *antenna, enum GNSSystem GNSS, enum MeasurementType meas, double orientation[3][3], double *LoS) {
	int		ind;
	double	dr[3];
	int		i,j;
	
	ind = getFrequencyInt(meas);
	for (j=0;j<3;j++) {
		dr[j] = 0;
		for (i=0;i<3;i++) {
			dr[j] += orientation[i][j]*antenna->PCO[ind][i];
		}
	}
	
	return -scalarProd(dr,LoS);
}

/*****************************************************************************
 * Name        : receiverPhaseCenterVarCorrection
 * Description : Get the receiver phase center correction for a specific measurement 
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TAntenna *antenna               I  N/A  Antenna information
 * GNSSystem  GNSS                 I  N/A  GNSS system of the satellite
 * MeasurementType meas            I  N/A  Measurement type
 * double  elevation               I  N/A  Elevation of the satellite in relation to the station (radians)
 * Returned value (double)         O  N/A  Receiver antenna phase center variation correction 
 *****************************************************************************/
double receiverPhaseCenterVarCorrection(TAntenna *antenna, enum GNSSystem GNSS, enum MeasurementType meas, double elevation){

        int     pos;
        int     ind;
        double  diff;

        //Only non azimuth dependent corrections implemented!!!!!

        //Check if angle is above the maximum zenital angle in the ANTEX file (some antennas the data goes from 0 to 80º)
        if(90-elevation*r2d>antenna->zen2) return 0;

        ind = getFrequencyInt(meas);
        pos=(int)((90-elevation*r2d-antenna->zen1)/antenna->dzen);

        //Check for the cases that are in the extreme
        if(pos==antenna->nzen-1) return antenna->noazi[GNSS][ind][pos];

        //Interpolate values
        diff=((90-elevation*r2d-antenna->zen1)/antenna->dzen)-pos;

        return antenna->noazi[GNSS][ind][pos]*(1-diff)+antenna->noazi[GNSS][ind][pos+1]*diff;


}


/*****************************************************************************
 * Name        : receiverARPCorrection
 * Description : Get the receiver Antenna Reference Point 
 * Parameters  :
 * Name                           |Da|Unit|Description
 * double  *ARP                    I  N/A  3D vector with dNEU data of the ARP
 * double  orientation[3][3]       I  N/A  Receiver orientation
 * double  *LoS                    I  N/A  Line of Sight satellite-receiver
 * Returned value (double)         O  N/A  Receiver ARP projection into the LoS
 *****************************************************************************/
double receiverARPCorrection (double *ARP, double orientation[3][3] , double *LoS) {
	int		i,j;
	double	dr[3];
	
	for (j=0;j<3;j++) {
		dr[j] = 0;
		for (i=0;i<3;i++) {
			dr[j] += orientation[i][j]*ARP[i];
		}
	}
	
	return -scalarProd(dr,LoS);
}

/*****************************************************************************
 * Name        : relativisticCorrection
 * Description : Returns the relativistic correction for the current satellite
 * Parameters  :
 * Name                           |Da|Unit|Description
 * double  *position               I  m    Position of the GNSS satellite
 * double  *velocity               I  m/s  Velocity of the GNSS satellite
 * Returned value (double)         O  m    Relativistic correction
 *****************************************************************************/
double relativisticCorrection(double *position, double *velocity) {
	return scalarProd(position,velocity)*2/c;
}

/*****************************************************************************
 * Name        : windUpCorrection
 * Description : Obtains the wind up correction for carrier-phase measurements
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  I  N/A  TEpoch structure
 * int  satIndex                   I  N/A  Index of the satellite to obtain the
 *                                         correction
 * Returned value (double)         O  rad  Wind up correction
 *****************************************************************************/
double windUpCorrection(TEpoch *epoch, int satIndex) {
	double	satelliteAzimuth;
	double	satelliteElevation;
	double	satelliteInc;
	double	receiverInc;
	
	getAzimuthElevation(epoch->sat[satIndex].orientation,epoch->sat[satIndex].position,epoch->receiver.aproxPosition,&satelliteAzimuth,&satelliteElevation);
	
	satelliteInc = satelliteAzimuth - epoch->sat[satIndex].windUpRadAccumSatellite;
	receiverInc = epoch->sat[satIndex].azimuth - epoch->sat[satIndex].windUpRadAccumReceiver;
	
	epoch->sat[satIndex].windUpRadAccumSatellite += atan2(sin(satelliteInc),cos(satelliteInc));
	epoch->sat[satIndex].windUpRadAccumReceiver += atan2(sin(receiverInc),cos(receiverInc));

	return epoch->sat[satIndex].windUpRadAccumSatellite - epoch->sat[satIndex].windUpRadAccumReceiver;
}

/*****************************************************************************
 * Name        : interpolateNiell
 * Description : Support function to interpolate values for the computation
 *               the Niell tropospheric correction
 * Parameters  :
 * Name                           |Da|Unit|Description
 * double  x                       I  deg  Station latitude
 * double  v[3][5]                 I  N/A  Matrix of vector to interpolate
 * int  i                          I  N/A  Selector of 'v' row
 * Returned value (double)         O  N/A  Interpolated value of v[i] in x
 *****************************************************************************/
double interpolateNiell (double x, double v[3][5], int i) {
  double x1,x2,y1,y2;
  int j;

  if (x<15) return v[i][0];
  if (x>75) return v[i][4];
  j = 1 + (int)(x/15-1);

  x1 = 15.*j;
  x2 = 15.*(j+1);
  y1 = v[i][j-1];
  y2 = v[i][j];

  return (x-x2)/(x1-x2)*y1 + (x-x1)/(x2-x1)*y2;
}

/*****************************************************************************
 * Name        : xmNiell
 * Description : Support function for the computation of the Niell tropospheric 
 *               mapping
 * Parameters  :
 * Name                           |Da|Unit|Description
 * double  elevation               I  rad  Elevation of the satellite in relation
 *                                         to the station
 * double  vector[3][5]            I  N/A  Mapping factors
 * Returned value (double)         O  N/A  Mapping
 *****************************************************************************/
double xmNiell (double elevation, double vector[3]) {
  double xn,xd,sinE;

  xn = 1. + vector[0]/(1. + vector[1]/(1. + vector[2]));
  sinE = sin(elevation);
  xd = sinE + vector[0]/(sinE+vector[1]/(sinE+vector[2]));

  return xn/xd;
}

/*****************************************************************************
 * Name        : tropNiell
 * Description : Obtains the Niell mappings
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TTime  *t                       I  N/A  Reference time
 * double  *positionNEU            I  rd/m 3D vector with latitude-longitude-height 
 *                                         of the station
 * double  elevation               I  rad  Elevation (from the station point of view) 
 *                                         of the pair station-satellite
 * double  *tropDryMap             O  N/A  Dry mapping factor (relates vertical delay 
 *                                         with slant delay)
 * double  *tropWetMap             O  N/A  Wet mapping factor (relates vertical delay
 *                                         with slant delay)
 *****************************************************************************/
void tropNiell(TTime *t,double *positionNEU,double elevation,double *tropDryMap,double *tropWetMap) {
	double	latitude;
	double	vector[3];
	int		i;
	double	DoY,d0;
	int 	year;
	
	// Coefficients of Hydrostatic Mapping function
	//   Average
	static double average[3][5] = {
          {1.2769934e-3,1.2683230e-3,1.2465397e-3,1.2196049e-3,1.2045996e-3},
          {2.9153695e-3,2.9152299e-3,2.9288445e-3,2.9022565e-3,2.9024912e-3},
          {62.610505e-3,62.837393e-3,63.721774e-3,63.824265e-3,64.258455e-3}};
	
	//   Amplitude
	static double amplitude[3][5] = {
          {0.e0,1.2709626e-5,2.6523662e-5,3.4000452e-5,4.1202191e-5},
          {0.e0,2.1414979e-5,3.0160779e-5,7.2562722e-5,11.723375e-5},
          {0.e0,9.0128400e-5,4.3497037e-5,84.795348e-5,170.37206e-5}};
          
	//   Height correction
	static double heightCorr[3] = {2.53e-5,5.49e-3,1.14e-3};

	// Coefficients of Wet Mapping Function
	static double awet[3][5] = {
          {5.8021897e-4,5.6794847e-4,5.8118019e-4,5.9727542e-4,6.1641693e-4},
          {1.4275268e-3,1.5138625e-3,1.4572752e-3,1.5007428e-3,1.7599082e-3},
          {4.3472961e-2,4.6729510e-2,4.3908931e-2,4.4626982e-2,5.4736038e-2}};
	
	latitude = fabs(positionNEU[0]/d2r);
	t2doy(t,&year,&DoY);

	if (positionNEU[0] >= 0) {
		d0=28;
	} else {
		d0=211;
	}
	
	//In Niell mapping, UT days are from 0 to 364.25. gLAB DoY go from 1 to 365.25
	DoY--;

	// Hydrostatic mapping...
	for (i=0;i<3;i++) {
		vector[i] = interpolateNiell(latitude,average,i) - interpolateNiell(latitude,amplitude,i)*cos(2*Pi*(DoY-d0)/365.25);
	}
	*tropDryMap = xmNiell(elevation,vector) + (1./sin(elevation) - xmNiell(elevation,heightCorr))*positionNEU[2]/1000.;
	
	// Wet mapping
	for (i=0;i<3;i++) {
		vector[i] = interpolateNiell(latitude,awet,i);
	}
	*tropWetMap = xmNiell (elevation,vector);
}

/*****************************************************************************
 * Name        : troposphericCorrection
 * Description : Computes the troposheric nominal values and the mapping used 
 *               to obtain the troposheric correction depending on the input
 *               options.
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TTime  *t                       I  N/A  Reference time
 * double  *positionNEU            I  rd/m 3D vector with latitude-longitude-height 
 *                                         of the station
 * double  elevation               I  rad  Elevation (from the station point of view) 
 *                                         of the pair station-satellite
 * TTROPOESA *TropoEsa		   I  N/A  TTROPOESA structure with ESA's data
 * double  *tropWetMap             O  N/A  Wet mapping factor (relates vertical delay
 *                                         with slant delay)
 * double  *ZTD                    O  m    Zenith Troposphreic Delay (without mapping)
 * Returned value (double)         O  m    Tropospheric slant correction
 *****************************************************************************/
double troposphericCorrection(TTime *t, double *positionNEU, double elevation, TTROPOESA *TropoEsa, double *tropWetMap, double *ZTD, TOptions *options) {
	double	tropDryMap;
	double	tropDryNominal,tropWetNominal;

	// Sanity check
	if (positionNEU[2]<-200) {
		*tropWetMap = 0;
		*ZTD = 0;
		return 0;
	}

	// Compute Nominal Tropospheric values
	if (options->tropNominal == SimpleNominal) {
		tropDryNominal = 1.013*2.27*exp((-0.116e-03)*positionNEU[2]);
		tropWetNominal = 0.10;
	} else if (options->tropNominal == UNB3Nominal) {
		troposphericCorrectionMOPS (t,positionNEU,elevation,&tropDryNominal,&tropWetNominal);		
	} else if (options->tropNominal == TropoESA) {
		//This is not ESA's correction, it is the same as SimpleNominal because ESA's code has been erased
		tropDryNominal = 1.013*2.27*exp((-0.116e-03)*positionNEU[2]);
		tropWetNominal = 0.10;
	}

	// Compute Tropospheric Mapping
	if (options->tropMapping == NiellMapping) {
		tropNiell(t,positionNEU,elevation,&tropDryMap,tropWetMap);
	} else if (options->tropMapping == SimpleMapping) {
		*tropWetMap = 1.001/sqrt(0.002001+sin(elevation)*sin(elevation));
		tropDryMap = (*tropWetMap);
	}

	// Obtain Zenith Tropospheric Delay
       *ZTD = tropDryNominal + tropWetNominal;

    
	// Return Tropospheric Slant Correction.
	return (tropDryNominal*tropDryMap + tropWetNominal*(*tropWetMap));
}


/*****************************************************************************
 * Name        : troposphericCorrectionMOPS
 * Description : Obtains, using the model described the RTCA "Minimum Operational 
 *               Performance Standards" (MOPS), version C (RTCA/DO-229C), in the
 *               Appendix A.4.2.4.
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TTime  *t                       I  N/A  Reference time
 * double  *positionNEU            I  rd/m 3D vector with latitude-longitude-height 
 *                                         of the station
 * double  elevation               I  rad  Elevation (from the station point of view) 
 *                                         of the pair station-satellite
 * double  *tropWetMOPS_Nominal    O  m    Wet Nominal Value (part of vertical delay)                                        
 * double  *tropDryMOPS_Nominal    O  m    Dry Nominal Value (part of vertical delay) 
 *****************************************************************************/
void troposphericCorrectionMOPS (TTime *t, double *positionNEU, double elevation,double *tropWetMOPS_Nominal,double *tropDryMOPS_Nominal) {

        // Some MOPS specific constants
        double MOPSgm = 9.784;
        double MOPSg  = 9.80665;
        double MOPSk1 = 77.604;
        double MOPSk2 = 382000.0;
        double MOPSRd = 287.054;


        double MOPSzhydCT = ( 1E-6 * MOPSk1 * MOPSRd / MOPSgm );
        double MOPSzwetCT = ( 1E-6 * MOPSk2 * MOPSRd );

	// MOPS Latitude Grid
	double MOPSlat[5] = {   15.0,   30.0,   45.0,   60.0,   75.0};
	
        // Meteorological Average Parameters : Po, To, Eo, Betao, Lambdao
	double MOPSaverages[5][5] = {
          {1013.25,1017.25,1015.75,1011.75,1013.00},
          { 299.65, 294.15, 283.15, 272.15, 263.65},
          {  26.31,  21.79,  11.66,   6.78,   4.11},
	  {6.30E-3,6.05E-3,5.58E-3,5.39E-3,4.53E-3},
	  {   2.77,   3.15,   2.57,   1.81,   1.55}};
  
        // Meteorological Seasonal Variation : deltaP, deltaT, deltaE, deltaBeta, deltaLambda
	double MOPSvariations[5][5] = {
	  {    0.0,  -3.75,  -2.25,  -1.75,  -0.50},
	  {    0.0,    7.0,   11.0,   15.0,   14.5},
	  {    0.0,   8.85,   7.24,   5.36,   3.39},
	  { 0.0E-3,0.25E-3,0.32E-3,0.81E-3,0.62E-3},
	  {    0.0,   0.33,   0.46,   0.74,   0.30}};

        // Year Fraction Constants
	double DminNorth = 28.0;
	double DminSouth = 211.0;
        double MOPScosCT = ( 2.0 * Pi / 365.25 );


	// In-Function Constants
	double	latitude;
	double	vectorPTEBL[5],averagePTEBL[5],seasvarPTEBL[5];
	double  Tzo_dry,Tzo_wet,exp1,exp2,Tz_dry,Tz_wet;
	int	i, MOPSPhi,index;
	double	DoY, Dmin;
	int 	year;
	double  factor;
	
	// Latitude Quantimization
	latitude = positionNEU[0]/d2r;
	if ( latitude >=0.0 ){
		Dmin = DminNorth;
	}else {
		Dmin = DminSouth;
	}		
	latitude = fabs( latitude );
        if (  latitude <= 15.0 )                          index=0;
        if ( (latitude >  15.0 ) && (latitude <= 30.0)  ) index=1;
        if ( (latitude >  30.0 ) && (latitude <= 45.0)  ) index=2;
        if ( (latitude >  45.0 ) && (latitude <= 60.0)  ) index=3;
        if ( (latitude >  60.0 ) && (latitude <  75.0)  ) index=4;
        if (  latitude >= 75.0 )                          index=4;


	// Year Fraction
	t2doy(t,&year,&DoY);

	// Interpolate Metereological Parameters
	if ( latitude <= 15.0  || latitude >= 75.0 ) {
		latitude = MOPSlat[index];
		for (i=0; i<5; i++){ 
			averagePTEBL[i] = MOPSaverages[i][index];
			seasvarPTEBL[i] = MOPSvariations[i][index];
			}
	} else { 
		latitude = ( latitude - MOPSlat[index-1] ) / ( MOPSlat[index] - MOPSlat[index-1] ); 
		for (i=0; i<5; i++){ 
			averagePTEBL[i] = MOPSaverages[i][index-1]   + ( MOPSaverages[i][index]   - MOPSaverages[i][index-1]   ) * latitude;
			seasvarPTEBL[i] = MOPSvariations[i][index-1] + ( MOPSvariations[i][index] - MOPSvariations[i][index-1] ) * latitude;
			}
	}

	// Compute metereologial parameters, from the interpolated Average and Seasonal:
	for (i=0; i<5; i++){ 
		vectorPTEBL[i] = averagePTEBL[i] - seasvarPTEBL[i] * cos ( (DoY - Dmin)*MOPScosCT );
	}

	// Zero-altitude zenith delay terms:
        Tzo_dry = MOPSzhydCT * vectorPTEBL[0];	
        Tzo_wet = ( MOPSzwetCT /(MOPSgm*(vectorPTEBL[4]+1.0)-vectorPTEBL[3]*MOPSRd) ) * (vectorPTEBL[2]/vectorPTEBL[1]);

	factor=vectorPTEBL[3]*positionNEU[2]/vectorPTEBL[1];

	//Sanity check: The result of the calculus Beta*Height/Temperature must be under 1
	if (factor<1) {
		// zenith delay terms at H meters of height above mean-sea-level        
        	exp1=MOPSg/(MOPSRd*vectorPTEBL[3]);
	        Tz_dry=Tzo_dry*pow(1.0-factor,exp1);
       
       		exp2=((vectorPTEBL[4]+1.0)*MOPSg)/(MOPSRd*vectorPTEBL[3])-1.0 ;
	        Tz_wet=Tzo_wet*pow(1.0-factor,exp2);
	} else {
		Tz_dry=0;
		Tz_wet=0;
	}

	// Return Nominals
	*tropWetMOPS_Nominal = Tz_dry;
	*tropDryMOPS_Nominal = Tz_wet;
	

}


/*****************************************************************************
 * Name        : klobucharModel
 * Description : Obtains the ionospheric correction using the klobuchar model
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  I  N/A  TEpoch structure
 * TGNSSproducts  *products        I  N/A  Ephemerides products to obtain the
 *                                         parameters for the model
 * int  satIndex                   I  N/A  Index of the satellite to obtain 
 *                                         the correction
 * Returned value (double)         O  TECU Klobuchar correction in STEC (in 
 *                                         electrons per area unit)
 *****************************************************************************/
double klobucharModel (TEpoch *epoch, TGNSSproducts *products, int satIndex) {
	double			fm;
	double			ilat;
	double			ilon;
	double			gmlat;
	double			tsub;
	double			slantFactor;
	double			alpha;
	double			beta;
	double			x;
	double			y;
	double			ionoDelay;
	double                  h = KLOBUCHAR_HEIGHT;

	
	if (products->type != BRDC) return 0;

        // Calculate the Earth-centered angle fm (in radians):
        fm = 0.5*Pi -1.0*epoch->sat[satIndex].elevation-asin(EARTH_RADIUS/(EARTH_RADIUS+h)*cos(epoch->sat[satIndex].elevation));

        // Compute the subionospheric latitude, ilat (in radians):
        ilat = asin(sin(epoch->receiver.aproxPositionNEU[0])*cos(fm) + cos(epoch->receiver.aproxPositionNEU[0])*sin(fm)*cos(epoch->sat[satIndex].azimuth));

        // Compute the subionospheric longitude, ilon (in radians):
        ilon = epoch->receiver.aproxPositionNEU[1] + fm*sin(epoch->sat[satIndex].azimuth)/cos(ilat);

        // Find the Geomagnetic latitude, gmlat (in radians):
        gmlat = asin(sin(ilat)*sin(GEOMAGNETIC_POLE_LAT) + cos(ilat)*cos(GEOMAGNETIC_POLE_LAT)*cos(ilon-GEOMAGNETIC_POLE_LON));

        // Find the local time at subionospheric point, tsub (in sec):
        tsub = 43200*ilon/Pi + epoch->t.SoD;
        if (tsub>86400) tsub -= 86400;
        else if (tsub<0) tsub += 86400;

        // Convert to slant time delay:
        slantFactor = 1.0/sqrt(1.0-pow(EARTH_RADIUS/(EARTH_RADIUS+h)*cos(epoch->sat[satIndex].elevation),2));



        // Compute the ionospheric delay STEC (in metres of delay):

        gmlat =gmlat/Pi;

	alpha = products->BRDC->ionA[0] + 
			products->BRDC->ionA[1] * gmlat +
			products->BRDC->ionA[2] * pow(gmlat,2) +
			products->BRDC->ionA[3] * pow(gmlat,3);
	beta =  products->BRDC->ionB[0] + 
			products->BRDC->ionB[1] * gmlat +
			products->BRDC->ionB[2] * pow(gmlat,2) +
			products->BRDC->ionB[3] * pow(gmlat,3);

	if (alpha<0) alpha = 0;
	if (beta<72000) beta = 72000;
	//else if (beta>144000) beta = 144000;  //Uncomment this line to fix Klobuchar excessive period

        x = 2.0*Pi*(tsub-50400)/beta;
        if (fabs(x)<=1.57) {
                y = alpha * cos(x);
        } else {
                y = 0.0;
        }


	ionoDelay = slantFactor*(5e-9+y)*c; // in L1 meters

	// Change from L1 meters to TECU 
	ionoDelay = ionoDelay / 40.3 * GPSf1 * GPSf1 * pow(10,-16); // in TECUs

	return ionoDelay;
}

/*****************************************************************************
 * Name        : beiDouKlobucharModel
 * Description : Obtains the ionospheric correction using the BeiDou 
 * 		 klobuchar model
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  I  N/A  TEpoch structure
 * TGNSSproducts  *products        I  N/A  Ephemerides products to obtain the
 *                                         parameters for the model
 * int  satIndex                   I  N/A  Index of the satellite to obtain 
 *                                         the correction
 * Returned value (double)         O  TECU Klobuchar correction in STEC (in 
 *                                         electrons per area unit)
 *****************************************************************************/
double beiDouKlobucharModel (TEpoch *epoch, TGNSSproducts *products, int satIndex) {
	double			fm;
	double			ilat;
	double			ilon;
	double			gmlat;
	double			tsub;
	double			slantFactor;
	double			alpha;
	double			beta;
	double			x;
	double			y;
	double			ionoDelay;
	double                  h = BEIDOU_HEIGHT;
	
        if (products->type != BRDC) return 0;                                                          
                                                                                                       
        // Calculate the Earth-centered angle fm (in radians):                                         
        fm = 0.5*Pi -1.0*epoch->sat[satIndex].elevation-asin(EARTH_RADIUS/(EARTH_RADIUS+h)*cos(epoch->sat[satIndex].elevation));        

        // Compute the subionospheric latitude, ilat (in radians):                                     
        ilat = asin(sin(epoch->receiver.aproxPositionNEU[0])*cos(fm) + cos(epoch->receiver.aproxPositionNEU[0])*sin(fm)*cos(epoch->sat[satIndex].azimuth));                                                    
        
        // Compute the subionospheric longitude, ilon (in radians):                                    
        ilon = epoch->receiver.aproxPositionNEU[1] + fm*sin(epoch->sat[satIndex].azimuth)/cos(ilat);   
        
        // Find the Geomagnetic latitude, gmlat (in radians):
        gmlat = asin(sin(ilat)*sin(GEOMAGNETIC_POLE_LAT) + cos(ilat)*cos(GEOMAGNETIC_POLE_LAT)*cos(ilon-GEOMAGNETIC_POLE_LON));         
        
        // Find the local time at subionospheric point, tsub (in sec):
        tsub = 43200*ilon/Pi + epoch->t.SoD;                                                           
        if (tsub>86400) tsub -= 86400;                                                                 
        else if (tsub<0) tsub += 86400;
        
        // Convert to slant time delay:                                                                
        slantFactor = 1.0/sqrt(1.0-pow(EARTH_RADIUS/(EARTH_RADIUS+h)*cos(epoch->sat[satIndex].elevation),2));                                                                                                  
                                                                                                       
        
        
        // Compute the ionospheric delay STEC (in metres of delay):                                    
                                                                                                       
        gmlat =gmlat/Pi;                                                                               
                                                                                                       
        alpha = products->BRDC->bdsA[0] + 
                        products->BRDC->bdsA[1] * gmlat +                                              
                        products->BRDC->bdsA[2] * pow(gmlat,2) +                                       
                        products->BRDC->bdsA[3] * pow(gmlat,3);                                        
        beta =  products->BRDC->bdsB[0] +                                                              
                        products->BRDC->bdsB[1] * gmlat +                                              
                        products->BRDC->bdsB[2] * pow(gmlat,2) +                                       
                        products->BRDC->bdsB[3] * pow(gmlat,3);                                        
        
        if (alpha<0) alpha = 0;
        if (beta<72000) beta = 72000;
	//New condition for BeiDou. This condition does not restrict that the semiperiod shall not be greater than 10 hours.
	//The right condition should be overlap with the next day should be the following:
	//if (beta>144000) beta = 144000; (instead of "if (beta>172800) beta = 172800;" given by BeiDou ICD)
	if (beta>172800) beta = 172800;
                        
        x = 2.0*Pi*(tsub-50400)/beta;                                                                  
        if (fabs(x)<=1.57) { 
                y = alpha * cos(x);
        } else {
                y = 0.0;
        }

        ionoDelay = slantFactor*(5e-9+y)*c; // in L1 meters
   	

        // Change from L1 meters to TECU 
        ionoDelay = ionoDelay / 40.3 * GPSf1 * GPSf1 * pow(10,-16); // in TECUs


	return ionoDelay;
}

/*****************************************************************************
 * Name        : ionexModel
 * Description : Obtains the ionospheric correction using IONEX file data 
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  I  N/A  TEpoch structure
 * int  satIndex                   I  N/A  Index of the satellite to obtain 
 * TIONEX *ionex		   I  N/A  TIONEX structure
 * TOptions  *options              I  N/A  TOptions structure
 * Returned value (double)         O  TECU Ionex correction in STEC (in 
 *                                         electrons per area unit)
 *					   9999 is returned when one of the
 *					   points needed for the interpolation
 *					   is not set in the IONEX file
 *****************************************************************************/
double ionexModel (TEpoch *epoch, int satIndex, TIONEX *ionex, TOptions *options) {
	
	int			i,j;
	int			diffEarlierMap;
	int			diffLaterMap;
	int			diffDaysIonex;
	int			diffDaysEpoch;
	int			diffSecIonex;
	int			diffSecEpoch;
	int			totalSecIonex;
	int			totalSecEpoch;
	int			mapTimePosition;
	int			posLongitude[2];
	int			posLatitude[2];
	int 			mapType = 0; // 0->TEC 1->RMS 2->HGT	
	int			ilat,ilon;
	int			onlyLastMapUsed = 0;
	double			ilonvalue;
	double			unitaryEarlierTimeDiff;
	double			currentHgtInterpolation;
	double			currentRMSInterpolation;
	double                  TECfinalInterpolation;
	double                  RMSfinalInterpolation;
        double                  TEClowLeft,TECupLeft,TEClowRight,TECupRight;
	double			RMSlowLeft,RMSupLeft,RMSlowRight,RMSupRight;
	double			unitaryLonDiff, unitaryLatDiff, unitaryHgtDiff;
	double			totalLongitude;
	double			currentMapAltitude[ionex->map[0].nhgt];
	double			lowLeftpoint[3]; //Phi,Longitude,Altitude
	double			upLeftpoint[3];
	double			lowRightpoint[3];
	double			upRightpoint[3];
	double			TECtimeInterpolation[2];
	double			RMStimeInterpolation[2];
	TIONEXPiercePoint       *ionexPiercePointvector = NULL;



	//Auxiliary variables to avoid repetitive calculus and make code more readable
	diffDaysIonex = ionex->endTime.MJDN - ionex->startTime.MJDN;
	diffSecIonex  = diffDaysIonex*NUMBERSECONDSPERDAY;
	totalSecIonex = diffSecIonex + (int)ionex->endTime.SoD ;

	diffDaysEpoch = (int) (epoch->t.MJDN - ionex->startTime.MJDN);
	diffSecEpoch  = diffDaysEpoch*NUMBERSECONDSPERDAY;
	totalSecEpoch = diffSecEpoch + (int)epoch->t.SoD;

	totalLongitude = fabs(ionex->map[0].lon1) + fabs(ionex->map[0].lon2);


	//Check that the IONO map day matches witch RINEX observation day
	if ( epoch->t.MJDN < ionex->startTime.MJDN || totalSecEpoch > totalSecIonex ) {
                epoch->sat[satIndex].validIonex = 0;
                return 9999;
	}

	//Look for map with current time. If current time is between two maps, it will start using the earlier one
 	//Also, it necessary to check if the interval value is 0. In that case, time interval is variable	
	if (ionex->interval==0)
	{
		mapTimePosition=0;
		do {
			mapTimePosition++;
			if ( ( (mapTimePosition +1) == ionex->numMaps) && ( totalSecEpoch <= totalSecIonex) ) { break; }
			if (mapTimePosition == ionex->numMaps ){
				                epoch->sat[satIndex].validIonex = 0;
				                return 9999;
			}
		}while ( (diffSecIonex + (int)ionex->map[mapTimePosition].t.SoD) < totalSecEpoch );

		if ( totalSecEpoch == totalSecIonex ){  //Special case: Only using the last map
			mapTimePosition = ionex->numMaps -1;
			diffEarlierMap = 0;
			diffLaterMap   = 0;
			unitaryEarlierTimeDiff = 0;
			onlyLastMapUsed =1;
		} else {
			mapTimePosition--;
			diffEarlierMap = totalSecEpoch - (ionex->map[mapTimePosition].t.MJDN - ionex->startTime.MJDN)*NUMBERSECONDSPERDAY  - (int)ionex->map[mapTimePosition].t.SoD;
			diffLaterMap = (ionex->map[mapTimePosition+1].t.MJDN - ionex->startTime.MJDN)*NUMBERSECONDSPERDAY + (int)(ionex->map[mapTimePosition+1].t.SoD - totalSecEpoch);
			unitaryEarlierTimeDiff = (double)diffEarlierMap / ((double)(diffEarlierMap + diffLaterMap) );
		}
	} else{

		if ( totalSecEpoch == totalSecIonex ){  //Special case: Only using the last map
                        mapTimePosition = ionex->numMaps -1;
                        diffEarlierMap = 0;
                        diffLaterMap   = 0;
                        unitaryEarlierTimeDiff = 0;
                        onlyLastMapUsed =1;
                } else {
		        diffEarlierMap = totalSecEpoch % ionex->interval;
  			diffLaterMap = ionex->interval - diffEarlierMap;
			unitaryEarlierTimeDiff = (double) (diffEarlierMap)/(double)(ionex->interval);
        		mapTimePosition = ( totalSecEpoch - diffEarlierMap ) / ionex->interval;
		}

	}
	//It will be necessary to find a Pierce Point for each altitude,
	//therefore needing the initial and ending height, and the 
	//number of steps, which are available in any map, so the first 
	//map will be used to gather this data.
	//The interpolation will be done twice, one for each map time

	TECtimeInterpolation[0]=TECtimeInterpolation[1] = 0;
	RMStimeInterpolation[0]=RMStimeInterpolation[1] = 0;
	
	for(j=0;j < 2 - onlyLastMapUsed;j++) {
		
		mapTimePosition += j;
		

		for(i=0;i < ionex->map[0].nhgt; i++) {
			
			//ionexPiercePointvector = malloc(sizeof(TIONEXPiercePoint)*(i+1));
			ionexPiercePointvector = (TIONEXPiercePoint*) (realloc(ionexPiercePointvector,sizeof(TIONEXPiercePoint)*(i+1)));
			initIonexPiercePoint(&ionexPiercePointvector[i]);
			currentMapAltitude[i] = ionex->map[0].hgt1 + ionex->map[0].dhgt*i; //Stored in Km 
			
			
			//Calculate the Iono Pierce Point at current height
			getIonoPiercePoint(epoch, &ionexPiercePointvector[i], satIndex, ionex->baseRadius, currentMapAltitude[i]);

			//Check if latitude value is in range (some IONEX files, latitude start or end at a value smaller than 90º, thus not covering the poles or more. If this PiercePoint is in the pole and IONEX file has no data, this points must be discarded
			if (ionex->map[mapTimePosition].dlat > 0 ) {
				if ( (ionexPiercePointvector[i].fi < ionex->map[mapTimePosition].lat1) || (ionexPiercePointvector[i].fi > ionex->map[mapTimePosition].lat2) ) {
                                	epoch->sat[satIndex].validIonex = 0;
                                	return 9999;
				}
			} else {
				if ( (ionexPiercePointvector[i].fi > ionex->map[mapTimePosition].lat1) || (ionexPiercePointvector[i].fi < ionex->map[mapTimePosition].lat2) ) {
                                	epoch->sat[satIndex].validIonex = 0;
                                	return 9999;
				}
			}

			//Rotate maps adding latitude proportional to time difference between maps
			if (j==0) { ionexPiercePointvector[i].longitude = ionexPiercePointvector[i].longitude + (double)diffEarlierMap*(1/(double)NUMBERSECONDSPERHOUR)*(360.0/24.0) ; }
			else { ionexPiercePointvector[i].longitude = ionexPiercePointvector[i].longitude - (double)diffLaterMap*(1/(double)NUMBERSECONDSPERHOUR)*(360.0/24.0) ; }
			
			//In this operation, we add 180 to move the range to 0..360 (only if longitude range is negative) and then do the module to assure the value is in range
			if ( (ionex->map[mapTimePosition].lon1 < 0) || ( ionex->map[mapTimePosition].lon2 < 0 ) ) {
				ionexPiercePointvector[i].longitude = modulo (ionexPiercePointvector[i].longitude+180.0+360.0,360);
			} else {
				ionexPiercePointvector[i].longitude = modulo (ionexPiercePointvector[i].longitude+360.0,360);
			}

			//Check if any of the values lon1 or lon2 (the limits in the range) are negative. If any is negative, it means that the range is from -180 to 180 (or 180..-180), therefore having to substract 180
			if ( (ionex->map[mapTimePosition].lon1 < 0) || ( ionex->map[mapTimePosition].lon2 < 0 ) ) {
				ionexPiercePointvector[i].longitude -=180.0;
			}

			//Check if longitude value is within the data given in the IONEX file (in the case that the longitude range is not 360 or 360-dlon)
			if ( (totalLongitude != (360.0 - fabs(ionex->map[mapTimePosition].dlon))) && (totalLongitude != (360.0 + fabs(ionex->map[mapTimePosition].dlon))) && (totalLongitude != 360.0)  ) {
				if (ionex->map[mapTimePosition].dlon > 0 ) {
					if ( (ionexPiercePointvector[i].longitude < ionex->map[mapTimePosition].lon1) || (ionexPiercePointvector[i].longitude > ionex->map[mapTimePosition].lon2) ) {
                               		 	epoch->sat[satIndex].validIonex = 0;
                                		return 9999;
					}
				} else {
					if ( (ionexPiercePointvector[i].longitude > ionex->map[mapTimePosition].lon1) || (ionexPiercePointvector[i].longitude < ionex->map[mapTimePosition].lon2) ) {
                               	 		epoch->sat[satIndex].validIonex = 0;
                                		return 9999;
					}
				}
			}


			//Calculate latitude and longitude for the four TEC values surrounding current pierce point
			//Given one position, it is first calculated the point which is on the bottom left hand corner of the TEC values surrounding the pierce point
       			if (ionex->map[mapTimePosition].nlat==1) {
               			 ilat=0;
        		} else {
               			 ilat = (int)((ionexPiercePointvector[i].fi - ionex->map[mapTimePosition].lat1)/ionex->map[mapTimePosition].dlat);
        		}

		        if (ionex->map[mapTimePosition].nlon==1) {
               			 ilon=0;
		        } else {
		        	ilon = (int)((ionexPiercePointvector[i].longitude - ionex->map[mapTimePosition].lon1)/ionex->map[mapTimePosition].dlon);
				ilonvalue = (ionexPiercePointvector[i].longitude - ionex->map[mapTimePosition].lon1)/ionex->map[mapTimePosition].dlon;
		        }

			lowLeftpoint[0] = ilat*ionex->map[mapTimePosition].dlat + ionex->map[mapTimePosition].lat1;
			lowLeftpoint[1] = ilon*ionex->map[mapTimePosition].dlon + ionex->map[mapTimePosition].lon1;
			lowLeftpoint[2] = currentMapAltitude[i];

			//Special case: If IONEX Range does not start at 0 or -180 (dlon positive) or at 180 or 360 (dlon positive) and total range = 360.0 - fabs(ionex->map[mapTimePosition].dlon and the pierce point is in range 0..dlon (ilonvalue<0) (or equivalent depending on range given), then lowLeftpoint[1] is lon2 (because position 0 is given at the end) and lowLeftpoint[1] is lon1
			if ( (totalLongitude == (360.0 - fabs(ionex->map[mapTimePosition].dlon))) && (ilonvalue < 0) ) {
				lowLeftpoint[1] = ionex->map[mapTimePosition].lon2;
			}
		
			if ( lowLeftpoint[0] != ionex->map[mapTimePosition].lat2 ){ //The point is not at the end of the map in latitude
				lowRightpoint[0] = lowLeftpoint[0] + ionex->map[mapTimePosition].dlat;
				lowRightpoint[1] = lowLeftpoint[1];
				lowRightpoint[2] = currentMapAltitude[i];
			}

			if ( (lowLeftpoint[1] == ionex->map[mapTimePosition].lon2) && (totalLongitude == (360.0 - fabs(ionex->map[mapTimePosition].dlon)) ) ) {
				//In this IONEX file, the maps are in range 0..375º (or -180..175). Instead of using value of lon=360º, use inital longitude
				upLeftpoint[0] = lowLeftpoint[0];
				upLeftpoint[1] = ionex->map[mapTimePosition].lon1;
				upLeftpoint[2] = currentMapAltitude[i];

			} else if ( lowLeftpoint[1] != ionex->map[mapTimePosition].lon2 ){ //The point is not at the end of the map in longitude
				upLeftpoint[0] = lowLeftpoint[0];
				upLeftpoint[1] = lowLeftpoint[1] + ionex->map[mapTimePosition].dlon;
				upLeftpoint[2] = currentMapAltitude[i];

			}

			if ( lowLeftpoint[0] != ionex->map[mapTimePosition].lat2 ) {
				if ( (lowLeftpoint[1] == ionex->map[mapTimePosition].lon2) && (totalLongitude == (360.0 - fabs(ionex->map[mapTimePosition].dlon)) ) ) { 
					upRightpoint[0] = lowLeftpoint[0] + ionex->map[mapTimePosition].dlat;
					upLeftpoint[1] = ionex->map[mapTimePosition].lon1;
					upRightpoint[2] = currentMapAltitude[i];
				} else if ( lowLeftpoint[1] != ionex->map[mapTimePosition].lon2 ) {
					upRightpoint[0] = lowLeftpoint[0] + ionex->map[mapTimePosition].dlat;
					upRightpoint[1] = lowLeftpoint[1] + ionex->map[mapTimePosition].dlon;
					upRightpoint[2] = currentMapAltitude[i];

				}
			}

			//Recover TEC data from the four points
			TEClowLeft = getIonoValue(&ionex->map[mapTimePosition], mapType, lowLeftpoint[2], lowLeftpoint[0], lowLeftpoint[1]);
			RMSlowLeft = getIonoValue(&ionex->map[mapTimePosition], 1+mapType, lowLeftpoint[2], lowLeftpoint[0], lowLeftpoint[1]);

			if ( lowLeftpoint[0] != ionex->map[mapTimePosition].lat2 ){ 	
				TEClowRight = getIonoValue(&ionex->map[mapTimePosition], mapType, lowRightpoint[2], lowRightpoint[0], lowRightpoint[1]); 
				RMSlowRight = getIonoValue(&ionex->map[mapTimePosition], 1+mapType, lowRightpoint[2], lowRightpoint[0], lowRightpoint[1]); 
			} else {
				TEClowRight = 0;
				RMSlowRight = 0;
			}

			if ( (lowLeftpoint[1] == ionex->map[mapTimePosition].lon2) && (totalLongitude == (360.0 - fabs(ionex->map[mapTimePosition].dlon)) ) ) {
				TECupLeft = getIonoValue(&ionex->map[mapTimePosition], mapType, upLeftpoint[2], upLeftpoint[0], upLeftpoint[1]);
				RMSupLeft = getIonoValue(&ionex->map[mapTimePosition], 1+mapType, upLeftpoint[2], upLeftpoint[0], upLeftpoint[1]);

			} else if ( lowLeftpoint[1] != ionex->map[mapTimePosition].lon2 ){
				TECupLeft = getIonoValue(&ionex->map[mapTimePosition], mapType, upLeftpoint[2], upLeftpoint[0], upLeftpoint[1]);
				RMSupLeft = getIonoValue(&ionex->map[mapTimePosition], 1+mapType, upLeftpoint[2], upLeftpoint[0], upLeftpoint[1]);
			} else {
				TECupLeft = 0;
				RMSupLeft = 0;
			}
			
			if ( lowLeftpoint[0] != ionex->map[mapTimePosition].lat2 ) {
				if ( (lowLeftpoint[1] == ionex->map[mapTimePosition].lon2) && (totalLongitude == (360.0 - fabs(ionex->map[mapTimePosition].dlon)) ) ) { 
					TECupRight = getIonoValue(&ionex->map[mapTimePosition], mapType, upRightpoint[2], upRightpoint[0], upRightpoint[1]);
					RMSupRight = getIonoValue(&ionex->map[mapTimePosition], 1+mapType, upRightpoint[2], upRightpoint[0], upRightpoint[1]);
				} else if ( lowLeftpoint[1] != ionex->map[mapTimePosition].lon2 ) {
					TECupRight = getIonoValue(&ionex->map[mapTimePosition], mapType, upRightpoint[2], upRightpoint[0], upRightpoint[1]);
					RMSupRight = getIonoValue(&ionex->map[mapTimePosition], 1+mapType, upRightpoint[2], upRightpoint[0], upRightpoint[1]);
				} else {		
					TECupRight = 0;
					RMSupRight = 0;
				}
			} else {		
				TECupRight = 0;
				RMSupRight = 0;
			}

			//Check if any value is 9999 (no data for that point)
			if ( TEClowLeft == 9999 || TEClowRight == 9999 || TECupLeft == 9999 || TECupRight == 9999 ) {
				epoch->sat[satIndex].validIonex = 0;
				return 9999;
			}
			if ( (RMSlowLeft == 9999 || RMSlowRight == 9999 || RMSupLeft == 9999 || RMSupRight == 9999) && options->useSigmaIono == 1 ) {
				epoch->sat[satIndex].validIonex = 0;
				return 9999;
			}
			
			
			/*//This commented lines were done for the case that instead of returning 9999 the program exited printing the point and time of the nonset value
			currentHour = (int) ( ionex->map[mapTimePosition].t.SoD/3600 );
			currentMinute = (int) ( (ionex->map[mapTimePosition].t.SoD -currentHour*3600)/60 );
			currentSecond = (int) ( ionex->map[mapTimePosition].t.SoD -currentHour*3600 - currentMinute*60 );
			if (TEClowLeft == 9999){
				sprintf(messagestr,"Value for %lfº latitude, %lfº longitude and %lf altitude at hour %d minute %d second %d in IONEX file is not set\n",lowLeftpoint[0], lowLeftpoint[1],lowLeftpoint[2],currentHour,currentMinute,currentSecond); 
				printError(messagestr,options);

			} else if (TEClowRight == 9999){
				sprintf(messagestr,"Value for %lfº latitude, %lfº longitude and %lf altitude at hour %d minute %d second %d in IONEX file is not set\n",lowRightpoint[0], lowRightpoint[1],lowRightpoint[2],currentHour,currentMinute,currentSecond);                     
                                printError(messagestr,options);

			} else if (TECupLeft == 9999){ 
				sprintf(messagestr,"Value for %lfº latitude, %lfº longitude and %lf altitude at hour %d minute %d second %d in IONEX file is not set\n",upLeftpoint[0], upLeftpoint[1],upLeftpoint[2],currentHour,currentMinute,currentSecond);                        
                                printError(messagestr,options);

			} else if (TECupRight == 9999){
				sprintf(messagestr,"Value for %lfº latitude, %lfº longitude and %lf altitude at hour %d minute %d second %d in IONEX file is not set\n",upRightpoint[0], upRightpoint[1],upRightpoint[2],currentHour,currentMinute,currentSecond);                                 
                                printError(messagestr,options);

			}*/

			//Calculate unitary distance between grid points
			unitaryLatDiff = fabs( (ionexPiercePointvector[i].fi - lowLeftpoint[0])/ionex->map[mapTimePosition].dlat );
			unitaryLonDiff = fabs( (ionexPiercePointvector[i].longitude - lowLeftpoint[1])/ionex->map[mapTimePosition].dlon );

			//Interpolate in 2D
			currentHgtInterpolation = TEClowLeft*(1-unitaryLonDiff)*(1-unitaryLatDiff) + TEClowRight*(1-unitaryLonDiff)*unitaryLatDiff + TECupLeft*unitaryLonDiff*(1-unitaryLatDiff) + TECupRight*unitaryLonDiff*unitaryLatDiff;
			currentRMSInterpolation = RMSlowLeft*(1-unitaryLonDiff)*(1-unitaryLatDiff) + RMSlowRight*(1-unitaryLonDiff)*unitaryLatDiff + RMSupLeft*unitaryLonDiff*(1-unitaryLatDiff) + RMSupRight*unitaryLonDiff*unitaryLatDiff;
			
			//Calculate mapping function
			getMappingFunction(&ionexPiercePointvector[i], ionex->baseRadius, currentMapAltitude[i], ionex->mappingFunction, options);

			//Add current height interpolation with previous interpolation
			TECtimeInterpolation[j] += currentHgtInterpolation*ionexPiercePointvector[i].mappingFunction;
			RMStimeInterpolation[j] += currentRMSInterpolation*ionexPiercePointvector[i].mappingFunction;

	
		} //End of height interpolation

		
	} //End of time interpolation

	if (onlyLastMapUsed == 1) {TECtimeInterpolation[1] = RMStimeInterpolation[1] = 0;} 
	TECfinalInterpolation = TECtimeInterpolation[0]*(1-unitaryEarlierTimeDiff) + TECtimeInterpolation[1]*unitaryEarlierTimeDiff;
	RMSfinalInterpolation = RMStimeInterpolation[0]*(1-unitaryEarlierTimeDiff) + RMStimeInterpolation[1]*unitaryEarlierTimeDiff;
		

	epoch->sat[satIndex].validIonex = 1;
	epoch->sat[satIndex].ionoSigma = RMSfinalInterpolation;	
	//Interpolation is in TECUs

	return TECfinalInterpolation;
}

/*****************************************************************************
 * Name        : FPPPModel
 * Description : Obtains the ionospheric correction using FPPP file data 
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  I  N/A  TEpoch structure
 * int  satIndex                   I  N/A  Index of the satellite to obtain 
 *                                         the parameters of the model
 * TFPPP *fppp			   I  N/A  Tfppp structure
 * TOptions  *options              I  N/A  TOptions structure
 * Returned value (double)         O  TECU Ionex correction in STEC (in 
 *                                         electrons per area unit)
 *					   9999 is returned when one of the
 *					   points needed for the interpolation
 *					   is not set in the IONEX file
 *****************************************************************************/
double FPPPModel (TEpoch *epoch, int satIndex, TFPPPIONEX *fppp, TOptions *options) {
	
	int			i,j;
	int			diffEarlierMap;
	int			diffLaterMap;
	int			diffDaysIonex;
	int			diffDaysEpoch;
	int			diffSecIonex;
	int			diffSecEpoch;
	int			totalSecIonex;
	int			totalSecEpoch;
	int			mapTimePosition;
	int			posLongitude[2];
	int			posLatitude[2];
	int 			mapType = 0; // 0->TEC 1->RMS 2->HGT	
	int			ilat,ilt;
	int			onlyLastMapUsed = 0;
	double			unitaryEarlierTimeDiff;
	double			currentHgtInterpolation;
	double			currentRMSInterpolation;
	double                  TECfinalInterpolation;
	double                  RMSfinalInterpolation;
        double                  TEClowLeft,TECupLeft,TEClowRight,TECupRight;
	double			RMSlowLeft,RMSupLeft,RMSlowRight,RMSupRight;
	double			unitaryLtDiff, unitaryLatDiff, unitaryHgtDiff;
	double			currentMapAltitude[NUMBERFPPPMAPS];
	double			lowLeftpoint[3]; //Phi,Longitude,Altitude
	double			upLeftpoint[3];
	double			lowRightpoint[3];
	double			upRightpoint[3];
	double			TECtimeInterpolation[2];
	double			RMStimeInterpolation[2];
	TIONEXPiercePoint       *ionexPiercePointvector = NULL;


	//Auxiliary variables to avoid repetitive calculus and make code more readable
	diffDaysIonex = fppp->endTime.MJDN - fppp->startTime.MJDN;
	diffSecIonex  = diffDaysIonex*NUMBERSECONDSPERDAY;
	totalSecIonex = diffSecIonex + (int)fppp->endTime.SoD ;

	diffDaysEpoch = (int) (epoch->t.MJDN - fppp->startTime.MJDN);
	diffSecEpoch  = diffDaysEpoch*NUMBERSECONDSPERDAY;
	totalSecEpoch = diffSecEpoch + (int)epoch->t.SoD;

	//Check that the IONO map day matches witch RINEX observation day

	if ( epoch->t.MJDN < fppp->startTime.MJDN || totalSecEpoch > totalSecIonex ) {

                epoch->sat[satIndex].validIonex = 0;
                return 9999;
	}

	//Look for map with current time. If current time is between two maps, it will start using the earlier one
 	//Also, it necessary to check if the interval value is 0. In that case, time interval is variable	
	if (fppp->interval==0)
	{
		mapTimePosition=0;
		do {
			mapTimePosition++;
			if ( ( (mapTimePosition +1) == fppp->numMaps) && ( totalSecEpoch <= totalSecIonex) ) { break; }
			if (mapTimePosition == fppp->numMaps ){
				epoch->sat[satIndex].validIonex = 0;
				return 9999;
			}
		}while ( (diffSecIonex + (int)fppp->map[mapTimePosition].t.SoD) < totalSecEpoch );

		if ( totalSecEpoch == totalSecIonex ){  //Special case: Only using the last map
			mapTimePosition = fppp->numMaps -1;
			diffEarlierMap = 0;
			diffLaterMap   = 0;
			unitaryEarlierTimeDiff = 0;
			onlyLastMapUsed =1;
		} else {
			mapTimePosition--;
			diffEarlierMap = totalSecEpoch - (fppp->map[mapTimePosition].t.MJDN - fppp->startTime.MJDN)*NUMBERSECONDSPERDAY  - (int)fppp->map[mapTimePosition].t.SoD;
			diffLaterMap = (fppp->map[mapTimePosition+1].t.MJDN - fppp->startTime.MJDN)*NUMBERSECONDSPERDAY + (int)(fppp->map[mapTimePosition+1].t.SoD - totalSecEpoch);
			unitaryEarlierTimeDiff = (double)diffEarlierMap / ((double)(diffEarlierMap + diffLaterMap) );
		}
	} else{

		if ( totalSecEpoch == totalSecIonex ){  //Special case: Only using the last map
                        mapTimePosition = fppp->numMaps -1;
                        diffEarlierMap = 0;
                        diffLaterMap   = 0;
                        unitaryEarlierTimeDiff = 0;
                        onlyLastMapUsed =1;
                } else {
		        diffEarlierMap = totalSecEpoch % fppp->interval;
  			diffLaterMap = fppp->interval - diffEarlierMap;
			unitaryEarlierTimeDiff = (double) (diffEarlierMap)/(double)(fppp->interval);
        		mapTimePosition = ( totalSecEpoch - diffEarlierMap ) / fppp->interval;
		}

	}
	//It will be necessary to find a Pierce Point for each altitude,
	//therefore needing the initial and ending height, and the 
	//number of steps, which are available in any map, so the first 
	//map will be used to gather this data.
	//The interpolation will be done twice, one for each map time

	TECtimeInterpolation[0]=TECtimeInterpolation[1] = 0;
	RMStimeInterpolation[0]=RMStimeInterpolation[1] = 0;
	
	for(j=0;j < 2 - onlyLastMapUsed;j++) {
		
		mapTimePosition += j;
		

		for(i=0;i < NUMBERFPPPMAPS; i++) {
			
			ionexPiercePointvector = (TIONEXPiercePoint*) (realloc(ionexPiercePointvector,sizeof(TIONEXPiercePoint)*(i+1)));
			initIonexPiercePoint(&ionexPiercePointvector[i]);
			currentMapAltitude[i] = fppp->map[0].hgt1 + fppp->map[0].dhgt*i; //Stored in Km 


			//Calculate the Iono Pierce Point at current height
			getIonoPiercePoint(epoch, &ionexPiercePointvector[i], satIndex, fppp->baseRadius, currentMapAltitude[i]);
			//Search for local time position
			ionexPiercePointvector[i].localtime = ionexPiercePointvector[i].longitude + epoch->t.SoD*(1/(double)NUMBERSECONDSPERHOUR)*15.0;
			//Move localtime value to range 0 360
			ionexPiercePointvector[i].localtime = modulo(ionexPiercePointvector[i].localtime+360.0,360);
			

			//Calculate latitude and local time for the four TEC values surrounding current pierce point
			//Given one position, it is first calculated the point which is on the bottom left hand corner of the TEC values surrounding the pierce point
       			if (fppp->map[mapTimePosition].nlat==1) {
               			 ilat=0;
        		} else {
               			 ilat = (int)((ionexPiercePointvector[i].fi - fppp->map[mapTimePosition].lat1)/fppp->map[mapTimePosition].dlat);
        		}

		        if (fppp->map[mapTimePosition].nlt==1) {
               			ilt=0;
		        } else if (ionexPiercePointvector[i].localtime >= fppp->map[mapTimePosition].lt2){
				ilt = (int)(fppp->map[mapTimePosition].lt2 /fppp->map[mapTimePosition].dlt);	
		        } else {
		                ilt = (int)((ionexPiercePointvector[i].localtime - fppp->map[mapTimePosition].lt1)/fppp->map[mapTimePosition].dlt);
		        }

			lowLeftpoint[0] = ilat*fppp->map[mapTimePosition].dlat + fppp->map[mapTimePosition].lat1;
			lowLeftpoint[1] = ilt*fppp->map[mapTimePosition].dlt + fppp->map[mapTimePosition].lt1;
			lowLeftpoint[2] = currentMapAltitude[i];
		
			
			if ( lowLeftpoint[0] != fppp->map[mapTimePosition].lat2 ){ //The point is not at the end of the map in latitude
				lowRightpoint[0] = lowLeftpoint[0] + fppp->map[mapTimePosition].dlat;
				lowRightpoint[1] = lowLeftpoint[1];
				lowRightpoint[2] = currentMapAltitude[i];

			}

			if ( lowLeftpoint[1] == fppp->map[mapTimePosition].lt2 ){ //The point is at the end of the map in local time
				upLeftpoint[0] = lowLeftpoint[0];
				upLeftpoint[1] = fppp->map[mapTimePosition].lt1;
				upLeftpoint[2] = currentMapAltitude[i];
			} else {
				upLeftpoint[0] = lowLeftpoint[0];
                                upLeftpoint[1] = lowLeftpoint[1] + fppp->map[mapTimePosition].dlt;
                                upLeftpoint[2] = currentMapAltitude[i];
			}

			//if ( lowLeftpoint[0] != fppp->map[mapTimePosition].lat2 && lowLeftpoint[1] != fppp->map[mapTimePosition].lt2 ){ //The point is not at the end of the map in latitude and local time
			if ( lowLeftpoint[0] != fppp->map[mapTimePosition].lat2){ //The point is not at the end of the map in latitude and local time
	                        if ( lowLeftpoint[1] == fppp->map[mapTimePosition].lt2 ){ //The point is at the end of the map in local time
					upRightpoint[0] = lowLeftpoint[0] + fppp->map[mapTimePosition].dlat;
					upRightpoint[1] = fppp->map[mapTimePosition].lt1;
					upRightpoint[2] = currentMapAltitude[i];
				} else {
					upRightpoint[0] = lowLeftpoint[0] + fppp->map[mapTimePosition].dlat;
                                        upRightpoint[1] = lowLeftpoint[1] + fppp->map[mapTimePosition].dlt;
                                        upRightpoint[2] = currentMapAltitude[i];
				}			
			}

			//Recover TEC data from the four points
			if( ionexPiercePointvector[i].fi >= fppp->map[mapTimePosition].lat1 && ionexPiercePointvector[i].fi <= fppp->map[mapTimePosition].lat2){
				TEClowLeft = getFPPPValue(&fppp->map[mapTimePosition], mapType, lowLeftpoint[2], lowLeftpoint[0], lowLeftpoint[1]);
				RMSlowLeft = getFPPPValue(&fppp->map[mapTimePosition], 1-mapType, lowLeftpoint[2], lowLeftpoint[0], lowLeftpoint[1]);
			}else{ //
				TEClowLeft = 9999;
                                epoch->sat[satIndex].validIonex = 0;
                                return 9999;
			}

			if ( lowLeftpoint[0] != fppp->map[mapTimePosition].lat2 ){ 	
				TEClowRight = getFPPPValue(&fppp->map[mapTimePosition], mapType, lowRightpoint[2], lowRightpoint[0], lowRightpoint[1]); 
				RMSlowRight = getFPPPValue(&fppp->map[mapTimePosition], 1-mapType, lowRightpoint[2], lowRightpoint[0], lowRightpoint[1]); 
			}else{
				TEClowRight = 0;
				RMSlowRight = 0;
			}

			TECupLeft = getFPPPValue(&fppp->map[mapTimePosition], mapType, upLeftpoint[2], upLeftpoint[0], upLeftpoint[1]);
			RMSupLeft = getFPPPValue(&fppp->map[mapTimePosition], 1-mapType, upLeftpoint[2], upLeftpoint[0], upLeftpoint[1]);
			
			if ( lowLeftpoint[0] != fppp->map[mapTimePosition].lat2 ){
				TECupRight = getFPPPValue(&fppp->map[mapTimePosition], mapType, upRightpoint[2], upRightpoint[0], upRightpoint[1]);
				RMSupRight = getFPPPValue(&fppp->map[mapTimePosition], 1-mapType, upRightpoint[2], upRightpoint[0], upRightpoint[1]);
			}else{		
				TECupRight = 0;
				RMSupRight = 0;
			}

			//Check if any value is 9999 (no data for that point)
			if ( TEClowLeft == 9999 || TEClowRight == 9999 || TECupLeft == 9999 || TECupRight == 9999 ) {
				epoch->sat[satIndex].validIonex = 0;
				return 9999;
			}

			if ( RMSlowLeft == 9999 || RMSlowRight == 9999 || RMSupLeft == 9999 || RMSupRight == 9999 ) {
				epoch->sat[satIndex].validIonex = 0;
				return 9999;
			}
			
			/*//This commented lines were done for the case that instead of returning 9999 the program exited printing the point and time of the nonset value
			currentHour = (int) ( fppp->map[mapTimePosition].t.SoD/3600 );
			currentMinute = (int) ( (fppp->map[mapTimePosition].t.SoD -currentHour*3600)/60 );
			currentSecond = (int) ( fppp->map[mapTimePosition].t.SoD -currentHour*3600 - currentMinute*60 );
			if (TEClowLeft == 9999){
				sprintf(messagestr,"Value for %lfº latitude, %lfº local time and %lf altitude at hour %d minute %d second %d in IONEX file is not set\n",lowLeftpoint[0], lowLeftpoint[1],lowLeftpoint[2],currentHour,currentMinute,currentSecond); 
				printError(messagestr,options);

			} else if (TEClowRight == 9999){
				sprintf(messagestr,"Value for %lfº latitude, %lfº local time and %lf altitude at hour %d minute %d second %d in IONEX file is not set\n",lowRightpoint[0], lowRightpoint[1],lowRightpoint[2],currentHour,currentMinute,currentSecond);                     
                                printError(messagestr,options);

			} else if (TECupLeft == 9999){ 
				sprintf(messagestr,"Value for %lfº latitude, %lfº local time and %lf altitude at hour %d minute %d second %d in IONEX file is not set\n",upLeftpoint[0], upLeftpoint[1],upLeftpoint[2],currentHour,currentMinute,currentSecond);                        
                                printError(messagestr,options);

			} else if (TECupRight == 9999){
				sprintf(messagestr,"Value for %lfº latitude, %lfº local time and %lf altitude at hour %d minute %d second %d in IONEX file is not set\n",upRightpoint[0], upRightpoint[1],upRightpoint[2],currentHour,currentMinute,currentSecond);                                 
                                printError(messagestr,options);

			}*/

			//Calculate unitary distance between grid points
			unitaryLatDiff = fabs( (ionexPiercePointvector[i].fi - lowLeftpoint[0])/fppp->map[mapTimePosition].dlat );
			unitaryLtDiff = fabs( (ionexPiercePointvector[i].localtime - lowLeftpoint[1])/fppp->map[mapTimePosition].dlt );

			//Interpolate in 2D
			currentHgtInterpolation = TEClowLeft*(1-unitaryLtDiff)*(1-unitaryLatDiff) + TEClowRight*(1-unitaryLtDiff)*unitaryLatDiff + TECupLeft*unitaryLtDiff*(1-unitaryLatDiff) + TECupRight*unitaryLtDiff*unitaryLatDiff;
			currentRMSInterpolation = RMSlowLeft*(1-unitaryLtDiff)*(1-unitaryLatDiff) + RMSlowRight*(1-unitaryLtDiff)*unitaryLatDiff + RMSupLeft*unitaryLtDiff*(1-unitaryLatDiff) + RMSupRight*unitaryLtDiff*unitaryLatDiff;
			
			//Calculate mapping function
			getMappingFunction(&ionexPiercePointvector[i], fppp->baseRadius, currentMapAltitude[i], fppp->mappingFunction, options);

			//Add current height interpolation with previous interpolation
			TECtimeInterpolation[j] += currentHgtInterpolation*ionexPiercePointvector[i].mappingFunction;
			RMStimeInterpolation[j] += currentRMSInterpolation*ionexPiercePointvector[i].mappingFunction;

	
		} //End of height interpolation

		
	} //End of time interpolation

	if (onlyLastMapUsed == 1) {TECtimeInterpolation[1]=RMStimeInterpolation[1]= 0;} 
	TECfinalInterpolation = TECtimeInterpolation[0]*(1-unitaryEarlierTimeDiff) + TECtimeInterpolation[1]*unitaryEarlierTimeDiff;
	RMSfinalInterpolation = RMStimeInterpolation[0]*(1-unitaryEarlierTimeDiff) + RMStimeInterpolation[1]*unitaryEarlierTimeDiff;
	


	//Interpolation is in TECUs 

	if (RMSfinalInterpolation > options->ionoRMS) {
		epoch->sat[satIndex].validIonex = 0;
		TECfinalInterpolation = 9999;
	} else {
		epoch->sat[satIndex].validIonex = 1;
		epoch->sat[satIndex].ionoSigma = RMSfinalInterpolation;
	}



	return TECfinalInterpolation;
}

/*****************************************************************************
 * Name        : gravitationalDelayCorrection
 * Description : Obtains the gravitational delay correction for the effect of 
 *               general relativity (red shift) to the GPS signal
 * Parameters  :
 * Name                           |Da|Unit|Description
 * double  *receiverPosition       I  m    Position of the receiver
 * double  *satellitePosition      I  m    Position of the satellite
 * Returned value (double)         O  m    Gravitational delay correction
 *****************************************************************************/
double gravitationalDelayCorrection (double *receiverPosition, double *satellitePosition) {
	double	receiverModule;
	double	satelliteModule;
	double	distance;
	
	receiverModule = sqrt(receiverPosition[0]*receiverPosition[0] + receiverPosition[1]*receiverPosition[1] + receiverPosition[2]*receiverPosition[2]);
	satelliteModule = sqrt(satellitePosition[0]*satellitePosition[0] + satellitePosition[1]*satellitePosition[1] + satellitePosition[2]*satellitePosition[2]);
	distance = sqrt((satellitePosition[0]-receiverPosition[0])*(satellitePosition[0]-receiverPosition[0]) +
					(satellitePosition[1]-receiverPosition[1])*(satellitePosition[1]-receiverPosition[1]) +
					(satellitePosition[2]-receiverPosition[2])*(satellitePosition[2]-receiverPosition[2]));
	
	return 2.0*MU/(c*c) * log((satelliteModule+receiverModule+distance)/(satelliteModule+receiverModule-distance));
}

/*****************************************************************************
 * Name        : getTGDfromDCB
 * Description : Returns the Total Group Delay from a DCB source. It allows
 *               specifying the source and final measurements (NA for absolute)
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TTGDdata  *tgdData              I  N/A  TTGDdata structure to extract the data
 * TReceiver  *rec                 I  N/A  Receiver data
 * GNSSystem  GNSS                 I  N/A  GNSS enumerator of the satellite
 * int  PRN                        I  N/A  PRN of the satellite
 * TTime  *t                       I  N/A  Reference time
 * enum MeasurementType  measFrom  I  N/A  Source measurement (from)
 * enum MeasurementType  measTo    I  N/A  Final measurement (to)
 * double  *dcb                    O    m  Total DCB
 * int  *retDCB                    O  N/A  Status of DCB seek
 *                                          0 => Failed to find proper DCBs
 *                                          1 => Fully OK (all DCBs found)
 * int  *retDCBC1P1                O  N/A  Status of DCB seek for C1-P1
 *                                         -1 => No info
 *                                          0 => Failed to find proper DCBs
 *                                          1 => OK
 *****************************************************************************/
void getTGDfromDCB (TTGDdata *tgdData, TReceiver *rec, enum GNSSystem GNSS, int PRN, TTime *t, enum MeasurementType measFrom, enum MeasurementType measTo, double *dcb, int *retDCB, int *retDCBC1P1) {
	int			i;
	int			check;
	double		factor;
	double		dcbAux[2];
	int			ret;
  int     retDCB_aux;
  int     retDCBC1P1_aux;
	
//	if (tgdData->tgdSource!=tgdDCB) {
//    *retDCB = 0;
//    *retDCBC1P1 = -1;
//    return;
//  }
	
	if (measFrom==measTo) {
		*dcb = 0;
    		*retDCB = 1;
   		 *retDCBC1P1 = -1;
		return;
	}
	
	if (whatIs(measFrom)==CarrierPhase) {
		*dcb = 0;
    		*retDCB = 1;
		*retDCBC1P1 = -1;
		return;
	}
	
	// This should be improved for anything that is not GPS
	// Usable measurements: C1C C1P C2P NA
	if (measFrom!=C1C && measFrom!=C1P && measFrom!=C2P && measFrom!=NA) {
    		*retDCB = 0;
    		*retDCBC1P1 = -1;
    		return;
	}
	if (measTo!=C1C && measTo!=C1P && measTo!=C2P && measTo!=NA) {
		*retDCB = 0;
    		*retDCBC1P1 = -1;
		return;
  	}
	
	// Check if there is a direct path
	for (i=0;i<tgdData->DCB->n;i++) {
		check = 0;
		// Measurements checking
		if (tgdData->DCB->block[i].measFrom == measFrom && tgdData->DCB->block[i].measTo == measTo) {
			check = 1;
			factor = 1;
		} else if (tgdData->DCB->block[i].measFrom == measTo && tgdData->DCB->block[i].measTo == measFrom) {
			check = 1;
			factor = -1;
		}

		if (check) {
			// Time checking
			if (tdiff(t,&tgdData->DCB->block[i].startTime)>=0 && tdiff(t,&tgdData->DCB->block[i].endTime)<=0) {
				// Availability checking
				if (tgdData->DCB->block[i].available[GNSS][PRN]) {
					*dcb = -tgdData->DCB->block[i].dcb[GNSS][PRN] * factor;
					if (rec->recType == rtCROSS) { // Additional correction for Cross-correlation receivers
						if (measFrom == C2P || measTo == C2P) { 
							getTGDfromDCB (tgdData, rec ,GNSS, PRN, t, C1C, C1P, &dcbAux[0], &retDCB_aux, &retDCBC1P1_aux);
							if (retDCB_aux) {
								if (measFrom == C2P) *dcb += dcbAux[0];
								else *dcb -= dcbAux[0];
							} else {
								*retDCB = 0;
								*retDCBC1P1 = -1;
								return;
							}
						}
					}
					*retDCB = 1;
					*retDCBC1P1 = -1;
					return;
				}
			}
		}
	}
	
	// If no direct path was found [To be improved]
	if (measFrom==C1C && measTo==NA) {
		getTGDfromDCB (tgdData, rec, GNSS, PRN, t, C1C, C1P, &dcbAux[0], &retDCB_aux, &retDCBC1P1_aux);
		if (retDCB_aux)  {
			getTGDfromDCB (tgdData, rec, GNSS, PRN, t, C1P, NA, &dcbAux[1], &retDCB_aux, &retDCBC1P1_aux);
			if (retDCB_aux) {
				*dcb = dcbAux[0] + dcbAux[1];
				*retDCB = 1;
				*retDCBC1P1 = 1;
				return;
			} else {
				*dcb = dcbAux[0];
				*retDCB = 0;
				*retDCBC1P1 = 1;
				return;
			}
			      *retDCB = 0;
			      *retDCBC1P1 = 0;
		}
	} else if (measFrom==C2P && measTo==NA) {
		getTGDfromDCB (tgdData, rec, GNSS, PRN, t, C2P, C1P, &dcbAux[0], &retDCB_aux, &retDCBC1P1_aux);
		if (retDCB_aux)  {
			getTGDfromDCB (tgdData, rec, GNSS, PRN, t, C1P, NA, &dcbAux[1], &retDCB_aux, &retDCBC1P1_aux);
			if (retDCB_aux) {
				*dcb = dcbAux[0] + dcbAux[1];
				*retDCB = 1;
				*retDCBC1P1 = -1;
				return;
			} else {
				*dcb = dcbAux[0];
				*retDCB = 0;
				*retDCBC1P1 = -1;
				return;
			}
		}
	}
	*dcb = 0;
	*retDCB = 0;
	return;
}

/*****************************************************************************
 * Name        : TGDCorrectionBRDC
 * Description : Obtains the TGD correction (Total Group Delay biases) from a 
 *               RINEX navigation file
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  I  N/A  TEpoch structure
 * int  satIndex                   I  N/A  Index of the satellite to obtain 
 *                                         the correction
 * TTime  *t                       I  N/A  Reference time
 * TBRDCproducts  *products        I  N/A  Ephemerides products to obtain the 
 *                                         parameters for the model
 * TOptions  *options              I  N/A  TOptions structure
 * Returned value (double)         O  mHz2 TGD correction (to apply it to a 
 *                                         specific frequency it must be divided 
 *                                         by f^2)
 *****************************************************************************/
double TGDCorrectionBRDC (TEpoch *epoch, int satIndex, TTime *t, TBRDCproducts *prod, TOptions *options) {
	TBRDCblock		*block;
	double			correction;

	if (prod == NULL) return 0;

	block = selectBRDCblock(prod,t,epoch->sat[satIndex].GNSS,epoch->sat[satIndex].PRN,options);

	if (block==NULL) return 0;

	correction = block->TGD*c*GPSf1*GPSf1;
	return correction;
}

/*****************************************************************************
 * Name        : TGDCorrectionIONEX
 * Description : Obtains the TGD correction (Total Group Delay biases) from a 
 *               IONEX file
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  I  N/A  TEpoch structure
 * int  satIndex                   I  N/A  Index of the satellite to obtain 
 *                                         the correction
 * TIonexDCB *ionexDCB             I  N/A  DCB data in IONEX file
 * TOptions  *options              I  N/A  TOptions structure
 * Returned value (double)         O  mHz2 DCB correction (to apply it to a 
 *                                         specific frequency it must be divided 
 *                                         by f^2)
 *****************************************************************************/
double TGDCorrectionIONEX (TEpoch *epoch, int satIndex,TIonexDCB *ionexDCB,TOptions *options) {

        double dcb;


        dcb = ionexDCB->DCB[epoch->sat[satIndex].GNSS][epoch->sat[satIndex].PRN];


        if (epoch->sat[satIndex].GNSS==GPS) {
                return -1*dcb*ALPHAGPS12*GPSf1*GPSf1*c*1E-9;
        } else {
        //Update for Galileo and other constellations !!
                return 0;
        }

}

/*****************************************************************************
 * Name        : TGDCorrectionFPPP
 * Description : Obtains the TGD correction (Total Group Delay biases) from a 
 *               FPPP file
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  I  N/A  TEpoch structure
 * int  satIndex                   I  N/A  Index of the satellite to obtain 
 *                                         the correction
 * TFPPPDCB *fpppDCB               I  N/A  DCB data in FPPP file
 * TOptions  *options              I  N/A  TOptions structure
 * Returned value (double)         O  mHz2 DCB correction (to apply it to a 
 *                                         specific frequency it must be divided 
 *                                         by f^2)
 *****************************************************************************/
double TGDCorrectionFPPP (TEpoch *epoch, int satIndex,TFPPPDCB *fpppDCB,TOptions *options) {

	double dcb;


	dcb = fpppDCB->IFB[epoch->sat[satIndex].GNSS][epoch->sat[satIndex].PRN];


	if (epoch->sat[satIndex].GNSS==GPS) {
		return -1*dcb*ALPHAGPS12*GPSf1*GPSf1;
	} else {
	//Update for Galileo and other constellations !!
		return 0;
	}
}

/*****************************************************************************
 * Name        : TGDCorrectionBRDC
 * Description : Obtains the TGD correction (Total Group Delay biases) from a 
 *               TTGDdata structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  I  N/A  TEpoch structure
 * int  satIndex                   I  N/A  Index of the satellite to obtain 
 *                                         the correction
 * TTime  *t                       I  N/A  Reference time
 * TTGDdata  *tgdData              I  N/A  Source of the TGD data
 * enum MeasurementType  meas      I  N/A  Measurement
 * TOptions  *options              I  N/A  TOptions structure
 * Returned value (double)         O  mHz2 TGD correction (to apply it to a 
 *                                         specific frequency it must be divided 
 *                                         by f^2)
 *****************************************************************************/
double TGDCorrection (TEpoch *epoch, int satIndex, TTime *t, enum MeasurementType meas, TTGDdata *tgdData,TOptions *options) {
	double	correction1,correction2;
	double	lambda;
	int		  retDCB, retDCBC1P1;
	
	if (whatIs(meas)==CarrierPhase) return 0;
	
	if (tgdData->p1p2dcbModel == p1p2RINEX) {
		correction2 = TGDCorrectionBRDC(epoch,satIndex,t,tgdData->BRDC,options);
		lambda = getLambda(epoch->sat[satIndex].GNSS,meas);
		correction2 = correction2 * lambda * lambda / (c*c);
	} else if (tgdData->p1p2dcbModel == p1p2DCB) {
	    if (meas==C1P || meas==C2P || meas==C1C) {
		if (meas==C1C) {
		    getTGDfromDCB (tgdData, &epoch->receiver, epoch->sat[satIndex].GNSS, epoch->sat[satIndex].PRN, t, C1P, NA, &correction2, &retDCB, &retDCBC1P1);
      		} else {
		    getTGDfromDCB (tgdData, &epoch->receiver, epoch->sat[satIndex].GNSS, epoch->sat[satIndex].PRN, t, meas, NA, &correction2, &retDCB, &retDCBC1P1);
      		}
      		if (retDCB==0) epoch->sat[satIndex].hasDCBs = 0;
		if (retDCBC1P1==0) epoch->sat[satIndex].hasDCBsofC1P1 = 0;
    	 	}
	} else if (tgdData->p1p2dcbModel == p1p2IONEX) {
		correction2 = TGDCorrectionIONEX(epoch,satIndex,&tgdData->ionexDCB,options);
		lambda = getLambda(epoch->sat[satIndex].GNSS,meas);
		correction2 = correction2 * lambda * lambda / (c*c);
	} else if (tgdData->p1p2dcbModel == p1p2FPPP) {
		correction2 = TGDCorrectionFPPP(epoch,satIndex,&tgdData->fpppDCB,options);
		lambda = getLambda(epoch->sat[satIndex].GNSS,meas);
		correction2 = correction2 * lambda * lambda / (c*c);
	} else {
		correction2 = 0;
    		epoch->sat[satIndex].hasDCBs = 1;
		epoch->sat[satIndex].hasDCBsofC1P1 = 1;
	}

	if (tgdData->p1c1dcbModel == p1c1STRICT) {
    		if (meas==C1C) {
		    getTGDfromDCB (tgdData, &epoch->receiver, epoch->sat[satIndex].GNSS, epoch->sat[satIndex].PRN, t, C1C, C1P, &correction1, &retDCB, &retDCBC1P1);
      		    if (retDCB==0) epoch->sat[satIndex].hasDCBs = 0;
		    if (retDCBC1P1==0) epoch->sat[satIndex].hasDCBsofC1P1 = 0;
    		} else {
	            correction1 = 0;
    		}
	} else {
    		correction1 = 0;
  	}

	return correction1 + correction2;
}

/*****************************************************************************
 * Name        : fillSolidTideDisplacement
 * Description : Fills the solid tide displacement of the receiver (XYZ) in the 
 *               TEpoch structure. Model extracted from IERS Conventions 2003, 
 *               Chapter 7 for solid tides of degree 2 and 3
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  IO N/A  TEpoch structure
 *****************************************************************************/
void fillSolidTideDisplacement (TEpoch *epoch) {
	const double	h0 =  0.6078;	// Love number h0 of degree 2
	const double	h2 = -0.0006;	// Love number h2 of degree 2
	const double	h3 =  0.292;	// Love number of degree 3
	const double	l0 =  0.0847;	// Shida number l0 of degree 2
	const double	l2 =  0.0002;	// Shida number l0 of degree 2
	const double	l3 =  0.015;	// Shide number of degree 3 
	const double	MS2E = 332946.0;	// Ratio of mass Sun to Earth
	const double	MM2E = 0.01230002;	// Ratio of mass Moon to Earth
	double			RecPos2SunPos;
	double			RecPos2MoonPos;
	double			scalarSunRec;
	double			scalarMoonRec;
	double			auxSun[3];
	double			auxMoon[3];
	double			h;
	double			l;
	double			sunPosition[3];
	double			sunDistance;
	double			sunPosUni[3];
	double			moonPosition[3];
	double			moonDistance;
	double			moonPosUni[3];
	double			receiverDistance;
	double			receiverPosUni[3];
	int				i;
	
	
	h = h0 + h2*(3*sin(epoch->receiver.aproxPositionNEU[0])*sin(epoch->receiver.aproxPositionNEU[0]) - 1)/2;
	l = l0 + l2*(3*sin(epoch->receiver.aproxPositionNEU[0])*sin(epoch->receiver.aproxPositionNEU[0]) - 1)/2;

	// Find Sun position
	findSun (&epoch->t, sunPosition);
	
	// Find Moon position
	findMoon(&epoch->t, moonPosition);

	sunDistance = sqrt(sunPosition[0]*sunPosition[0] + sunPosition[1]*sunPosition[1] + sunPosition[2]*sunPosition[2]);
	moonDistance = sqrt(moonPosition[0]*moonPosition[0] + moonPosition[1]*moonPosition[1] + moonPosition[2]*moonPosition[2]);
	receiverDistance = sqrt(epoch->receiver.aproxPosition[0]*epoch->receiver.aproxPosition[0] + epoch->receiver.aproxPosition[1]*epoch->receiver.aproxPosition[1] + epoch->receiver.aproxPosition[2]*epoch->receiver.aproxPosition[2]);

	for (i=0;i<3;i++) {
		sunPosUni[i] = sunPosition[i] / sunDistance;
		moonPosUni[i] = moonPosition[i] / moonDistance;
		receiverPosUni[i] = epoch->receiver.aproxPosition[i] / receiverDistance;
	}
	receiverDistance /= 1000; // To transform from m to km
	
	// Computing scalar products
	scalarSunRec = scalarProd(sunPosUni,receiverPosUni);
	scalarMoonRec = scalarProd(moonPosUni,receiverPosUni);
	
	// Computing auxiliar vectors
	for (i=0;i<3;i++) {
		auxSun[i] = sunPosUni[i] - scalarSunRec*receiverPosUni[i];
		auxMoon[i] = moonPosUni[i] - scalarMoonRec*receiverPosUni[i];
	}
	
	// Distance relations	
	RecPos2SunPos = receiverDistance/sunDistance;
	RecPos2MoonPos = receiverDistance/moonDistance;
	
	// Computing displacements
	for (i=0;i<3;i++) {
		epoch->solidTideDisplacement[i] = 0;
		
		// Degree 2
		epoch->solidTideDisplacement[i] += (
				MS2E * RecPos2SunPos * RecPos2SunPos * RecPos2SunPos * receiverDistance *( 	// Sun
					h * receiverPosUni[i] * (3.0/2.0*scalarSunRec*scalarSunRec - 1.0/2.0) +
					3.0 * l * scalarSunRec * auxSun[i]) +
				MM2E * RecPos2MoonPos * RecPos2MoonPos * RecPos2MoonPos * receiverDistance *( 	// Moon
					h * receiverPosUni[i] * (3.0/2.0*scalarMoonRec*scalarMoonRec - 1.0/2.0) +
					3.0 * l * scalarMoonRec * auxMoon[i])
				) * 1000; // To transform from km to m
		
		// Degree 3
		epoch->solidTideDisplacement[i] += (
				MM2E * RecPos2MoonPos * RecPos2MoonPos * RecPos2MoonPos * RecPos2MoonPos * receiverDistance *(   // Only Moon 
					h3 * receiverPosUni[i] * (5.0/2.0*scalarMoonRec*scalarMoonRec*scalarMoonRec - 3.0/2.0*scalarMoonRec) +
					l3 * (15.0/2.0*scalarMoonRec*scalarMoonRec - 3.0/2.0) * auxMoon[i])
				) * 1000; // To transform from km to m
	}
}

/*****************************************************************************
 * Name        : solidTidesCorrection
 * Description : Obtains the solid tides correction from the solidTideDisplacement 
 *               at TEpoch structure. It projects the XYZ displacement into the
 *               Line-of-Sight of the specified satellite
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  I  N/A  TEpoch structure
 * int  satIndex                   I  N/A  Index of the satellite
 * Returned value (double)         O  m    Solid tides correction
 *****************************************************************************/
double solidTidesCorrection (TEpoch *epoch, int satIndex) {
	return -scalarProd(epoch->sat[satIndex].LoS,epoch->solidTideDisplacement);
}

/*****************************************************************************
 * Name        : fillGeometricDistance
 * Description : Fills the geometric distance in the TEpoch structure from 
 *               the receiver and satellite positions
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  IO N/A  TEpoch structure
 * int  satIndex                   I  N/A  Index of the satellite
 *****************************************************************************/
void fillGeometricDistance(TEpoch *epoch,int satIndex) {
	epoch->sat[satIndex].geometricDistance = 
			sqrt((epoch->receiver.aproxPosition[0]-epoch->sat[satIndex].position[0])*(epoch->receiver.aproxPosition[0]-epoch->sat[satIndex].position[0]) +
				 (epoch->receiver.aproxPosition[1]-epoch->sat[satIndex].position[1])*(epoch->receiver.aproxPosition[1]-epoch->sat[satIndex].position[1]) +
				 (epoch->receiver.aproxPosition[2]-epoch->sat[satIndex].position[2])*(epoch->receiver.aproxPosition[2]-epoch->sat[satIndex].position[2]));
}

/*****************************************************************************
 * Name        : fillMeasurementWeights
 * Description : Fills the measurement weights in the TEpoch structure from 
 *               the options
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  IO N/A  TEpoch structure
 * int  satIndex                   I  N/A  Index of the satellite
 * TOptions  *options              I  N/A  TOptions structure
 *****************************************************************************/
void fillMeasurementWeights(TEpoch *epoch, int satIndex,TOptions *options) {

	int			i;
	int			smoothN;
	double			scale;
	double                  lambda;
	double                  ionosigma;
	enum MeasurementType    meas;
	
	for (i=0;i<options->totalFilterMeasurements;i++) {
		// Compute nominal weights
		epoch->sat[satIndex].measurementWeights[i] = options->fixedMeasurementWeight[i] + options->elevationMeasurementWeight[i][0] * exp(-epoch->sat[satIndex].elevation/options->elevationMeasurementWeight[i][1]);

                //Add IONEX or F-PPP ionosphere sigma in the measurement weight
                if ( (options->ionoModel == IONEX || options->ionoModel == FPPP) && options->useSigmaIono==1 ) {
                        meas=options->measurement[i];
                        if(meas>=ENDMEAS) lambda=0;
                        else  lambda = getLambda(epoch->sat[satIndex].GNSS,meas);
                        ionosigma = epoch->sat[satIndex].ionoSigma*40.3*(lambda*lambda)/(c*c)*pow(10,16);
                        epoch->sat[satIndex].measurementWeights[i] = sqrt ( pow(epoch->sat[satIndex].measurementWeights[i],2) + pow (ionosigma,2));
                }

		// Scale weights if measurement is smoothed
		if (options->smoothEpochs!=0 && options->smoothMeas[i]!=NA) {
			smoothN = epoch->sat[satIndex].arcLength;
			if (smoothN > options->smoothEpochs) {
				smoothN = options->smoothEpochs;
			}
			// Sanity check
			if (smoothN==0) smoothN=1;
			
			scale = sqrt((double)options->smoothEpochs/smoothN);
			//epoch->sat[satIndex].measurementWeights[i] *= scale;
		}
	}
}

/*****************************************************************************
 * Name        : getAntenna
 * Description : Get the antenna structure from its name (as a string)
 * Parameters  :
 * Name                           |Da|Unit|Description
 * char  *name                     I  N/A  Name of the antenna
 * TAntennaList  *antennaList      I  N/A  List of antennas to look for
 * TOptions  *options              I  N/A  TOptions structure
 * Returned value (*TAntenna)      O  N/A  Pointer to the antenna inside TAntennaList
 *                                         If it is NULL, the antenna has not been
 *                                         found
 *****************************************************************************/
TAntenna *getAntenna (char *name, TAntennaList *antennaList, TOptions *options) {
	int			i;
	char		*ca;
	char		ant_name[MAX_ANTENNA_NAME];
	char		ant_radome[MAX_ANTENNA_NAME];
	TAntenna	*aux;
	
	ant_radome[0]='\0';
	strcpy(ant_name,name);
	ca = strstr(ant_name," ");
	if (ca != NULL) {
		*ca = '\0';
		strcpy(ant_radome,&ca[1]);
		trim(ant_radome);
	}

	for (i=0;i<antennaList->numAnt;i++) {
		if (strcmp(name,antennaList->ant[i].type)==0) {
			return &antennaList->ant[i];
		} else if (strcmp(ant_name,antennaList->ant[i].type_ant)==0) {
			if (strcmp(ant_radome,antennaList->ant[i].type_radome)==0) {
				return &antennaList->ant[i];
			} else if (strcmp(ant_radome,"")==0 && strcmp(antennaList->ant[i].type_radome,"NONE")==0) {
				return &antennaList->ant[i];
			}
		}
	}
	
	// If the radome parsing is not strict, try to get the same antenna, but without radome
	if (!options->strictradome && ant_radome[0]!='\0') {
		aux = getAntenna(ant_name,antennaList,options);
		if (aux!=NULL) {
			sprintf(messagestr,"WARNING antenna '%s' not found... using '%s'. Enable radome strict option to avoid this behaviour",name,aux->type);
			printInfo(messagestr,options);
		}
		return aux;
	}
	
	return NULL;
}

/*****************************************************************************
 * Name        : getReceiverType
 * Description : Get the receiver structure from its name (as a string)
 * Parameters  :
 * Name                           |Da|Unit|Description
 * char  *name                     I  N/A  Name of the antenna
 * TReceiverList  *recList         I  N/A  List of receivers to look for
 * Returned value (*TReceiverData) O  N/A  Pointer to the receiver inside TReceiverList
 *                                         If it is NULL, the receiver has not been
 *                                         found
 *****************************************************************************/
TReceiverData *getReceiverType (char *name, TReceiverList *recList) {
	int 	i;
	
	for (i=0;i<recList->n;i++) {
		if (strcmp(name,recList->rec[i].name)==0) return &recList->rec[i];
	}
	
	return NULL;
}

/*****************************************************************************
 * Name        : modelEpoch
 * Description : Model the common part of all satellites for a specific epoch 
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  IO N/A  TEpoch structure
 * TOptions  *options              I  N/A  TOptions structure
 * Returned value (int)            I  N/A  Status of the function
 *                                         1 => No errors found
 *                                         Right now this function does not find
 *                                         any errors, thence always returns 1
 *****************************************************************************/
int modelEpoch(TEpoch *epoch, TOptions *options) {
	// Station NEU coordinates 
	fillGroundStationNEU(epoch);
	
	// Station orientation
	fillGroundStationOrientation(epoch);
	
	// Compute tides displacements
	// Solid tides
	if (options->solidTidesCorrection) {
		fillSolidTideDisplacement (epoch);
	}
	
	return 1;
}
 
/*****************************************************************************
 * Name        : modelSatellite
 * Description : Model the satellite-dependant corrections for a specific epoch
 *               and satellite
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  IO N/A  TEpoch structure
 * int  satIndex                   I  N/A  Index of the satellite to model
 * TGNSSproducts  *products        I  N/A  Orbit and clock GNSS products to obtain the model
 * TGNSSproducts  *klbProd         I  N/A  Support products to use for Klobuchar modeling, when
 * TGNSSproducts  *beiProd         I  N/A  Support products to use for BeiDou modelling, when
 *                                         applicable
 * TGNSSproducts  *neqProd         I  N/A  Support products to use for NeQuick modelling, when
 *                                         applicable
 * TIONEX *Ionex                   I  N/A  IONEX structure, when applicable
 * TFPPPIONEX *Fppp                I  N/A  FPPP structure, when applicable
 * TTROPOESA *TropoESA             I  N/A  TTROPOESA structure, when applicable
 * TTGDdata  *tgdData              I  N/A  Support products to use for TGD correction, when
 *                                         applicable
 * TConstellation  *constellation  I  N/A  Constellation information of the products
 * TOptions  *options              I  N/A  TOptions structure
 * Returned value (int)            I  N/A  Status of the function
 *                                         0 => Any problem found in the modeling
 *                                              (something is missing)
 *                                         1 => Model properly computed
 *****************************************************************************/
int modelSatellite(TEpoch *epoch, int satIndex, TGNSSproducts *products, TGNSSproducts *klbProd, TGNSSproducts  *beiProd, TGNSSproducts *neqProd, TIONEX *Ionex, TFPPPIONEX *Fppp, TTROPOESA *TropoESA, TTGDdata *tgdData, TConstellation *constellation, TOptions *options) {
	double	flightTime;
	double	tropWetMap;
	double	relativityModel;
	double	windUpRadModel;
	double	windUpMetModel;
	double	troposphereModel;
	double	ionosphereModel;
	double	ionosphereModelElectrons;
	double	gravitationalDelayModel;
	double	TGDModel;
	double	solidTidesModel;
	double	satelliteClockModel;
	double	satellitePhaseCenterProjection;
	double  satellitePhaseCenterVarProjection;
	double	receiverPhaseCenterProjection;
	double	receiverARPProjection;
	double	model;
	double	lambda;
	double	geometricDistance;
	double  satearthcentredistance;
	TTime	t;
	int		i;
	int		res1,res2;
	enum	MeasurementType	meas;
	
	// Model not available for non GPS satellites
	if (epoch->sat[satIndex].GNSS != GPS) return 0;
	
	// Check if satellite has been deselected (do not compute any model)
	if (!options->includeSatellite[epoch->sat[satIndex].GNSS][epoch->sat[satIndex].PRN])
		return 0;
	
	res1 = fillTransmissionTimeSat(epoch,products,satIndex,options);
	
	if (res1==0) return 0;
	
	if (options->timeTrans) {
		memcpy(&t,&epoch->sat[satIndex].transTime,sizeof(TTime));
	} else {
		memcpy(&t,&epoch->t,sizeof(TTime));
	}
	flightTime = computeFlightTime(epoch,products,satIndex,options);

	res2 = fillSatellitePVT(epoch,products,&t,flightTime,satIndex,1,1,options->satelliteClockCorrection,constellation,options);
	
	// Stop computing models if any problem had arisen to save computing time
	if (res2==0) return 0;
	
	// Line of Sight
	fillLineOfSight(epoch,satIndex);
	
	// Azimuth and elevation
	fillAzimuthElevation(epoch,satIndex);
	
	// Models	
	// Geometric distance
	fillGeometricDistance(epoch,satIndex);
	geometricDistance = epoch->sat[satIndex].geometricDistance;

	// Satellite clock correction
	if (options->satelliteClockCorrection) {
		satelliteClockModel = epoch->sat[satIndex].clockCorrection;
	} else {
		satelliteClockModel = 0;
	}
	
	// Relativistic correction
	if (options->relativisticCorrection) {
		relativityModel = relativisticCorrection(epoch->sat[satIndex].position,epoch->sat[satIndex].velocity);
	} else {
		relativityModel = 0;
	}
	
	// Wind up correction
	if (options->windUpCorrection) {
		windUpRadModel = windUpCorrection(epoch,satIndex);
	} else {
		windUpRadModel = 0;
	}
	
	// Tropospheric correction
	if (options->troposphericCorrection ) {
		troposphereModel = troposphericCorrection(&epoch->t, epoch->receiver.aproxPositionNEU, epoch->sat[satIndex].elevation, TropoESA, &tropWetMap, &epoch->ZTD,options);
		epoch->sat[satIndex].tropWetMap = tropWetMap;
	} else {
		troposphereModel = 0;
		epoch->sat[satIndex].tropWetMap = 0;
	}
	
	// Ionospheric correction
	if (options->ionoModel == Klobuchar) {
		ionosphereModelElectrons = klobucharModel(epoch, klbProd, satIndex);
	} else if (options->ionoModel == BeiDou) {
		ionosphereModelElectrons = beiDouKlobucharModel(epoch, beiProd, satIndex);
	} else if (options->ionoModel == IONEX) {
		ionosphereModelElectrons = ionexModel(epoch, satIndex, Ionex, options); 
		if (ionosphereModelElectrons == 9999 ) return 0;
	} else if (options->ionoModel == FPPP) {
		ionosphereModelElectrons = FPPPModel(epoch, satIndex, Fppp, options);
		if (ionosphereModelElectrons == 9999 ) return 0;	
        } else if (options->ionoModel == neQuick) {
                ionosphereModelElectrons = 0;
                if (ionosphereModelElectrons == 9999 ) return 0;
	} else {
		ionosphereModelElectrons = 0;
	}
	// Gravitational delay correction
	if (options->gravitationalCorrection) {
		gravitationalDelayModel = gravitationalDelayCorrection(epoch->receiver.aproxPosition,epoch->sat[satIndex].position);
	} else {
		gravitationalDelayModel = 0;
	}
	
	// Solid tides correction
	if (options->solidTidesCorrection) {
		solidTidesModel = solidTidesCorrection(epoch,satIndex);
	} else {
		solidTidesModel = 0;
	}
	
	// Fill measurement weights
	fillMeasurementWeights(epoch,satIndex,options);

	for (i=0;i<epoch->measOrder[epoch->sat[satIndex].GNSS].nDiffMeasurements;i++) {
		// Measurement dependant effects
		meas = epoch->measOrder[epoch->sat[satIndex].GNSS].ind2Meas[i];
		
		// Ionosphere model
		lambda = getLambda(epoch->sat[satIndex].GNSS,meas);

		ionosphereModel = ionosphereModelElectrons*40.3*(lambda*lambda)/(c*c)*pow(10,16);

		// Satellite phase center correction for SP3 products (may depend in the frequency)
		if (options->satellitePhaseCenter) {
			if (constellation!=NULL && products->type == SP3) {
				//res1 = satellitePhaseCenterCorrection (&epoch->t, epoch->sat[satIndex].GNSS, epoch->sat[satIndex].PRN, epoch->measOrder[epoch->sat[satIndex].GNSS].ind2Meas[i], epoch->sat[satIndex].orientation, epoch->sat[satIndex].LoS, constellation, &satellitePhaseCenterProjection);
				res1 = satellitePhaseCenterCorrection (&epoch->t, epoch->sat[satIndex].GNSS, epoch->sat[satIndex].PRN, meas, epoch->sat[satIndex].orientation, epoch->sat[satIndex].LoS, constellation, &satellitePhaseCenterProjection);
				
				if (!res1) {
					epoch->sat[satIndex].hasOrbitsAndClocks = 0;	// No constellation information
					return 0; 
				}
                                if(options->satellitePhaseVar==1) {
                                        satearthcentredistance=sqrt(epoch->sat[satIndex].position[0]*epoch->sat[satIndex].position[0]+epoch->sat[satIndex].position[1]*epoch->sat[satIndex].position[1]+epoch->sat[satIndex].position[2]*epoch->sat[satIndex].position[2]);
                                        res1=satellitePhaseCenterVarCorrection (&epoch->t,epoch->sat[satIndex].GNSS,epoch->sat[satIndex].PRN,meas,satearthcentredistance,epoch->sat[satIndex].elevation,constellation, &satellitePhaseCenterVarProjection);
                                        if(res1) {
                                                satellitePhaseCenterProjection+=satellitePhaseCenterVarProjection;
                                        }
                                }
			} else {
				satellitePhaseCenterProjection = 0;
			}
		} else {
			satellitePhaseCenterProjection = 0;
		}
		
		// Receiver antenna phase center
		if (options->antennaData == adSET || options->antennaData == adANTEX) {
			receiverPhaseCenterProjection = receiverPhaseCenterCorrection(&epoch->receiver.antenna,epoch->sat[satIndex].GNSS,meas,epoch->receiver.orientation,epoch->sat[satIndex].LoS);
                        if(options->receiverPhaseVar==1 && options->antennaData == adANTEX) {
                                receiverPhaseCenterProjection += receiverPhaseCenterVarCorrection(&epoch->receiver.antenna,epoch->sat[satIndex].GNSS,meas,epoch->sat[satIndex].elevation);
                        }

		} else {
			receiverPhaseCenterProjection = 0;
		}
		
		// Receiver Antenna Reference Point (ARP)
		if (options->ARPData == arpSET || options->ARPData == arpRINEX) {
			receiverARPProjection = receiverARPCorrection (epoch->receiver.ARP,epoch->receiver.orientation,epoch->sat[satIndex].LoS);
		} else {
			receiverARPProjection = 0;
		}
			
		// TGD Model
//		if (options->p1c1dcbModel != p1c1NONE) {
			epoch->sat[satIndex].hasDCBs = 1;
			epoch->sat[satIndex].hasDCBsofC1P1 = 1;
			TGDModel = TGDCorrection(epoch,satIndex,&t,meas,tgdData,options);
//		} else {
//			TGDModel = 0;
//		}
		
		// Modeling computing
		switch (whatIs(meas)) {
			case CarrierPhase:
				windUpMetModel = windUpRadModel/(2*Pi)*lambda;
				ionosphereModel = -ionosphereModel;
				model = geometricDistance + satelliteClockModel + satellitePhaseCenterProjection + receiverPhaseCenterProjection + receiverARPProjection + relativityModel + windUpMetModel + troposphereModel + ionosphereModel + gravitationalDelayModel + TGDModel + solidTidesModel;
				break;
			case Pseudorange:
				windUpMetModel = 0;
				model = geometricDistance + satelliteClockModel + satellitePhaseCenterProjection + receiverPhaseCenterProjection + receiverARPProjection+ relativityModel + troposphereModel + ionosphereModel + gravitationalDelayModel + TGDModel + solidTidesModel;
				break;
			default:
				model = 0;
				continue;
				break;
		}
		epoch->sat[satIndex].meas[i].model = model;
		
		printModel(epoch, satIndex, i, flightTime, geometricDistance, satelliteClockModel, satellitePhaseCenterProjection, receiverPhaseCenterProjection, receiverARPProjection, relativityModel, windUpMetModel, troposphereModel, ionosphereModel, gravitationalDelayModel, TGDModel, solidTidesModel, options);
	}

	return 1;
}
